package cpmodel;

import choco.Solver;
import choco.ContradictionException;
import choco.integer.search.*;
import choco.integer.IntDomainVar;
import instance.Precedence;

import java.util.Arrays;

import magicsearch.impact.ImpactStrategie;
import magicsearch.impact.Iref;
import magicsearch.impact.impactMAC.IBSProblem;
import magicsearch.impact.impactMAC.IBSAssignVar;
import magicsearch.impact.impactMAC.ImpactBasedVarSearch;
import magicsearch.impact.impactMAC.ImpactBasedValSearch;
import magicsearch.domwdeg.domwdegMAC.DWDAssignVar;
import magicsearch.domwdeg.DomOverWdeg;

/**
 * The first "naive" model
 */
public class MediumModel extends PrimalModel {


    public MediumModel(String s) {
        super(s);
    }

    protected IntDomainVar[] cstR;

    public void buildModel() {
        // Coloring + simple allocation
        super.buildModel();
        this.buildColoring(PrimalModel.CONFLICTS);
        IntDomainVar maxRooms = pb.makeBoundIntVar("MR", ins.getNRooms(), ins.getNRooms());
        for (int i = 0; i < nHours * nDays; i++) {
            pb.post(pb.occurenceMax(eventTime, (i + 1), maxRooms));
        }
    }

    /**
     * build the coloring problematic common to all models involving
     * the et variables
     */
    public void buildColoring(int coltype) {
        // Takes care of the graph coloring + precedence aspect
        // The graph coloring
        eventTime = new IntDomainVar[ins.getNEvents()];
        for (int i = 0; i < ins.getNEvents(); i++) {
            eventTime[i] = pb.makeEnumIntVar("Et" + (i + 1), 1, nDays * nHours);
        }


        postConflicts(coltype);

        // 4. Lecturer's availability
        for (int i = 0; i < ins.getNEvents(); i++) {
            for (int j = 0; j < nDays * nHours; j++) {
                if (!ins.eventIsCompatibleWithTimeslot(i, j)) {
                    try {
                        eventTime[i].remVal(j + 1);                      
                    } catch (ContradictionException e) {
                        throw new Error("buildColoring: inconsistent initial model ??");
                    }
                }
            }
        }
//        // 5. Precedences
        for (Precedence p : ins.eventPrecedences()) {
            pb.post(pb.lt(eventTime[p.firstEvent], eventTime[p.secondEvent]));
            //System.out.println(eventTime[p.firstEvent] + " < " + eventTime[p.secondEvent]);
        }
    }


    public void solve() {
        // Solving
        super.solve();
        Solver s = pb.getSolver();
        setHeuristic();
        s.launch();
        int nbSol = 0; // DG Added for all solution search
        if (pb.isFeasible() == Boolean.FALSE)
            System.out.println("no solution");
        else {
            do {
                int[] evtPerT = new int[46];
                for (int i = 0; i < ins.getNEvents(); i++) {
                    System.out.println("event " + i + " : " + eventTime[i].getVal());
                    evtPerT[eventTime[i].getVal()]++;
                }
                System.out.println(Arrays.toString(evtPerT));
                //checkSolution();
                //printFullTimetable();
                nbSol++;

                if (nbSol % 1000 == 0) System.out.println("bibi " + nbSol);
            } while (nbSol < 10 && pb.nextSolution());
        }
        System.out.println("NB solution found " + pb.getSolver().getNbSolutions());
        int nodes = pb.getSolver().getSearchSolver().getNodeCount(); // get nodes
        //            pb.getSolver().getSearchSolver().getTimeCount();
        System.out.println("Nodes: " + nodes + " " + pb.getSolver().getSearchSolver().getTimeCount());
    }


    public void setHeuristic() {
        Solver s = pb.getSolver();
        s.generateSearchSolver(pb);
        if (heuristictype == Impact) {
            restart = false;
            initialization = true;
            ImpactStrategie ibs = new Iref(pb, eventTime);
            ((IBSProblem) pb).setIbs(ibs);
            if (initialization) {
                long tps = System.currentTimeMillis();
                ((IBSProblem) pb).initImpact(timelimit);
                inittime = (int) (System.currentTimeMillis() - tps);
                System.out.println("initialization time " + inittime);
            }

            s.attachGoal(new IBSAssignVar(
                    new ImpactBasedVarSearch(eventTime, ibs),
                    new ImpactBasedValSearch(ibs)));
        } else if (heuristictype == DomWDeg) {
            s.attachGoal(new DWDAssignVar(new DomOverWdeg(pb, eventTime), new IncreasingDomain()));
        } else {
           // s.attachGoal(new AssignVar(new DomOverDeg(pb, eventTime), new IncreasingDomain()));
             s.attachGoal(new AssignVar(new DomOverDeg(pb, eventTime), new RandomIntValSelector(0)));
        }
    }

    public static void main(String[] args) {
        MediumModel pm = new MediumModel("data/Track2/comp-2007-2-8.tim");
        //MediumModel pm = new MediumModel("data/medium2008/medium01.tim");//,LightModel.Impact);
        //LightModel pm = new LightModel("data/Track2/comp-2007-2-4.tim", 2);

        //LightModel pm = new LightModel("data/test/comp-2007-2-2-5-1.tim", 2);

        pm.nDays = 5;
        pm.nHours = 9;
        pm.heuristictype = PrimalModel.Impact;

        pm.buildModel(); //PrimalModel.CONFLICTS);
        pm.solve();
        System.out.println("" + pm.getNode());
        //pm.printColoring();
        //pm.printFullTimetable();
        //pm.checkSolution();
    }


}
