package cpmodel.chocoext;

import choco.Problem;
import choco.integer.IntDomainVar;

import java.util.Arrays;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: 10 sept. 2007
 * Time: 14:32:13
 * To change this template use File | Settings | File Templates.
 */
public class TimeTablingProblem extends Problem {


	protected IntDomainVar createIntVar(String name, int domainType, int min, int max) {
		return new IndexedIntVar(this, name, domainType, min, max);
	}

	protected IntDomainVar createIntVar(String name, int[] sortedValues) {
		return new IndexedIntVar(this, name, sortedValues);
	}


	public IntDomainVar makeIntVar(String name, int domainType, int min, int max) {
		 IntDomainVar v = createIntVar(name, domainType, min, max);
		 intVars.add(v);
		 ((IndexedIntVar) v).setIdx(intVars.size() - 1);
		 return v;
	 }

	 /**
	  * Creates a new search variable with an enumerated domain
	  *
	  * @param name the name of the variable
	  * @param min  minimal allowed value (included in the domain)
	  * @param max  maximal allowed value (included in the domain)
	  * @return the variable
	  */
	 public IntDomainVar makeEnumIntVar(String name, int min, int max) {
		 IntDomainVar v = createIntVar(name, IntDomainVar.BITSET, min, max);
		 intVars.add(v);
		 ((IndexedIntVar) v).setIdx(intVars.size() - 1);
		 return v;
	 }

	 /**
	  * Creates a new search variable with an enumerated domain
	  *
	  * @param name   the name of the variable
	  * @param values allowed in the domain (may be unsorted, but not with duplicates !)
	  * @return the variable
	  */
	 public IntDomainVar makeEnumIntVar(String name, int[] values) {
		 int[] values2 = new int[values.length];
		 System.arraycopy(values, 0, values2, 0, values.length);
		 Arrays.sort(values2);
		 IntDomainVar v = createIntVar(name, values2);
		 intVars.add(v);
		 ((IndexedIntVar) v).setIdx(intVars.size() - 1);
		 return v;
	 }

	 /**
	  * Creates a new search variable with an interval domain
	  *
	  * @param name the name of the variable
	  * @param min  minimal allowed value (included in the domain)
	  * @param max  maximal allowed value (included in the domain)
	  * @return the variable
	  */
	 public IntDomainVar makeBoundIntVar(String name, int min, int max) {
		 IntDomainVar v = createIntVar(name, IntDomainVar.BOUNDS, min, max);
		 intVars.add(v);
		 ((IndexedIntVar) v).setIdx(intVars.size() - 1);
		 return v;
	 }

	 public IntDomainVar makeBoundIntVar(String name, int min, int max, boolean toAdd) {
		 IntDomainVar v = createIntVar(name, IntDomainVar.BOUNDS, min, max);
		 if (toAdd) {
			 intVars.add(v);
			 ((IndexedIntVar) v).setIdx(intVars.size() - 1);
		 }
		 return v;
	 }

}
