package cpmodel.constraints;

import java.util.Arrays;

import choco.ContradictionException;
import choco.integer.IntDomainVar;
import choco.integer.constraints.AbstractLargeIntConstraint;



/**
 * A single constraint to maintain all the differences
 */
public class Colouring extends AbstractLargeIntConstraint {

	//the graph of incompatibilities
	protected int[][] igraph;

	protected int[] nbneighbours;

	//temporary data
	protected int tcpt;

	public Colouring(IntDomainVar[] vars) {
		super(vars);
		igraph = new int[vars.length][vars.length + 1];
		nbneighbours = new int[vars.length];
		for (int i = 0; i < vars.length; i++) {
			Arrays.fill(igraph[i],-1);
		}
	}

	//todo: this more properly once alex will have the graph data structure
	public void setIncomp(int i, int j) {
		nbneighbours[i] ++;
		nbneighbours[j] ++;
		int cpt = 0;
		while(igraph[i][cpt] != -1) {
			cpt++;
		}
		igraph[i][cpt] = j;
		cpt = 0;
		while(igraph[j][cpt] != -1) {
			cpt++;
		}
		igraph[j][cpt] = i;
	}

	public int getNbN(int i) {
		return nbneighbours[i];
	}

	public void propagate() throws ContradictionException {
		for (int i = 0; i < vars.length; i++) {
			if (vars[i].isInstantiated())
				awakeOnInst(i);
		}
	}

	public void awakeOnInf(int idx) throws ContradictionException {

	}

	public void awakeOnSup(int idx) throws ContradictionException {

	}

	public void awakeOnInst(int idx) throws ContradictionException {
		int val = vars[idx].getVal();
		tcpt = 0;
		while(igraph[idx][tcpt] != -1) {
			vars[igraph[idx][tcpt]].removeVal(val,-1);
			tcpt++;
		}
	}

	public void awakeOnRem(int idx, int x) throws ContradictionException {

	}

	public boolean isSatisfied() {
		throw new Error("is satisfied not yet implemented on Coloring");
	}
}
