package cpmodel.constraints;

import instance.TimetablingInstance;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.util.Arrays;
import java.util.BitSet;

import tools.InstanceReader;
import choco.ContradictionException;
import choco.Problem;
import choco.util.IntIterator;
import choco.integer.IntDomainVar;
import choco.integer.constraints.AbstractLargeIntConstraint;
import choco.integer.search.RandomIntValSelector;
import choco.integer.search.RandomIntVarSelector;
import choco.mem.IStateBitSet;
import choco.mem.IStateInt;
import choco.mem.trailing.EnvironmentTrailing;
import choco.mem.trailing.StoredBitSet;
import choco.mem.trailing.StoredInt;


/**
 * The constraint that maintains the first cost
 * (number of students in a final timeslot).
 * 
 * @author alexp
 *
 */
public class Cost3Constraint extends AbstractLargeIntConstraint {

//	protected int[][] maxCount;
//	protected int[][] minCount;
//	protected boolean[] canGo;
//	protected boolean[] mustGo;
//	protected int lowBounb;
//	protected int uppBounb;
//	
//	protected void computeBounds()
//	{
////		int day, stu, evt, nStu = instance.getNStudents(), nEvt = instance.getNEvents();
////		for( day=0; day<5; ++day )
////		{
////			for( stu=0; stu<nStu; ++stu )
////				maxCount[day][stu] = 0;
////		}
////		
////		for( evt=0; evt<nEvt; ++evt )
////		{
////			for( day=0; day<5; ++day )
////				canGo[day] = canGoInDay(evt, day);
////					
////			stu = -1;
////			BitSet students = instance.studentsAttendingEvent(evt);
////			do {
////				stu = students.nextSetBit(stu+1);
////				for( day=0; day<5; ++day )
////					if( canGo[day] )
////						++maxCount[day][stu];
////			}
////		}	
//	}
	
	protected void setEventDayTo( final int evt, final int day )
	{
		n_lost_days = 0;
		for(int i=0; i<5; ++i) 
			if( i != day && eventDay[evt].get(i) ) {
				lost_days[n_lost_days++] = i;
				eventDay[evt].clear(i);
		}
		eventDay[evt].set(day);
	}
	
	protected int getStudent( final int pointer )
	{
		return ( pointer >> 1 );
	}
	
	protected int getW( final int pointer )
	{
		return ( pointer & 1 );
	}
	
	protected void add( final int evt, final int day, final int stu, final int w )
	{
		watchers[evt][day][nWatcher[evt][day]++] = (stu << 1)+w;
		watchedBy[w][day][stu] = evt;
	}
	
	protected void remove( final int evt, final int day, final int idx )
	{
//		if( nWatcher[evt][day] > watchers[evt][day].length )
//			printWatchedStudents( evt, day );
		watchers[evt][day][idx] = watchers[evt][day][--nWatcher[evt][day]];
	}
	
	protected boolean cannotGoInDay( final int evt, final int day )
	{	
		int val = events[evt].getDomain().getNextValue(day*9);
		return ( val > (day+1)*9);
	}
	
	protected boolean canGoInDay( final int evt, final int day )
	{	
		return eventDay[evt].get(day);
	}
	
	protected boolean mustGoInDay( final int evt, final int day )
	{
		return (eventDay[evt].get(day) && eventDay[evt].cardinality() == 1);
	}
	
	protected int findNewWatcher( final int stu, final int day, final int old_evt, final int w )
	{		
		BitSet the_events = instance.eventsAttendedByStudent(stu);
		int evt = -1, second_old_evt = watchedBy[1-w][day][stu];
		do{
			evt = the_events.nextSetBit(evt+1);
		} while( evt >= 0 &&
				(evt == old_evt ||
				 evt == second_old_evt ||
				 !canGoInDay(evt, day)) );
//		
//		assert( evt != old_evt && evt != second_old_evt && 
//				(evt < 0 || (canGoInDay(evt, day) && instance.studentAttendsEvent(stu, evt))) );
		
		return evt;
	}

	protected void printWatchedStudents( final int evt, final int day )
	{
		System.out.print( "evt" + evt + " watches " );
		for(int i=0; i<nWatcher[evt][day]; ++i)
			System.out.print( getStudent(watchers[evt][day][i]) + " " );
		System.out.println();
	}
	
	/**
	 * The variables corresponding to the event timeslots,
	 */
	protected IntDomainVar[] events;
	/**
	 * The variable corresponding to the cost.
	 */
	protected IntDomainVar cost;
	
	protected TimetablingInstance instance;
	
	// watchers
	protected int[][][]  watchers;
	protected int[][]    nWatcher;
	protected int[][][] watchedBy;

	// 
	protected StoredInt reversible_lower_bound;
	protected StoredBitSet[] eventDay;

	// buffers
	protected int[]     lost_days;
	protected int     n_lost_days;

	

	public Cost3Constraint(IntDomainVar[] events, IntDomainVar cost, TimetablingInstance instance) {
		
		super(toSingleArray(cost, events));	
		
		this.instance = instance;
		this.events = events;
		this.cost = cost;
		
		eventDay = new StoredBitSet[instance.getNEvents()];
		for( int i=0; i<instance.getNEvents(); ++i ) {
			eventDay[i] = new StoredBitSet( (EnvironmentTrailing) this.getProblem().getEnvironment(), 5 );
			for( int j=0; j<5; ++j )
				eventDay[i].set(j);
		}
		
		
		int maxStudentsPerEvent = 0, nEvents = instance.getNEvents(), nStudents = instance.getNStudents();
		
		for(int i=0; i<nEvents; ++i) {
			if( maxStudentsPerEvent < instance.getNbStudentInEvent(i) )
				maxStudentsPerEvent = instance.getNbStudentInEvent(i);
		}
		
		int e1, e2;
	
		lost_days = new int[5];
		watchers = new int[nEvents][5][maxStudentsPerEvent];
		nWatcher = new int[nEvents][5];
		watchedBy = new int[2][5][nStudents];
		reversible_lower_bound = new StoredInt( (EnvironmentTrailing) this.getProblem().getEnvironment(), 0 );
		
		for(int i=0; i<nStudents; ++i)
		{
			e1 = instance.eventsAttendedByStudent(i).nextSetBit(0);
			e2 = instance.eventsAttendedByStudent(i).nextSetBit(e1+1);
			
			for(int j=0; j<5; ++j) 
			{
				add( e1, j, i, 0 );
				add( e2, j, i, 1 );
			}
		}
	}


	protected static IntDomainVar[] toSingleArray(IntDomainVar var, IntDomainVar[] vars){
		IntDomainVar[] array = new IntDomainVar[vars.length+1];
		System.arraycopy(vars, 0, array, 0, vars.length);
		array[vars.length] = var;
		return array;
	}
	

	public boolean isSatisfied() {
		return false;
	}
	
	public void printDays( int evt )
	{
		for(int i=0; i<5; ++i)
			if( eventDay[evt].get(i) )
				System.out.print( i );
			else	
				System.out.print( " " );
	}

	
	
	@Override
	public void awakeOnRemovals(int arg0, IntIterator arg1)
			throws ContradictionException {
		awakeOnVar(arg0);
	}

	public void awakeOnInst(int evt) throws ContradictionException {
//		if( evt < events.length )
//		{
//			if( eventDay[evt].cardinality() > 1 )
//			{
//				n_lost_days=0;
//				int day = ((events[evt].getVal()-1) / 9);
//				setEventDayTo( evt, day );
//				throwLostDays( evt );
//				throwInstDay( evt, day );
//			} 
//		}
		awakeOnVar(evt);

	}
	
	
	
	@Override
	public void awakeOnBounds(int arg0) throws ContradictionException {
		// TODO Auto-generated method stub
		awakeOnVar(arg0);
	}

	public void awakeOnVar(int evt) throws ContradictionException {
		if (evt < events.length) {			
			if( eventDay[evt].cardinality() > 1 )
			{
				n_lost_days=0;
				int remaining_day = -1;
				for(int i=0; i<5; ++i)
				{
					if( eventDay[evt].get(i) && cannotGoInDay(evt, i) ) {
						lost_days[n_lost_days++] = i;
						eventDay[evt].clear(i);
					}
					
					if( eventDay[evt].get(i) ) {
						if( remaining_day < 0 )
							remaining_day = i;	
						else 
							remaining_day = 5;
					}
				}

				if( n_lost_days > 0 ) {
					throwLostDays( evt );
					if( remaining_day < 5 )
						throwInstDay( evt, remaining_day );
				} 
			} 	
		}
	}
	
	
//	public void awakeOnInf(int evt) throws ContradictionException {
//		if (evt < events.length) {
//		
//			int lb = ((events[evt].getInf()-1) / 9);
//			
//			System.out.println( events[evt].getInf() + " l-> " + lb );
//			checkB( lb, events[evt].getInf() );
//			
//			n_lost_days=0;
//			for(int i=0; i<lb; ++i)
//			{
//				if( eventDay[evt].get(i) ) {
//					lost_days[n_lost_days++] = i;
//					eventDay[evt].clear(i);
//				}
//			}
//			if( n_lost_days > 0 ) {
//				throwLostDays( evt );
//			} 
//		} 	
//	}
//	
//	
//	protected void checkB(int b, int v)
//	{
//		assert( v > 0  && v <= 45 );
//		assert( b >=0 && b < 5 );
//		assert( b != 0 || ( v > 0 && v < 10 ) );
//		assert( b != 1 || ( v > 9 && v < 19 ) );
//		assert( b != 2 || ( v > 18 && v < 28 ) );
//		assert( b != 3 || ( v > 27 && v < 37 ) );
//		assert( b != 4 || ( v > 36 && v < 46 ) );
//		
//	}
//	
//	public void awakeOnSup(int evt) throws ContradictionException {
//		if (evt < events.length) {
//			
//			int ub = ((events[evt].getSup()-1) / 9);
//			
//			System.out.println( events[evt].getSup() + " u-> " + ub );
//			checkB( ub, events[evt].getSup() );
//			
//			n_lost_days=0;
//			for(int i=ub+1; i<5; ++i)
//			{
//				if( eventDay[evt].get(i) ) {
//					lost_days[n_lost_days++] = i;
//					eventDay[evt].clear(i);
//				}
//			}
//			if( n_lost_days > 0 ) {
//				throwLostDays( evt );
//			} 
//		} 
//	}
	
	
	public void throwInstDay(int evt, int day) throws ContradictionException {
		
		int stu, w, i=nWatcher[evt][day];
		while( --i >= 0 )
		{
			stu = getStudent(watchers[evt][day][i]);
			w = getW(watchers[evt][day][i]);
			if( !canGoInDay( watchedBy[1-w][day][stu], day ) )
				reversible_lower_bound.add(1);
		}
		cost.setInf( reversible_lower_bound.get() );
	}
	

	public void throwLostDays(int evt) throws ContradictionException {
		int new_evt, day, stu, w, j, i;
		for(j=0; j<n_lost_days; ++j)
			{
				day = lost_days[j];
				i = nWatcher[evt][day];
				while( --i >= 0 )
				{
					stu = getStudent(watchers[evt][day][i]);
					w = getW(watchers[evt][day][i]);
					new_evt = findNewWatcher(stu, day, evt, w);
						
					//assert( new_evt != evt );
				
					if( new_evt >= 0 ) {
					
						remove( evt, day, i );
						add( new_evt, day, stu, w );
					
					} else {
						// there is no more watcher
						if( mustGoInDay( watchedBy[1-w][day][stu], day ) )
							reversible_lower_bound.add(1);
					}
				}
			}	
		cost.setInf( reversible_lower_bound.get() );
	}
}
		

// the assumption (invariant?) is that if a pair (student, day)
// is 'watched' by an event, then this event can possibly
// happen on that day (and the student goes to that event).

// Therefore, when a watcher is no longer available,
// if we fail to find a new watcher *and* if the second
// watcher-event must happen on this day, this is the
// one and only event for that student and that day.
// On the other hand, when the remaining event can
// be assigned to another day, we must wait for it to
// get assigned.

// Hence we also check the case when a watcher-event
// is 'stuck' to the day it watches. If the other
// watcher event cannot go to that day, then this
// is also the one and only event for that student/day.



// the domain of events[idx] has changed, we go through 
// the pairs (day, students) watched by this event and remove 
// them if 'day' is not possible anymore

	
