package cpmodel.constraints;

import choco.integer.constraints.AbstractTernIntConstraint;
import choco.integer.IntDomainVar;
import choco.integer.search.RandomIntVarSelector;
import choco.integer.search.RandomIntValSelector;
import choco.ContradictionException;
import choco.Problem;
import choco.util.IntIterator;

import java.util.Iterator;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: 15 aot 2007
 * Time: 09:34:42
 * To change this template use File | Settings | File Templates.
 */
public class ModelChannel extends AbstractTernIntConstraint {

    protected int a;
    protected int b;

    /**
     * Enforce a channeling between the following variables :
     * x0 = {1,...,ab}
     * x1 = {1,...,a}
     * x2 = {1,...,b}
     * Message to <alex> : peux tu quand meme avoir l'option regular pour maintenir
     * le channeling de maniere a verifier que je fais bien l ac avec cette contrainte en regardant
     * le nombre de noeud par rapport a la regular. thx
     * Elle devrait etre utilisable comme ca (tests random dans le main) mais il faut que je reflechisse un peu pour voir
     * si je peux pas simplifier des trucs :)
     * @param x0
     * @param x1
     * @param x2
     */
    public ModelChannel(final IntDomainVar x0, final IntDomainVar x1, final IntDomainVar x2, int a, int b) {
        super(x0, x1, x2);
        this.a = a;
        this.b = b;
	    //System.out.println("" + x0.pretty() + " " + x1.pretty() + " " + x2.pretty());
	    assert (x0.getSup() == a * b);
    }


    public void propagate() throws ContradictionException {
	    for (int i = 1; i <= a * b; i++) {
            if (!v0.canBeInstantiatedTo(i)) {
                awakeOnRem(0, i);
            }
        }
        for (int i = 1; i <= a; i++) {
            if (!v1.canBeInstantiatedTo(i)) {
                awakeOnRem(1, i);
            }
        }
        for (int i = 1; i <= b; i++) {
            if (!v2.canBeInstantiatedTo(i)) {
                awakeOnRem(2, i);
            }
        }
    }

    public void awakeOnInf(int idx) throws ContradictionException {
        int k;
        if (idx == 0) {
            k = v0.getInf();
            v1.updateInf(k - (b - 1) * a, cIdx1);
            v2.updateInf((int) Math.ceil((double) k / (double) a), cIdx2);
        } else if (idx == 1) {
            k = v1.getInf();
            for (int i = 1; i < k; i++) {
                awakeOnRem(1, i);
            }
        } else {
            k = v2.getInf();
            v0.updateInf((k - 1) * a + 1, cIdx0);
        }
    }

    public void awakeOnSup(int idx) throws ContradictionException {
        int k;
        if (idx == 0) {
            k = v0.getSup();
            v1.updateSup(k, cIdx1);
            v2.updateSup((int) Math.ceil((double) k / (double) a), cIdx2);
        } else if (idx == 1) {
            k = v1.getSup();
            for (int i = k + 1; i <= a; i++) { //todo: improve that to avoid taking known removal into account
                awakeOnRem(1, i);
            }
        } else {
            k = v2.getSup();
            v0.updateSup(k * a, cIdx0);
        }
    }

    public void awakeOnInst(int idx) throws ContradictionException {
        int k;
        if (idx == 0) {
            k = v0.getVal();
            int v1val = k % a;
            if (v1val == 0) v1val = a;
            v1.instantiate(v1val, cIdx1);
            v2.instantiate((int) Math.ceil((double) k / (double) a), cIdx2);
        } else if (idx == 1) {
            k = v1.getVal();
            IntIterator it = v0.getDomain().getIterator();
            while (it.hasNext()) {
                int v = it.next();
                if ((v-k) % a != 0) v0.removeVal(v, -1);
            }
        } else {
            k = v2.getVal();
            v0.updateInf((k - 1) * a + 1, cIdx0);
            v0.updateSup(k * a, cIdx0);
        }
    }

    public void awakeOnRem(int idx, int k) throws ContradictionException {
        if (idx == 0) {
            int rest = k % a;
            if (rest == 0) rest = a;
            boolean pruneV1 = true;
            for (int i = 0; i < b; i++) {
                if (v0.canBeInstantiatedTo(i * a + rest)) {
                    pruneV1 = false;
                    break;
                }
            }
            if (pruneV1)
               v1.removeVal(rest, cIdx1);

            int div = (int) Math.ceil((double) k / (double) a);
            int lb = (div-1)*a + 1;
            int ub = div*a;
            boolean pruneV2 = true;
            for (int i = lb; i <= ub; i++) {
                if (v0.canBeInstantiatedTo(i)) {
                    pruneV2 = false;
                    break;
                }
            }
            if (pruneV2)
                v2.removeVal(div, cIdx2);
       
        } else if (idx == 1) {
            IntIterator it = v0.getDomain().getIterator();
            while (it.hasNext()) {
                int v = it.next();
                if ((v-k) % a == 0) v0.removeVal(v, -1);
            }
        } else {
            v0.removeInterval((k - 1) * a + 1, k * a, cIdx0);
        }
    }

    public String toString() {
		return "model channel : " + v0.getVal() +" = "+ v1.getVal() +" + "+ ( v2.getVal() - 1 ) * a;
    }

    public boolean isSatisfied() {
        return v0.getVal() ==  v1.getVal() + ( v2.getVal() - 1 ) * a;
    }

    public static void main(String[] args) {
        for (int seed = 0; seed <200; seed++) {
            Problem pb = new Problem();
            IntDomainVar v0 = pb.makeEnumIntVar("v0", 1, 12);
            IntDomainVar v1 = pb.makeEnumIntVar("v0", 1, 3);
            IntDomainVar v2 = pb.makeEnumIntVar("v0", 1, 4);
            pb.post(new ModelChannel(v0, v1, v2, 3,4));

            pb.getSolver().setVarSelector(new RandomIntVarSelector(pb, seed));
            pb.getSolver().setValSelector(new RandomIntValSelector(seed));
            pb.solve();
            do {
                System.out.println(v0.getVal() + " = " + v1.getVal() + " " + v2.getVal());
            } while (pb.nextSolution() == Boolean.TRUE);
            System.out.println("nbSol = " + pb.getSolver().getNbSolutions());
        }
    }
}
