package cpmodel.constraints;

import choco.integer.constraints.AbstractLargeIntConstraint;
import choco.integer.IntDomainVar;
import choco.integer.IntVar;
import choco.integer.search.RandomIntVarSelector;
import choco.integer.search.RandomIntValSelector;
import choco.ContradictionException;
import choco.Problem;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: 19 nov. 2007
 * Time: 16:20:25
 * To change this template use File | Settings | File Templates.
 */
public class NotAllEqual extends AbstractLargeIntConstraint {


	public NotAllEqual(IntDomainVar[] vars) {
		super(vars);
	}

	public void filter() throws ContradictionException {
		int firstValueFound = Integer.MIN_VALUE;
		int nbUnInstantiated = 0;
		int freeVar = -1;
		boolean filtering = true;
		for (int i = 0; i < vars.length && nbUnInstantiated <= 1; i++) {
			if (vars[i].isInstantiated()) {
				int val = vars[i].getVal();
				if (firstValueFound == Integer.MIN_VALUE) {
					firstValueFound = val;
				} else {
					if (val != firstValueFound) {
						filtering = false;
						setEntailed();
						break;
					}
				}
			} else {
				nbUnInstantiated++;
				freeVar = i;
			}
		}
		if (nbUnInstantiated == 0 && filtering) {
			fail();
		}
		if (nbUnInstantiated == 1 && filtering) {
			vars[freeVar].removeVal(firstValueFound, cIndices[freeVar]);
		}
	}

	public void propagate() throws ContradictionException {
		filter();
	}

	public void awakeOnInf(int idx) throws ContradictionException {

	}

	public void awakeOnSup(int idx) throws ContradictionException {

	}

	public void awakeOnInst(int idx) throws ContradictionException {
		filter();
	}

	public void awakeOnRem(int idx, int x) throws ContradictionException {

	}

	public boolean isSatisfied() {
		throw new Error("isSatisfied of NotAllEqual normaly not called");
	}

	public static void main(String[] args) {
		for (int i = 0; i < 100; i++) {
			Problem pb = new Problem();
			IntDomainVar[] vs = pb.makeEnumIntVarArray("tab", 5, 0, 2);
			pb.post(new NotAllEqual(vs));
			pb.getSolver().setVarSelector(new RandomIntVarSelector(pb, i));
			pb.getSolver().setValSelector(new RandomIntValSelector(i));
			pb.solveAll();
			System.out.println("nb sol " + pb.getSolver().getNbSolutions());
		}
	}
}
