package cpmodel.constraints;

import choco.integer.constraints.AbstractBinIntConstraint;
import choco.integer.IntVar;
import choco.integer.IntDomainVar;
import choco.integer.search.RandomIntVarSelector;
import choco.integer.search.RandomIntValSelector;
import choco.ContradictionException;
import choco.Problem;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: 28 aot 2007
 * Time: 17:32:44
 * To change this template use File | Settings | File Templates.
 */
public class SimpleInverse extends AbstractBinIntConstraint {

    public int i, j;

    /**
     * Enforce x = i <=> y = j
     */
    public SimpleInverse(IntDomainVar x, int i, IntDomainVar y, int j) {
        super(x, y);
        this.i = i;
        this.j = j;
    }

    public void filterOnV0() throws ContradictionException {
        if (!v0.canBeInstantiatedTo(i))
            v1.removeVal(j, cIdx1);
    }

    public void filterOnV1() throws ContradictionException {
        if (!v1.canBeInstantiatedTo(j))
            v0.removeVal(i, cIdx0);
    }

    public void propagate() throws ContradictionException {
        if (v0.isInstantiated()) awakeOnInst(0);
        if (v1.isInstantiated()) awakeOnInst(1);
        filterOnV0();
        filterOnV1();
    }

    public void awakeOnRem(int idx, int val) throws ContradictionException {
        if (idx == 0) {
            if (val == i) v1.removeVal(j, cIdx1);
        } else {
            if (val == j) v0.removeVal(i, cIdx0);
        }
    }

    public void awakeOnInf(int idx) throws ContradictionException {
        if (idx == 0) {
            filterOnV0();
        } else filterOnV1();
    }

    public void awakeOnSup(int idx) throws ContradictionException {
        if (idx == 0) {
            filterOnV0();
        } else filterOnV1();
    }

    public void awakeOnInst(int idx) throws ContradictionException {
        if (idx == 0) {
            int val = v0.getVal();
            if (val == i) v1.instantiate(j, cIdx1);
        } else {
            int val = v1.getVal();
            if (val == j) v0.instantiate(i, cIdx0);
        }
    }

    public boolean isSatisfied() {
        if (v0.getVal() == i) return v1.getVal() == j;
        if (v1.getVal() == j) return v0.getVal() == i;
        if (v0.getVal() != i) return v1.getVal() != j;
        if (v1.getVal() != i) return v0.getVal() != j;
        return false;
    }

	public String toString() {
		return v0 + " == " + i + " <=> " + v1 + " == " + j;
	}

    public static void main(String[] args) {
        for (int seed = 0; seed < 100; seed++) {
            Problem pb = new Problem();
            IntDomainVar x = pb.makeEnumIntVar("x", 1, 5);
            IntDomainVar y = pb.makeEnumIntVar("y", 1, 5);
            pb.post(new SimpleInverse(x, 2, y, 3));

            pb.getSolver().setVarSelector(new RandomIntVarSelector(pb, seed));
            pb.getSolver().setValSelector(new RandomIntValSelector(seed));
            pb.solve();
            do {
                System.out.println(x.getVal() + " ? " + y.getVal());
            } while (pb.nextSolution() == Boolean.TRUE);
            System.out.println("nbSol " + pb.getSolver().getNbSolutions());
            assert(pb.getSolver().getNbSolutions() == 17);
        }
    }
}
