package cpmodel.heuristics;

import choco.integer.search.ValSelector;
import choco.integer.IntDomainVar;
import cpmodel.chocoext.IndexedIntVar;
import cpmodel.constraints.ConsecutiveConstraint;
import magicsearch.impact.ImpactStrategie;
import magicsearch.impact.impactMAC.IBSIntVarImpl;
import instance.TimetablingInstance;


public class OptimizationValueSelector implements ValSelector {

	protected ConsecutiveConstraint conscons;
	
	public IntDomainVar[] evts;
	public TimetablingInstance ins;
	protected ImpactStrategie ibs;

	public OptimizationValueSelector(TimetablingInstance ins, IntDomainVar[] evts, 
									 ConsecutiveConstraint cc, ImpactStrategie ibs) {
		this.conscons = cc;
		this.evts = evts;
		this.ins = ins;
		this.ibs = ibs;
	}

	
	protected int lastTimeslotCost( final int evt, final int tst )
	{
		int cost = 0;
		if( (tst % 9) == 0 )
		{
			cost += ins.getNbStudentInEvent(evt);
		}
		return cost;
	}
	
	
	protected int consecutiveCost( final int evt, final int tst )
	{
		return conscons.getPenalty(evt, tst);
	}
	

	// attention tu dois imperativement renvoyer une valeur qui est
	// dans le domaine de x !
	public int getBestVal(IntDomainVar x) {
		
		int evt = ((IBSIntVarImpl) x).getIndex();
		int bestval = x.getInf();
		double cost, bestcost = Integer.MAX_VALUE;
		
		int vali = bestval;
		
		if( !x.isInstantiated() ) {
		do {
			//System.out.print( "\nvalue:\n" + vali + ":" ); 
			
			cost = (lastTimeslotCost( evt, vali ) + 
					consecutiveCost( evt, vali ) +
					ibs.getImpactVal(x, vali));
			if( cost < bestcost )
			{
				bestcost = cost;
				bestval = vali;
			}
			
			//System.out.print( ibs.getImpactVal(x, vali) + ":" ); 
			//System.out.print( (cost + 1 - ibs.getImpactVal(x, vali)) + ",  " ); 
			
			vali = x.getNextDomainValue(vali);
		} while( vali <= x.getSup() );
		}
		
		//System.out.println( x + " = " + bestval );
		
		return bestval;
	}

}
