package cpmodel.heuristics;

import choco.integer.search.AbstractIntVarSelector;
import choco.integer.search.DoubleHeuristicIntVarSelector;
import choco.integer.IntDomainVar;
import choco.AbstractProblem;
import choco.AbstractConstraint;
import choco.ContradictionException;
import choco.util.IntIterator;
import choco.mem.IStateInt;
import choco.mem.PartiallyStoredVector;
import cpmodel.chocoext.IndexedIntVar;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: 7 sept. 2007
 * Time: 14:36:38
 * To change this template use File | Settings | File Templates.
 */
public class WeightedDomOverDeg extends AbstractIntVarSelector {

	protected int nEvents;
	protected int nTimeslots;
	protected int nRooms;
	
	
	
	public double timeContention[];
	protected IntDomainVar vars[];
	protected int useContention;

	//Give for each variable its set of neigbhours
	protected IntDomainVar[][] vargraph;

	public WeightedDomOverDeg() {
		super();
	}

	public WeightedDomOverDeg(IntDomainVar[] vs, int ne, int nt, int nr) {
		super();
		vars = vs;
		nEvents = ne;
		nTimeslots = nt;
		nRooms = nr;
		timeContention = new double[nTimeslots];
		vars = vs;
	}

//******************************************************//
//*********** incompatibility graph ********************//
//******************************************************//

	public void initvgraph() {
		vargraph = new IntDomainVar[vars.length][]; 
	}

	public void setNbNeigbhours(int i, int nbN) {
		vargraph[i] = new IntDomainVar[nbN];
	}

	public void addNeigbhours(int i, int j) {
		int cpt = 0;
		while(vargraph[i][cpt] != null) {
			cpt++;
		}
		vargraph[i][cpt] = vars[j];
		cpt = 0;
		while(vargraph[j][cpt] != null) {
			cpt++;
		}
		vargraph[j][cpt] = vars[i];
	}

	//return the static degree in the graph of differences
	//IT ASSUMES that v is a eventTime variable
	public int getGraphDegree(IntDomainVar v) {
		if (vargraph == null) return 0;
		int idx = ((IndexedIntVar) v).getIdx();
		return vargraph[idx].length;
	}

	//return the dynamic degree in the graph of differences
	//IT ASSUMES that v is a eventTime variable
	public int getGraphDynDegree(IntDomainVar v) {
		if (vargraph == null) return 0;
		int idx = ((IndexedIntVar) v).getIdx();
		int deg = 0;
		for (int i = 0; i < vargraph[idx].length; i++) {
		   if (vargraph[idx][i].isInstantiated()) deg++;
		}
	    return deg;
	}

//******************************************************//

	public int getDynDeg(IntDomainVar v) {
		int ddeg = 0;
		PartiallyStoredVector cts = v.getConstraintVector();
		IntIterator it = cts.getIndexIterator();
		while (it.hasNext()) {
			AbstractConstraint ct = (AbstractConstraint) v.getConstraint(it.next());
			if (!ct.isCompletelyInstantiated()) {
				ddeg += 1;
			}
		}
		return ddeg + getGraphDynDegree(v);

	}

	/// Computes the current contention values
	public void updateContention()
	{
		int dom;
		int vali;
		
		for (int i = 0; i < nTimeslots; i++) {
			timeContention[i] = 0;
		}
		
		for (int i = 0; i < nEvents; i++) {
			dom = vars[i].getDomainSize();
			vali = vars[i].getInf();
			do {
				timeContention[vali-1] += (1/(double)dom);
				vali = vars[i].getNextDomainValue(vali);
			} while( vali <= vars[i].getSup() );
		}
	
		
//		for (int i = 0; i < nEvents; i++) {
//			dom = vars[i].getDomainSize();
//			vali = vars[i].getInf();
//			do {
//				eventContention[i] += 1/(timeContention[vali-1]);
//				vali = vars[i].getNextDomainValue(vali);
//			} while( vali <= vars[i].getSup() );
//		}
		
//		for (int i = 0; i < nEvents; i++) {
//			
//			System.out.println( vars[i] + ": " + (vars[i].getDomain().pretty()) 
//					+ "  " + 
//					((double)(vars[i].getDomainSize()) / 
//							(double)(vars[i].getNbConstraints())) 
//					+ "  -> " + eventContention[i]);
//		}
//		
//		for (int i = 0; i < nTimeslots; i++) {
//			System.out.println( (i+1) + ": " + timeContention[i]);
//		}
//		System.out.println( );
		
	
//		try {
//			System.in.wait(1000);
//		} catch (InterruptedException e) {
//			// TODO Auto-generated catch block
//			e.printStackTrace();
//		}
		
	}
	
	
	/// Domain over degree
	public IntDomainVar dod() {
		
		int dom;
		int deg;
		int minDomain = nTimeslots+1;
		int maxDegree = -1;
		IntDomainVar v0 = null;
		for (IntDomainVar v:vars) {
			if (!v.isInstantiated()) {
				dom = v.getDomainSize();
				deg = v.getNbConstraints() + getGraphDegree(v) - 1;
				if (dom * maxDegree < minDomain * deg)  {
					minDomain = dom;
					maxDegree = deg;
					v0 = v;
				}
			}
		}
		return v0;
	}
	
//	public IntDomainVar com() {
//	
//		int i;
//		int x = -1;
//		int y = -1;
//		int dom;
//		int deg;
//		int minDomain = nTimeslots+1;
//		int maxDegree = 0;
//		IntDomainVar v0 = null;
//		for (i = 0; i < nEvents; i++) {
//			if (!vars[i].isInstantiated()) {
//				dom = vars[i].getDomainSize();
//				deg = vars[i].getNbConstraints() + getGraphDegree(vars[i]) - 1;
//				if (dom * maxDegree < minDomain * deg)  {
//					minDomain = dom;
//					maxDegree = deg;
//					v0 = vars[i];
//					x = i;
//				}
//			}
//		}
//	
//		maxDegree = 0;
//		updateContention();
//		double minContention = nTimeslots+1;
//		IntDomainVar v1 = null;
//		for (i = 0; i < nEvents; i++) {
//			if (!vars[i].isInstantiated()) {
//				
//
//				dom = vars[i].getDomainSize();
//				int vali = vars[i].getInf();
//				eventContention[i] = 0;
//				do {
//					eventContention[i] += 1/(timeContention[vali-1]);
//					vali = vars[i].getNextDomainValue(vali);
//				} while( vali <= vars[i].getSup() );
//		
//				
//				deg = vars[i].getNbConstraints() + getGraphDegree(vars[i]) - 1;
//				//deg = this.getDynDeg(vars[i]);
//				if ( eventContention[i] * maxDegree < minContention * deg)  {
//					minContention = eventContention[i];
//					maxDegree = deg;
//					v1 = vars[i];
//					y = i;
//				}
//			}
//		}
//		
//		
//		if( x >= 0  && v0 != v1 ) {
//			
//			System.out.println( v0 + ": " + (v0.getDomain().pretty()) 
//					+ "  " + 
//					((double)(v0.getDomainSize()) / 
//							(double)(v0.getNbConstraints())) 
//					+ "  -> " + (eventContention[x] /
//							(double)(v0.getNbConstraints())));
//			System.out.println( v1 + ": " + (v1.getDomain().pretty()) 
//					+ "  " + 
//					((double)(v1.getDomainSize()) / 
//							(double)(v1.getNbConstraints())) 
//					+ "  -> " + (eventContention[y] /
//							(double)(v1.getNbConstraints())));
//			System.out.println( );
//			
//		}
//		
//		
//		return v0;
//	}
	
	public double getEventContention( IntDomainVar x )
	{
		double ec = 0;
		int vali = x.getInf();
		do {
			ec += 1/(timeContention[vali-1]);
			vali = x.getNextDomainValue(vali);
		} while( vali <= x.getSup() );
		return ec;
	}
	
	public IntDomainVar con() {
		double minContention = nTimeslots+1;
		double ec;
		IntDomainVar v0 = null;
		for (int i = 0; i < nEvents; i++) {
			if (!vars[i].isInstantiated()) {
				ec = getEventContention( vars[i] );
				if ( ec < minContention )  {
					minContention = ec;
					v0 = vars[i];
				}
			}
		}
		return v0;
	}
	
	public IntDomainVar cod() {
		int deg;
		int maxDegree = 0;
		double minContention = nTimeslots+1;
		double ec;
		IntDomainVar v0 = null;
		for (int i = 0; i < nEvents; i++) {
			if (!vars[i].isInstantiated()) {
				ec = getEventContention( vars[i] );
				deg = vars[i].getNbConstraints() + getGraphDegree(vars[i]) - 1;
				if ( ec * maxDegree < minContention * deg)  {
					minContention = ec;
					maxDegree = deg;
					v0 = vars[i];
				}
			}
		}
		return v0;
	}
	
	public void setDod()
	{
		useContention = 0;
	}
	
	public void setCod()
	{
		useContention = 1;
	}
	
	public void setCon()
	{
		useContention = 2;
	}
	
	public IntDomainVar selectIntVar() throws ContradictionException {
		IntDomainVar v0 = null;
		if( useContention == 2 ) {
			updateContention();
			v0 = con();
		} else if( useContention == 1 ) {
			updateContention();
			v0 = cod();
		} else 
			v0 = dod();
		return v0;
	}

}
