package instance.clique;

import instance.ConflictGraph;

import java.util.BitSet;

/**
 * The class representing compound cliques, ie the disjoint
 * conjunction of another compound clique and an elementary
 * clique (implemented as a linked list).
 */
public class CompoundClique extends Clique{
	
	protected MaxCliqueFinder mc;
	protected ElementaryClique elementaryClique;
	protected CompoundClique baseClique;
	
	/**
	 * New compound clique.
	 * @param mc
	 * @param elementaryClique
	 * @param baseClique the current compound clique. <code>null</code> if none.
	 */
	public CompoundClique(MaxCliqueFinder mc,
			ElementaryClique elementaryClique,
			CompoundClique baseClique){
		this.mc = mc;
		this.elementaryClique = elementaryClique;
		this.baseClique = baseClique;
	}
	
	public boolean derivesFromRoomClique(){
		CompoundClique clique = this;
		while(clique.baseClique != null) clique = clique.baseClique;
		return clique.elementaryClique.getClass() == RoomClique.class;
		
	}
	
	public boolean derivesFromStudentClique(){
		CompoundClique clique = this;
		while(clique.baseClique != null) clique = clique.baseClique;
		return clique.elementaryClique.getClass() == StudentClique.class;
	}

	@Override
	public ConflictGraph getGraph() {
		return mc.getGraph();
	}

	@Override
	public int getSize() {
		return elementaryClique.getSize()+(baseClique == null ? 0 : baseClique.getSize());
	}

	@Override
	public BitSet retrieveNodes() {
		BitSet nodes = elementaryClique.retrieveNodes();
		if(baseClique != null) nodes.or(baseClique.retrieveNodes());
		return nodes;
	}
	
	@Override
	public String toString() {
		String type = derivesFromRoomClique()?"Room":"Student";
		return type+" "+super.toString();
	}
			
}