/**
 * 
 */
package instance.colouring;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.util.Arrays;

import tools.InstanceReader;
import util.IntSet;
import instance.ColourableConflictGraph;
import instance.TimetablingInstance;

/**
 * @author alexp
 *
 */
public class DegeneracyOrder extends GraphColouring {

	protected int[] sortedNodes;
	protected int currentIndex;
	
	/**
	 * @param graph
	 */
	public DegeneracyOrder(ColourableConflictGraph graph) {
		super(graph);
		
		sortedNodes = new int[graph.getNNodes()];
		sortNodes();
	}
	
	protected void sortNodes(){
		IntSet tempNodes = new IntSet(unassignedNodes);
		for(int i=0;i<sortedNodes.length;i++){
			int node = minDegreeNode(tempNodes);
			sortedNodes[sortedNodes.length-1-i] = node;
			tempNodes.remove(node);
		}
				
	}
	
	protected int minDegreeNode(IntSet nodes){
		int min = Integer.MAX_VALUE;
		IntSet minNodes = null;
		for(int n=nodes.first();n>-1;n=nodes.next(n)){
			int deg = degree(n, nodes);
			if(deg < min){
				min = deg;
				minNodes = new IntSet(sortedNodes.length);
				minNodes.setRandom(random);
			}
			if(deg == min){
				minNodes.add(n);
			}
		}
		assert(minNodes != null);
		return minNodes.randomValue();
	}
	
	protected int degree(int node, IntSet allowedNodes){
		int deg = 0;
		for(int m=graph.firstNeighbour(node);m>-1;m=graph.nextNeighbour(node, m))
			if(allowedNodes.contains(m)) ++deg;
		return deg;
	}

	/* (non-Javadoc)
	 * @see instance.colouring.GraphColouring#nextNode()
	 */
	@Override
	protected int nextNode() {
		return sortedNodes[currentIndex++];
	}

}
