package localsearch;

import choco.integer.IntDomainVar;

import localsearch.solvers.complete.*;
import localsearch.solvers.LSSolver;
import localsearch.solvers.coloring.TwoStepHardSolver;
import localsearch.solvers.coloring.TwoStepColoringSolver;

import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: 24 ao�t 2007
 * Time: 09:14:19
 * To change this template use File | Settings | File Templates.
 */
public class Benchmark {

    /**
     * compare the performances of different kind of neighbourhood
     *
     * @param setup
     * @param s
     */
    public static void setupNeighbourhood(int setup, LSSolver s) {
        switch (setup) {
            case 0:
                s.addNeighbourhood(AbstractSolver.TRANSLATE_EVENTS);
                s.addNeighbourhood(AbstractSolver.SWAP_EVENTS);
                break;
            case 1:
                s.addNeighbourhood(AbstractSolver.TRANSLATE_EVENTS);
                s.addNeighbourhood(AbstractSolver.SWAP_EVENTS);
                s.addNeighbourhood(AbstractSolver.SWAP_TIMESLOTS);
                break;
            case 2:
                s.addNeighbourhood(AbstractSolver.TRANSLATE_EVENTS);
                s.addNeighbourhood(AbstractSolver.SWAP_EVENTS);
                s.addNeighbourhood(AbstractSolver.HORIZ_HUNGARIAN_MOVE);
                break;
            case 3:
                s.addNeighbourhood(AbstractSolver.SWAP_TIMESLOTS);
                s.addNeighbourhood(AbstractSolver.TRANSLATE_EVENTS);
                s.addNeighbourhood(AbstractSolver.SWAP_EVENTS);
                s.addNeighbourhood(AbstractSolver.HORIZ_HUNGARIAN_MOVE);
                break;
            case 4:
                s.addNeighbourhood(AbstractSolver.SWAP_TIMESLOTS);
                s.addNeighbourhood(AbstractSolver.TRANSLATE_EVENTS_MATCHING);
                s.addNeighbourhood(AbstractSolver.HORIZ_HUNGARIAN_MOVE);
                s.addNeighbourhood(AbstractSolver.SWAP_EVENTS);
                s.addNeighbourhood(AbstractSolver.TRANSLATE_EVENTS);
                break;
            case 5:
                s.dynamicNeighbourhood();
                s.setGreedy();
                break;
            case 6:
                s.dynamicNeighbourhood();
                break;
            case 7:
                s.dynamicNeighbourhood();
                s.setGreedy();
                ((HardSolver) s).setLNS();
                break;
            case 8:
                s.addNeighbourhood(AbstractSolver.SWAP_TIMESLOTS);
                s.addNeighbourhood(AbstractSolver.TRANSLATE_EVENTS_MATCHING);
                s.addNeighbourhood(AbstractSolver.HORIZ_HUNGARIAN_MOVE);
                s.addNeighbourhood(AbstractSolver.SWAP_EVENTS);
                s.addNeighbourhood(AbstractSolver.TRANSLATE_EVENTS);
                s.setGreedy();
                break;
            default:
                throw new Error("Setup " + setup + " unknown");
        }
    }

    /**
     * compare the performances of different kind of neighbourhood
     *
     * @param setup
     * @param s
     */
    public static void setupSoftNeighbourhood(int setup, SoftSolver s) {
        switch (setup) {
            case 0:
                s.addNeighbourhood(AbstractSolver.TRANSLATE_EVENTS);
                break;
            case 1:
                s.addNeighbourhood(AbstractSolver.TRANSLATE_EVENTS);
                s.addNeighbourhood(AbstractSolver.SWAP_EVENTS);
                break;
            case 2:
                s.addNeighbourhood(AbstractSolver.TRANSLATE_EVENTS);
                s.addNeighbourhood(AbstractSolver.SWAP_TIMESLOTS);
                break;
            case 3:
                s.addNeighbourhood(AbstractSolver.CP_MOVE);
                break;
            default:
                throw new Error("Setup " + setup + " unknown");
        }
    }


    //A class to gather statistics over runs
    public static class Stat {
        String nameI;
        ArrayList<Integer> points;
        int numI = 0;
        int avgtime = 0;
        int avgiter = 0;
        int avgquality = 0;
        int nbfeasfound = 0;
        int minquality = Integer.MAX_VALUE;
        int worstquality = Integer.MIN_VALUE;
        int mintime = Integer.MAX_VALUE;
        int maxtime = Integer.MIN_VALUE;

        int nbLNSAttempt = 0;
        double avg_lns_node;
        double avg_lns_time;
        int nbTreached;

        public Stat(int i, String nameIns) {
            numI = i;
            nameI = nameIns;
            points = new ArrayList<Integer>();
        }

        public double computeDeviation() {
            double avgSquares= 0;
            double avg = 0;
            for (Iterator<Integer> iterator = points.iterator(); iterator.hasNext();) {
                int val = iterator.next();
                avgSquares += Math.pow(val,2);
                avg += val;
            }
            avgSquares = avgSquares / (double) points.size();
            avg = avg / (double) points.size();
            return Math.sqrt((avgSquares - Math.pow(avg,2)));
        }

        public static double computeMedian(List<Integer> intlist) {
            if (!intlist.isEmpty()) {
                Collections.sort(intlist);
                if (intlist.size() % 2 == 1) {
                    int middle = (intlist.size() - 1) / 2;
                    return intlist.get(middle);
                } else {
                    int center = intlist.size() / 2;
                    return (double) (intlist.get(center - 1) + intlist.get(center)) / 2d;
                }
            } else return -1;
        }

        public void add(boolean feas, LSSolver s2) {
            int tottime = s2.getTotalTime();
            if (feas) {
                nbfeasfound++;
            }
            avgtime += tottime;
            avgiter += s2.getNbIter();
            if (tottime < mintime) {
                mintime = tottime;
            }
            if (tottime > maxtime) {
                maxtime = tottime;
            }
            if (s2 instanceof SoftSolver) {
                nbLNSAttempt += ((SoftSolver) s2).nbCPMove;
                avg_lns_node += ((SoftSolver) s2).sumnodes / (double) ((SoftSolver) s2).nbCPMove;
                avg_lns_time += ((SoftSolver) s2).time / (double) ((SoftSolver) s2).nbCPMove;
                nbTreached += ((SoftSolver) s2).nbtthresholdReached;
            }

            if (feas) {
                avgquality += s2.getBestCost();
                points.add(s2.getBestCost());
                if (minquality > s2.getBestCost())
                    minquality = s2.getBestCost();
                if (worstquality < s2.getBestCost())
                    worstquality = s2.getBestCost();
            }
        }

        public void print(int ech) {
            System.out.print(nameI + " & ");
            System.out.print(roundTok((double) nbfeasfound / (double) ech, 2) + " & ");
            System.out.print(roundInSec((double) avgtime / (double) ech) + " & ");
            System.out.print(round((double) avgiter / (double) ech) + " & ");
            double aq = Math.round((double) avgquality / (double) nbfeasfound);
            System.out.print((int) aq + " & ");
            System.out.print((int) Math.round(computeMedian(points)) + " & ");
            System.out.print(minquality + " & ");
            System.out.print(worstquality + " & ");
            System.out.print(roundInSec((double) mintime) + " & ");
            System.out.print(roundInSec((double) maxtime) + " & ");


            System.out.print(round((double) nbLNSAttempt / (double) ech) + " & ");
            System.out.print(round(avg_lns_node / (double) ech) + " & ");
            System.out.print(round(avg_lns_time / (double) ech) + " & ");
            System.out.print(round(nbTreached / (double) ech) + " & ");

            System.out.println(" \\\\");
        }

        public void fastprint(boolean feas, AbstractSolver s2) {
            if (feas)
                System.out.print(s2.getBestCost() + ",");
            else
                System.out.print(-1 + ",");
        }


    }


    /**
     * Benchmarck the complete resolution hard + soft
     *
     * @param from2002
     * @param numI
     * @param name
     * @param tabusize
     */
    public static void testSearchLoop(int numI, String name, int tabusize) {
        Stat statistic = new Stat(numI, name);
        for (int i = seed; i < (ech + seed); i++) {
            SearchLoop ss = new SearchLoop(name, i, SA_Then_LNS_ratio);
            ss.setParamSolver(timelimit, tabusize);
            boolean feas = ss.solveLoop();
            if (feas) {
                statistic.add(true, ss.softS);
                ss.softS.outputSolution(true);
            } else {
                statistic.add(false, ss.hardS);
            }
            //statistic.fastprint(feas, ss.softS);
        }
        statistic.print(ech);
    }

    /**
     * Benchmark only the search for feasible solution
     *
     * @param from2002
     * @param d
     * @param f
     * @param numI
     * @param name
     * @param tabusize
     */
    public static void testHardSolver(int numI, String name, int tabusize) {
            Stat statistic = new Stat(numI, name);
            for (int i = seed; i < (ech + seed); i++) {
                LSSolver s;
                if (twoStepHardSolverOnly) {
                    s = new TwoStepHardSolver(name, i);
                    ((TwoStepHardSolver) s).setDefaultNeighbourhood();
                } else {
                    s = new HardSolver(name, i);
                    setupNeighbourhood(defFeasNeig, s);
                }
                s.setTabuMetaHeuristic(tabusize);
                s.setTimelimit(timelimit);
                //((HardSolver) s).fastDescentStep(true);
                s.solve(); //solve the feasible part
                if (s.getBestCost() == 0)
                    s.outputSolution(false);
                statistic.add(s.getBestCost() == 0, s);
            }
            statistic.print(ech);
    }

    /**
     * Resolution of hard + soft constraints by solving first the
     * colouring relaxation
     *
     * @param from2002
     * @param d
     * @param f
     * @param numI
     * @param name
     * @param tabusize
     */
    public static void testFourStepsApproach(int numI, String name, int tabusize) {
        Stat statistic = new Stat(numI, name);
        for (int i = seed; i < (ech + seed); i++) {
            TwoStepColoringSolver s = new TwoStepColoringSolver(name, i);
            s.setTimelimit(timelimit);
            s.setDefaultNeighnourhood();
            s.setTabuMetaHeuristic(tabusize);
            s.setSAMetaHeuristic(1);
            s.solve();
            if (s.isFeasible()) {
                s.outputSolution(false);
                statistic.add(s.isFeasible(), s.ss);
            } else {
                statistic.add(false, s.hs);
            }
        }
        statistic.print(ech);
    }

    public static void test(int numI, String name) {
        int tabusize = 10;
        if (fourStepsSolver) {
            testFourStepsApproach(numI, name, tabusize);
        } else if (hardsolverOnly) {
            testHardSolver(numI, name, tabusize);//4
        } else
            testSearchLoop(numI, name, tabusize);
    }

    public static int round(double d) {
        return (int) Math.round(d);
    }

    public static double roundTok(double d, int k) {
        return (double) Math.round(d * Math.pow(10,k)) / Math.pow(10,k);
    }

    public static double roundInSec(double d) {
        return (double) Math.round(d / 10d) / 100d;
    }

    //number of run to gather statictics
    public static int ech = 20;  // passer a 50
    public static int seed = 0;
    //an identifier for the files generated containing the solution
    public static int unique_ID = 0;
    public static int[] instancelist = new int[]{1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16};

    //benchmarck the soft solver by generating multiple feasible solutions
    public static boolean cyclicSoft = false;
    // benchmark only the hard solver
    public static boolean hardsolverOnly = false;
    //benchmarck the two steps hard solver
    public static boolean twoStepHardSolverOnly = false;
    //Four steps approach working on coloring first and then matching
    public static boolean fourStepsSolver = false;
    //only use a lns scheme
    public static boolean pureLNS = false;
    //use the CP solver as a move in the SA
    public static boolean SA_LNS = false;
    //use LNS as an intensification phase on SA_Then_LNS_ratio % of the timelimit
    public static double SA_Then_LNS_ratio = 0;

    //default neighborhood for feasibility
    public static int defFeasNeig = 8;


    //
    public static int timelimit = 372000;
    public static boolean checkResults = true;


    public static void testsHardFeas() {
        //test(false, 8, "data/Track2/comp-2007-2-8.tim");
        test(1, "data/Track2/comp-2007-2-1.tim");
        test(2, "data/Track2/comp-2007-2-2.tim");
        test(9, "data/Track2/comp-2007-2-9.tim");
        test(10, "data/Track2/comp-2007-2-10.tim");
    }


    public static void tests() {
        for (int i = 0; i < 16; i++) {
            test(i + 1, "data/Track2/comp-2007-2-" + (i + 1) + ".tim");
        }
    }

    public static void tests(int[] instances) {
        for (int i = 0; i < instances.length; i++) {
            test(instances[i], "data/Track2/comp-2007-2-" + instances[i] + ".tim");
        }
    }

    public static void tests2002() {
        for (int i = 0; i < 20; i++) {
            if (i < 9)
                test(i + 1, "data/2002/competition0" + (i + 1) + ".tim");
            else test(i + 1, "data/2002/competition" + (i + 1) + ".tim");
        }
    }

    public static void testsSmall2002() {
        for (int i = 0; i < 5; i++) {
            test(i + 1, "data/small2008/small" + (i + 1) + ".tim");
        }
    }

    public static void testsMedium2002() {
        for (int i = 0; i < 5; i++) {
            test(i + 1, "data/medium2008/medium0" + (i + 1) + ".tim");
        }
    }

    //******************************************************************* //
    //*********************** MAIN ************************************** //
    //******************************************************************* //

    //results submitted with seed 1

    public static int[] getInstances(String s) {
        StringTokenizer st = new StringTokenizer(s, " ");
        int[] instances = new int[st.countTokens()];
        int cpt = 0;
        while (st.hasMoreTokens()) {
            instances[cpt] = Integer.parseInt(st.nextToken());
            cpt++;
        }
        return instances;
    }

    // timelimit, nbruns, uniqueID, approach among 0..6, instances,
    // an initial seed, threshold to apply intensification
    public static void main(String[] args) {
          int timlim = Integer.parseInt(args[0]);//372000;
          int nbruns = Integer.parseInt(args[1]);
          int uniqueID = Integer.parseInt(args[2]);
          int mode = Integer.parseInt(args[3]);
          if (args.length > 4)
              instancelist =  getInstances(args[4]);
          if (args.length > 5)
              seed = Integer.parseInt(args[5]);
          if (args.length > 6)
              HardSolver.threshold = Integer.parseInt(args[6]);
          switch(mode) {
              case 0 : main_feasLS1(uniqueID,timlim,nbruns); break;
              case 1 : main_feasLS2(uniqueID,timlim,nbruns); break;
              case 2 : main_feasLS3(uniqueID,timlim,nbruns); break;
              case 3 : main_SA(uniqueID,timlim,nbruns); break;
              case 4 : main_SA2(uniqueID,timlim,nbruns); break;
              case 5 : main_SA_LNS(uniqueID,timlim,nbruns); break;
              case 6 : main_LNS(uniqueID,timlim,nbruns); break;
          }
    }


    public static void testChocoSol(
            boolean completesolve, String
            nameInst, IntDomainVar[] events) {
        HardSolver s = new HardSolver(nameInst, 0);
        setupNeighbourhood(3, s);
        int[] intevents = new int[events.length];
        for (int i = 0; i < intevents.length; i++) {
            intevents[i] = events[i].getVal() - 1;
        }
        s.setGreedyInit(intevents);
        if (completesolve)
            s.solve();
        else {
            s.start();
            System.out.println("best cost: " + s.bestCost);
        }
        s.outputSolution(true);
    }


    //******************************************************************* //
    //****** MAIN to reproduce the experiments of the paper ************ //
    //******************************************************************* //

    //ech = 100 for LS1, LS2, LS3
    //timelimit = 324 on the cluster

    public static void commonSetup(int uniqueID, int tl) {
        //TRACE
        AbstractSolver.trace = false;
        AbstractSolver.debug = false;
        //PARAMETERS
        timelimit = tl * 1000;
        //seed = 0;
        unique_ID = uniqueID;
    }

    public static void feasSetup(int nbrun) {
        ech = nbrun;
        hardsolverOnly = true;        
    }

    public static void softSetup(int nbrun) {
        ech = nbrun;
    }

    //result for LS1 in section 3.1
    public static void main_feasLS1(int uniqueID, int timelimit, int ech) {
        System.out.println("Results for LS1 on feasibility only");
        commonSetup(uniqueID, timelimit);
        feasSetup(ech);
        defFeasNeig = 4;
        tests(instancelist);
    }

    //result for LS2 in section 3.1
    public static void main_feasLS2(int uniqueID, int timelimit, int ech) {
        System.out.println("Results for LS2 on feasibility only");
        commonSetup(uniqueID, timelimit);
        feasSetup(ech);
        tests(instancelist);
    }

    //result for LS3 in section 3.1    
    public static void main_feasLS3(int uniqueID, int timelimit, int ech) {
        System.out.println("Results for LS3 on feasibility only");        
        commonSetup(uniqueID, timelimit);
        feasSetup(ech);
        twoStepHardSolverOnly = true;
        tests(instancelist);
    }

    public static void main_SA(int uniqueID, int timelimit, int ech) {
        System.out.println("Standart SA");
        commonSetup(uniqueID, timelimit);
        softSetup(ech);
        tests(instancelist);
    }

    public static void main_SA2(int uniqueID, int timelimit, int ech) {
        System.out.println("SA on colouring first");
        commonSetup(uniqueID, timelimit);
        softSetup(ech);
        fourStepsSolver = true;
        tests(instancelist);
    }

    public static void main_LNS(int uniqueID, int timelimit, int ech) {
        System.out.println("LNS");        
        commonSetup(uniqueID, timelimit);
        softSetup(ech);
        pureLNS = true;
        SoftSolver.nodelimit = 800;
        SoftCostProjection.s_nbct = 1;
        SoftCostProjection.s_nbrt = 1;
        SoftSolver.growingSizeN = true;
        SoftSolver.grow_threshold = 50;
        SoftSolver.diversification = true;
        tests(instancelist);
    }

    public static void main_SA_LNS(int uniqueID, int timelimit, int ech) {
        System.out.println("SA with LNS move");
        commonSetup(uniqueID, timelimit);
        softSetup(ech);
        SA_LNS = true;
        SoftSolver.nodelimit = 5000;
        SoftCostProjection.s_nbct = 2;
        SoftCostProjection.s_nbrt = 2;
        SoftSolver.growingSizeN = true;
        SoftSolver.grow_threshold = 100;
        SoftSolver.discrepency = 7;
        tests(instancelist);
    }
}
