package localsearch;

import instance.TimetablingInstance;

import java.util.Arrays;
import java.util.BitSet;
import java.util.ArrayList;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: 17 sept. 2007
 * Time: 13:24:33
 * To change this template use File | Settings | File Templates.
 */
public class SolutionSpace extends HardSolutionSpace {

	//******** Instance data ************//

	//Dim: NbEvents * NbTimeslots
	//give  for each student the list of feasible timeslots
	public int[][] feastimeslots;

	// Dim: NbEvents * NbStudents
	// give for each event the list of student following the course
	public int[][] studentlist;

	// Dim: NbEvents * NbEvents * at most NbStudents
	// deltaStudentlist[i][j] give the list of students who follow
	// event i but not j
	public int[][][] deltaStudentlist;

	// Dim: NbEvents * NbEvents
	// possibleswap[i] gives the list of events that have no precedences with i
	public int[][] possibleswap;

	//Dim: NbEvents
	//give for each event the number of student following it
	public int[] evtSize;

	//Dim: NbTimeslots
	//specify for each timeslot if it is the end of the day or not
	public boolean[] finaltimeslots;

	//Dim: NbTimeslots
	//specify for each timeslot the corresponding day
	public int[] day;

	//Dim: NbTimeslots
	//daytimeslots[t] gives for a timeslot t, its index (between 1 and 8) in the day.
	public int[] timeDayIdx;

	//******** Cost computation ************//

	//global soft cost of the objective
	//softObj = softObj1 + softObj2 + softObj3
	public int softObj, softObj1, softObj2, softObj3;

	//softobj[i][j] is the projection of the soft cost i on day j
	public int[][] sobj_day;

	//******** Description of a solution ************//

	// Dim: NbStudent * NbDay(5) * NbHours + 1 (10)
	// gives the timetabling of each student
	// sudentPlanning[i][j][0] :
	// 		the number of courses followed by student i the day j
	// sudentPlanning[i][j][ech] with ech != 0 :
	// 		the course followed by student i the day j, the ech-ieme timeslot of the day
	public int[][][] studentPlanning;


    //default constructor
    public SolutionSpace() {

    }

    public SolutionSpace(TimetablingInstance ins, int seed) {
		super(ins, seed);
	}

	public void baseSoftInit() {
		studentlist = new int[E][];
		evtSize = new int[E];
		deltaStudentlist = new int[E][E][];
		possibleswap = new int[E][];
		finaltimeslots = new boolean[T];
		timeDayIdx = new int[T];
		feastimeslots = new int[E][];
		day = new int[T];
		sobj_day = new int[3][D];

		for (int i = 0; i < T; i++) {
			day[i] = i / H;
			timeDayIdx[i] = i % H + 1;
		}
		for (int i = 0; i < D; i++) {
			finaltimeslots[(i + 1) * H - 1] = true;
		}
		for (int i = 0; i < E; i++) {
			possibleswap[i] = getPossibleSwapTable(i);
			feastimeslots[i] = getTable(ins.timeslotsAssignableToEvent(i));
			evtSize[i] = ins.getNbStudentInEvent(i);
			studentlist[i] = getTable(ins.studentsAttendingEvent(i));
			for (int j = 0; j < feastimeslots.length; j++) {
				if (i != j)
					deltaStudentlist[i][j] = getDeltaEventTable(i, j);
			}
		}
	}

	/**
	 * Initialize all "static" data structures needed for
	 * evaluating quickly the soft cost
	 */
	public void softInit() {
		studentPlanning = new int[S][D][H + 1];
		softRestart();
		for (int i = 0; i < D; i++) {
			for (int j = 0; j < H; j++) {
				int timeslot = i * H + j;
				for (int k = 0; k < nbept[timeslot]; k++) {
					int evt = listevts[timeslot][k];
					for (int l = 0; l < evtSize[evt]; l++) {
						int s = studentlist[evt][l];
						studentPlanning[s][i][j + 1] = evt;
						studentPlanning[s][i][0]++;
					}
				}
			}
		}
	}

	/**
	 * @param e1 first event
	 * @return the table of events that have no precedences with e1
	 */
	public int[] getPossibleSwapTable(int e1) {
		int[] events = new int[E - ins.getNInvprec(e1) - ins.getNPrec(e1)];
		int cpt = 0;
		for (int i = 0; i < E; i++) {
			if (i != e1 && !before[e1][i] && !before[i][e1]) {
				events[cpt] = i;
				cpt++;
			}
		}
		return events;
	}

	/**
	 * @param e1 first event
	 * @param e2 second event
	 * @return the table of students who follow e1 but not e2
	 */
	public int[] getDeltaEventTable(int e1, int e2) {
		ArrayList<Integer> templ = new ArrayList<Integer>();
		BitSet b1 = ins.studentsAttendingEvent(e1);
		BitSet b2 = ins.studentsAttendingEvent(e2);
		for (int i = b1.nextSetBit(0); i >= 0; i = b1.nextSetBit(i + 1)) {
			if (!b2.get(i)) templ.add(i);
		}
		int[] studenttable = new int[templ.size()];
		for (int i = 0; i < studenttable.length; i++) {
			studenttable[i] = templ.get(i);
		}
		return studenttable;
	}

	/**
	 * @param b the bitset
	 * @return a table of the bitset b
	 */
	public int[] getTable(BitSet b) {
		int[] studenttable = new int[b.cardinality()];
		int cpt = 0;
		for (int i = b.nextSetBit(0); i >= 0; i = b.nextSetBit(i + 1)) {
			studenttable[cpt] = i;
			cpt++;
		}
		return studenttable;
	}

	/**
	 * Compute the real number of student by agregating identical
	 * student (todo)
	 */
	public void reduceStudents() {
		BitSet student = new BitSet(S); //real students
		int nbStudent = 0;
		for (int i = 0; i < S; i++) {
			if (!student.get(i)) {
				student.set(i);
				nbStudent++;
				for (int j = 0; j < S; j++) {
					if (ins.eventsAttendedByStudent(i).equals(ins.eventsAttendedByStudent(j))) {
						student.set(j);
					}
				}
			}
		}
		System.out.println(nbStudent + " vs " + ins.getNStudents());
	}

	public boolean sameDay(int t1, int t2, int t3) {
		return (day[t1] == day[t2])  && (day[t2] == day[t3]);
	}



	public void softRestart() {
		for (int i = 0; i < S; i++) {
			for (int j = 0; j < D; j++) {
				Arrays.fill(studentPlanning[i][j], -1);
				studentPlanning[i][j][0] = 0;
			}
		}
		softObj = 0;
		softObj1 = 0;
		softObj2 = 0;
		softObj3 = 0;
	}

	//******** Basic Api *******************//

	public boolean isFinalTimeslot(int timeslot) {
		return finaltimeslots[timeslot];
	}

	public int getDay(int timeslot) {
		return day[timeslot];
	}

	public int getSingleEvt(int d, int s, int guessEvt) {
		if (guessEvt != -1 && getDay(timeslot[guessEvt]) == d) return guessEvt;
		for (int i = 1; i <= H; i++) {
			if (studentPlanning[s][d][i] != -1) return studentPlanning[s][d][i];
		}
		throw new Error("there should be a single event in this day !");
	}

	public int getCostProjection(int cost_idx) {
		int c1 = 0;
		for (int i = 0; i < D; i++) {
			c1 += sobj_day[cost_idx][i];
		}
		return c1;
	}


	public int getMostConflictingDay() {
		int cmaxday = Integer.MIN_VALUE;
		int maxday = -1;
		for (int i = 0; i < D; i++) {
			int ci = sobj_day[0][i] + sobj_day[1][i] + sobj_day[2][i];
			if (ci > cmaxday) {
				maxday = i;
				cmaxday = ci;
			}
		}
		return maxday;
	}

	/**
	 * get the timeslot with the least number of student
	 *
	 * @param except
	 * @return
	 */
	public int getEmptiestTimeslot(BitSet except) {
		int smallestT = -1;
		int smallestS = Integer.MAX_VALUE;
		for (int i = 0; i < T; i++) {
			if (!except.get(i) && nbept[i] > 0) {
				if (smallestS > nbept[i]) {
					smallestT = i;
					smallestS = nbept[i];
				} else if (smallestS == nbept[i] && rand.nextBoolean()) {
					smallestT = i;
				}
			}
		}
		return smallestT;
	}

	/**
	 * get the timeslot with the least number of student
	 *
	 * @param except
	 * @return
	 */
	public int getRandomTimeslot(BitSet except) {
		int cpt = except.cardinality();
		int timeslot_idx = rand.nextInt(T - cpt);
		int t, tempIdx = 0;
		for (t = 0; t < T && (tempIdx != timeslot_idx); t++) {
			if (!except.get(t)) {
				tempIdx++;
			}
		}
		return t;
	}

	//**********************************************************//
	//******** Operation on Events for Hard data ***************//
	//**********************************************************//

	/**
	 * Remove event evt from its current location while
	 * maintaining only the hard data structures needed by softmoves
	 *
	 * @param evt the event removed
	 */
	public void softRemoveEvent(int evt) {
		int e1t = timeslot[evt];
		int e1r = rooms[evt];
		matrix[e1t][e1r] = -1;
		rooms[evt] = -1;
		timeslot[evt] = -1;
		int pevt = -1;
		for (int i = 0; i < R; i++) { //todo : improve with a real list data structure
			if (listevts[e1t][i] == evt) {
				pevt = i;
				break;
			}
		}
		maintainListEvt(evt, e1t, pevt);
		nbept[e1t]--;
	}

	/**
	 * Place event evt in the EMPTY place corresponding to the given
	 * timeslot and room while maintaining only the hard data structures needed
	 * by soft moves.
	 *
	 * @param evt     the event to add
	 * @param timeslo the timeslot where we want to put the event
	 * @param room    the room where the event has to go
	 */
	public void softAddEvent(int evt, int timeslo, int room) {
		assert (matrix[timeslo][room] == -1);
		matrix[timeslo][room] = evt;
		timeslot[evt] = timeslo;
		rooms[evt] = room;
		listevts[timeslo][nbept[timeslo]] = evt;
		nbept[timeslo]++;
	}

	/**
	 * change the room of an event who is staying in the same timeslot
	 *
	 * @param evt the event to remove
	 */
	public void fastSoftRemoveEvent(int evt) {
		int e1t = timeslot[evt];
		int e1r = rooms[evt];
		matrix[e1t][e1r] = -1;
		rooms[evt] = -1;
	}

	/**
	 * add the event in the same timeslot it has been removed
	 *
	 * @param evt
	 * @param room the new room in the same timeslot
	 */
	public void fastSoftAddEvent(int evt, int room) {
		assert (matrix[timeslot[evt]][room] == -1);
		matrix[timeslot[evt]][room] = evt;
		rooms[evt] = room;
	}

	//**********************************************************//
	//******** Api for swapping timeslots **********************//
	//**********************************************************//


	public void fastSwapTimeSlot(int t1, int t2) {
		int[] levtst1 = listevts[t1];
		int[] rooms1 = saveRooms(t1);
		int[] levtst2 = listevts[t2];
		int[] rooms2 = saveRooms(t2);
		for (int i = 0; i < rooms1.length; i++) {
			int e = levtst1[i];
			fastSoftRemoveEvent(e);
			timeslot[e] = t2;
		}
		for (int i = 0; i < rooms2.length; i++) {
			int e = levtst2[i];
			fastSoftRemoveEvent(e);
			timeslot[e] = t1;
		}
		for (int i = 0; i < rooms1.length; i++) {
			int e = levtst1[i];
			fastSoftAddEvent(e, rooms1[i]);
		}
		for (int i = 0; i < rooms2.length; i++) {
			int e = levtst2[i];
			fastSoftAddEvent(e, rooms2[i]);
		}
		nbept[t1] = rooms2.length;
		nbept[t2] = rooms1.length;
		listevts[t1] = levtst2;
		listevts[t2] = levtst1;
	}

	public int[] saveRooms(int t) {
		int[] room = new int[nbept[t]];
		for (int i = 0; i < room.length; i++) {
			room[i] = rooms[listevts[t][i]];
		}
		return room;
	}

	//**********************************************************//
	//******** Operation on Events for Soft data ***************//
	//**********************************************************//

	/**
	 * maintain only the soft data structures that changed
	 * because of the removal of event evt from its current location
	 *
	 * @param evt the event removed
	 */
	public void removeEventSoftCost(int evt, int timeslo) {
		int eday = getDay(timeslo);
		//maintain cost1
		if (isFinalTimeslot(timeslo)) {
			softObj -= evtSize[evt];
			sobj_day[0][eday] -= evtSize[evt];
		}
		//maintain cost2
		for (int i = 0; i < studentlist[evt].length; i++) {
			int s = studentlist[evt][i];
			int varObj2 = computeSoftCost2Variation(s, timeslo, evt);
			softObj -= varObj2;
			sobj_day[1][eday] -= varObj2;
			studentPlanning[s][eday][0]--;
			studentPlanning[s][eday][timeDayIdx[timeslo]] = -1; //mapping inverse
		}
		//maintain cost3
		int qvar = 0;
		for (int i = 0; i < studentlist[evt].length; i++) {
			int s = studentlist[evt][i];
			if (studentPlanning[s][eday][0] == 0) {//because the student planning has been updated before
				qvar--;
			}
			if (studentPlanning[s][eday][0] == 1) {//because the student planning has been updated before
				qvar++;
			}
		}
		sobj_day[2][eday] += qvar;
		softObj += qvar;
	}

	/**
	 * maintain only the soft data structures that changed
	 * because of the addition of an event in a given timeslot
	 *
	 * @param evt     the event to add
	 * @param timeslo the timeslot where we want to put the event
	 */
	public void addEventSoftCost(int evt, int timeslo) {
		int eday = getDay(timeslo);
		//maintain cost1
		if (isFinalTimeslot(timeslo)) {
			softObj += evtSize[evt];
			sobj_day[0][eday] += evtSize[evt];
		}
		//maintain cost3
		int qvar = 0;
		for (int i = 0; i < studentlist[evt].length; i++) {
			int s = studentlist[evt][i];
			if (studentPlanning[s][eday][0] == 1) {
				qvar--;
			} else if (studentPlanning[s][eday][0] == 0) {
				qvar++;
			}
		}
		sobj_day[2][eday] += qvar;
		softObj += qvar;

		//maintain cost2
		for (int i = 0; i < studentlist[evt].length; i++) {
			int s = studentlist[evt][i];
			int varObj2 = computeSoftCost2Variation(s, timeslo, evt);
			softObj += varObj2;
			sobj_day[1][eday] += varObj2;
			studentPlanning[s][eday][0]++;
			studentPlanning[s][eday][timeDayIdx[timeslo]] = evt;
		}
	}

	/**
	 * fast remove if the event is planned to be added in the same timeslo
	 * it typically avoid to maintain cost3
	 *
	 * @param evt the event removed
	 */
	public void fastRemoveEventSoftCost(int evt, int timeslo) {
		int eday = getDay(timeslo);
		//maintain cost1
		if (isFinalTimeslot(timeslo)) {
			softObj -= evtSize[evt];
			sobj_day[0][eday] -= evtSize[evt];
		}
		//maintain cost2
		for (int i = 0; i < studentlist[evt].length; i++) {
			int s = studentlist[evt][i];
			int varObj2 = computeSoftCost2Variation(s, timeslo, evt);
			softObj -= varObj2;
			sobj_day[1][eday] -= varObj2;
			studentPlanning[s][eday][timeDayIdx[timeslo]] = -1; //mapping inverse
		}

	}

	/**
	 * fast addition if the event is coming from a timeslot in the same day.
	 * This avoid to maintain cost3
	 *
	 * @param evt     the event to add
	 * @param timeslo the timeslot where we want to put the event
	 */
	public void fastAddEventSoftCost(int evt, int timeslo) {
		int eday = getDay(timeslo);
		//maintain cost1
		if (isFinalTimeslot(timeslo)) {
			softObj += evtSize[evt];
			sobj_day[0][eday] += evtSize[evt];
		}
		//maintain cost2
		for (int i = 0; i < studentlist[evt].length; i++) {
			int s = studentlist[evt][i];
			int varObj2 = computeSoftCost2Variation(s, timeslo, evt);
			softObj += varObj2;
			sobj_day[1][eday] += varObj2;
			studentPlanning[s][eday][timeDayIdx[timeslo]] = evt;
		}
	}

	//**********************************************************//
	//******** Hard cost feasability tests *********************//
	//**********************************************************//

	/**
	 * evaluate if a precedence will be violated by moving evt from timeslot t1 to t2
	 * It returns the new cost projection of evt regarding precedences AND
	 * the NEW cost of other violated precedences of other events with evt
	 */
	public boolean evalFeasPrecedenceFromMove(int evt, int evt2, int t1, int t2) {
		for (int i = 0; i < greaterthan[evt].length; i++) {
			if (timeslot[greaterthan[evt][i]] >= t2) {
				return false;
			}
		}
		for (int i = 0; i < lessthan[evt].length; i++) {
			int e2 = lessthan[evt][i];// we should have e2 > evt

			if (evt2 != e2) {
				//if the precedences was satisfied before but is violated now
				if (timeslot[e2] > t1 && timeslot[e2] <= t2) {
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * evaluate if a precedence linked to evt will be violated (without counting
	 * the precedences violated from events placed on timeslot t1 and t2)
	 */
	public boolean evalFeasPrecCost(int evt, int ts1, int ts2) {
		int t1 = timeslot[evt];
		for (int i = 0; i < greaterthan[evt].length; i++) { //we should evt > greaterthan[evt][cpt]
			if (timeslot[greaterthan[evt][i]] >= t1) {
				return false;
			}
		}
		for (int i = 0; i < lessthan[evt].length; i++) {
			int e2 = lessthan[evt][i];// we should have e2 > evt
			int t2 = timeslot[e2];
			if (t2 != ts1 && t2 != ts2 && timeslot[e2] <= t1)
				return false;
		}
		return true;
	}

	//**********************************************************//
	//******** Soft Cost computation ***************************//
	//**********************************************************//
	//******** computation from scratch ************************//

	public void computeSoftCostFromScratch() {
		computeC1FromScratch();
		computeC2FromScratch();
		computeC3FromScratch(true);
		softObj = softObj1 + softObj2 + softObj3;
	}

	/**
	 * Compute the cost involved by students having events in the last timeslot
	 * of the day
	 */
	public void computeC1FromScratch() {
		softObj1 = 0;
		for (int d = 0; d < 5; d++) {
			int t = d + 8 * (d + 1);
			for (int e = 0; e < nbept[t]; e++) {
				int evt = listevts[t][e];
				softObj1 += evtSize[evt];
				sobj_day[0][d] += evtSize[evt];
			}
		}
	}

	/**
	 * Compute the cost involved by successive events for a student
	 */
	public void computeC2FromScratch() {
		softObj2 = 0;
		for (int s = 0; s < S; s++) {
			for (int d = 0; d < D; d++) {
				computeC2OfStudentAndDay(s, d);
			}
		}
	}

	public void computeC2OfStudentAndDay(int s, int d) {
		int[] day = studentPlanning[s][d];
		int seqsize = 0;
		for (int i = 1; i < day.length; i++) {
			int e = day[i];
			if (e == -1) {
				seqsize = 0;
			} else {
				seqsize++;
				if (seqsize > 2) {
					softObj2++;
					sobj_day[1][d]++;
				}
			}
		}
	}

	/**
	 * Compute the cost involved by students having a single event
	 * per day
	 */
	public void computeC3FromScratch(boolean updatedata) {
		softObj3 = 0;
		for (int s = 0; s < S; s++) {
			for (int d = 0; d < D; d++) {
				if (studentPlanning[s][d][0] == 1) {
					softObj3++;
					sobj_day[2][d]++;
				}
			}
		}
	}

//******** soft cost incremental computation *********************//

	/**
	 * Compute the reduction/increase of cost2 related to the removal/addition of
	 * student s from the timeslot t (excluding evt which is about to be moved)
	 */
	public int computeSoftCost2Variation(int s, int t, int evt) {
		int d = getDay(t);
		int pos = timeDayIdx[t];
		int nba = 0; // nb consecutive timeslots after t where s is involved
		int nbb = 0; // nb consecutive timeslots before t where s is involved

		for (int i = pos + 1; i <= H; i++) {
			if (studentPlanning[s][d][i] != -1 &&
					studentPlanning[s][d][i] != evt)
				nba++;
			else break;
		}
		for (int i = pos - 1; i >= 1; i--) {
			if (studentPlanning[s][d][i] != -1 &&
					studentPlanning[s][d][i] != evt)
				nbb++;
			else break;
		}
		//we break a sequence of size nba + nbb + 1 (thus a cost of (nba + nbb - 1))
		//however we can leave behind too subsequences that cost something :) !
		return Math.max(nba + nbb - 1, 0) - Math.max(nba - 2, 0) - Math.max(nbb - 2, 0);
	}

//**********************************************************//
//******** assert data structures **************************//
//**********************************************************//

	public boolean assertFeasible() {
		int[] cardEvtsPerTimeslot = new int[T];
		for (int e = 0; e < E; e++) {
			int er = rooms[e];
			int et = timeslot[e];
			cardEvtsPerTimeslot[et]++;
			assert (er != -1);
			assert (et != -1);
			assert (roomFeas[e][er] + timeslotFeas[e][et] == 0);
			for (int r = 0; r < R; r++) {
				if (er != r) {
					int e2 = matrix[et][r];
					if (e2 != -1) {
						assert (conflicts[e][e2] == 0);
					}
				}
			}
			for (int i = 0; i < greaterthan[e].length; i++) {
				assert (timeslot[greaterthan[e][i]] < et);
			}
		}
		for (int i = 0; i < T; i++) {
			assert (nbept[i] == cardEvtsPerTimeslot[i]);
		}
		for (int i = 0; i < E; i++) {
			int eit = timeslot[i];
			boolean isInList = false;
			for (int j = 0; j < nbept[eit] && !isInList; j++) {
				if (listevts[eit][j] == i) isInList = true;
			}
			if (!isInList) System.out.println("evt " + i + " in " + eit);
			assert (isInList);
		}
		return true;
	}

	public boolean assertAllSoftCosts() {
		assertFeasible();
		for (int s = 0; s < S; s++) {
			for (int d = 0; d < D; d++) {
				int nbevt = studentPlanning[s][d][0];
				int cpt = 0;
				for (int t = 1; t < studentPlanning[s][d].length; t++) {
					int evt = studentPlanning[s][d][t];
					if (evt != -1) {
						cpt++;
						int expectedtimslot = (d * H + (t - 1));
						if (timeslot[evt] != expectedtimslot)
							System.out.println(evt + " in timeslot " + timeslot[evt] + " or " + expectedtimslot);
						assert (timeslot[evt] == expectedtimslot);
					}
				}
				if (nbevt != cpt) {
					System.out.println("student " + s + " nb event in day " + d + " : " + cpt + " or " + nbevt + " " + Arrays.toString(studentPlanning[s][d]));
				}
				assert (nbevt == cpt);
			}
		}
		computeC1FromScratch();
		computeC2FromScratch();
		computeC3FromScratch(false);
		assert (softObj == softObj1 + softObj2 + softObj3);
		return true;
	}

//*********** print cost ************//

	public void printCosts() {
		System.out.println("Soft cost 1: " + softObj1);
		System.out.println("Soft cost 2: " + softObj2);
		System.out.println("Soft cost 3: " + softObj3);
		System.out.println("Soft cost : " + softObj);
	}
}
