package localsearch.graphalgos;

import localsearch.SolutionSpace;

import java.util.Arrays;
import java.util.Collections;
import java.util.ArrayList;

/**
 * (c) 2005, Bouygues, e-lab.
 * <p/>
 * Guillaume Rochart
 */
public class Hongrois {

  protected int rowNb;
  protected int colNb;
  protected int[][] costs;
  protected int[] affectation;
  protected int cost;
  protected ArrayList<Integer> permutation;	

  public Hongrois(int[][] costs, SolutionSpace space) {
    rowNb = costs.length;
    colNb = costs[0].length;
    this.costs = new int[rowNb][colNb];
    permutation = new ArrayList<Integer>();
	for (int i = 0; i < costs.length; i++) {
        permutation.add(i);
	}
	Collections.shuffle(permutation,space.rand);
	for (int i = 0; i < costs.length; i++) {
    	int pi = permutation.get(i);
		System.arraycopy(costs[i], 0, this.costs[pi], 0, costs[i].length);
	}
    this.affectation = new int[rowNb];
  }

  public Hongrois(int[][] costs) {
	  rowNb = costs.length;
	  colNb = costs[0].length;
	  this.costs = new int[rowNb][colNb];
	  for (int i = 0; i < costs.length; i++) {
		System.arraycopy(costs[i], 0, this.costs[i], 0, costs[i].length);
	  }
	  this.affectation = new int[rowNb];
	}


  public String toString() {
    StringBuffer buf = new StringBuffer();
    for (int i = 0; i < costs.length; i++) {
      int[] cost = costs[i];
      for (int j = 0; j < cost.length; j++) {
        int c = cost[j];
        buf.append(c + " \t\t");
      }
      buf.append("\n");
    }
    return buf.toString();
  }

  public void solve() {
    int[] colMate = new int[rowNb];
    int[] unchosenRow = new int[rowNb];
    int[] rowDec = new int[rowNb];
    int[] slackRow = new int[rowNb];

    int[] rowMate = new int[colNb];
    int[] parentRow = new int[colNb];
    int[] colInc = new int[colNb];
    int[] slackCol = new int[colNb];

    cost = 0;
    int t = 0;

    for (int row = 0; row < rowNb; row++) {
        this.affectation[row] = -1;
    }

    // 1. On retire les minima sur les lignes et colonnes pour simplifier
    for(int row = 0; row < rowNb; row++) {
      int s = costs[row][0];
      for(int col = 1; col < colNb; col++) if (costs[row][col] < s) s= costs[row][col];
      cost += s;
      if (s != 0) for(int col = 0; col < colNb; col++) costs[row][col] -= s;
    }
    for(int col = 0; col < colNb; col++) {
      int s = costs[0][col];
      for(int row = 1; row < rowNb; row++) if (costs[row][col]  < s) s = costs[row][col];
      cost += s;
      if (s != 0) for(int row = 0; row < rowNb; row++) costs[row][col] -= s;
    }

    // 2. Initialisation
    for(int col = 0; col < colNb; col++) {
      rowMate[col] = -1;
      parentRow[col] = -1;
      colInc[col] = 0;
      slackCol[col] = Integer.MAX_VALUE;
    }

    loop1: for(int row=0; row < rowNb; row++) {
      int s = costs[row][0];
      for(int col = 1; col < colNb; col++) if (costs[row][col] < s) s = costs[row][col];
      //System.out.println(row + " dec = " + s);
      rowDec[row] = s;
      for(int col = 0; col < colNb; col++) if (s==this.costs[row][col] && rowMate[col]<0) {
        colMate[row] = col;
        rowMate[col] = row;
        //System.out.println(col + " -> " + row + " (" + costs[row][col] + ")");
        continue loop1;
      }
      colMate[row] = -1;
      unchosenRow[t++] = row;
    }

    //System.out.println("current state : \n" + this);

    if (t!=0) {
      int unmatched = t;
      while (true) {
        //System.out.println("Matched " + (rowNb - t) + " rows.");
        int q = 0;
        int row,col;
        loop2: while (true) {
          while (q < t) {
            row = unchosenRow[q];
            int dec = rowDec[row];
            //System.out.println("Handling row " + row + " (" + dec + ")");
            for(col = 0; col < colNb; col++) {
              if (slackCol[col] != 0) {
                int del = costs[row][col] - dec + colInc[col];
                if (del < slackCol[col]) {
                  if (del == 0) {
                    if (rowMate[col] < 0) break loop2;
                    slackCol[col] = 0;
                    parentRow[col] = row;
                    unchosenRow[t++] = rowMate[col];
                  } else {
                    //System.out.println("Slack[" + col + "] = " + del);
                    //System.out.println("SlackRow[" + col + "] = " + row);
                    slackCol[col] = del;
                    slackRow[col] = row;
                  }
                }
              }
            }
            q++;
          }

          int s = Integer.MAX_VALUE;
          for(col=0; col < colNb; col++) if (slackCol[col] != 0 && slackCol[col] < s) s = slackCol[col];
          //System.out.println("s = " + s);
          for(q=0; q<t; q++) rowDec[unchosenRow[q]] += s;
          for(col = 0; col < colNb; col++) if (slackCol[col] != 0) {
            slackCol[col] -= s;
            if (slackCol[col] == 0) {
              row = slackRow[col];
              if (rowMate[col] < 0) {
                for(int j = col+1; j < colNb; j++) if (slackCol[j] == 0) colInc[j] += s;
                break loop2;
              } else {
                parentRow[col] = row;
                unchosenRow[t++]=rowMate[col];
              }
            }
          } else colInc[col] += s;
        } // fin loop 2

        while(true) {
          int j = colMate[row];
          colMate[row] = col;
          rowMate[col] = row;
          if (j < 0) break;
          row=parentRow[j];
          col=j;
        }

        if (--unmatched == 0) break;

        t=0;
        for(col=0; col<colNb; col++) {
          parentRow[col] = -1;
          slackCol[col] = Integer.MAX_VALUE;
        }
        for(row=0; row < rowNb; row++) {
          if (colMate[row] < 0) {
            unchosenRow[t++] = row;
          }
        }
      }

    }


    // Check
    /*for(int row = 0; row < rowNb; row++) for(int col= 0 ; col < colNb; col++)
      if (costs[row][col] < rowDec[row] - colInc[col]) System.out.println("Problme 1 !!");
    for(int row = 0; row < rowNb; row++) {
      int col = colMate[row];
      if (col < 0 || costs[row][col] < rowDec[row] - colInc[col]) System.out.println("Problme 2 !!");
    }
    int k = 0;
    for(int col = 0; col < colNb; col++) if (colInc[col] != 0) k++;
    if (k > rowNb) System.out.println("Problme 3 !!");
    */
    // Fin Check

    for(int row = 0; row < rowNb; row++) {
      for(int col =0; col < colNb; col++) {
        costs[row][col] += colInc[col] - rowDec[row];
      }
    }

    for(int row=0; row < rowNb; row++) {
      this.affectation[row] = colMate[row];
    }

    for(int row=0; row < rowNb; row++) cost+=rowDec[row];
    for(int col=0; col < colNb; col++) cost-=colInc[col];
  }

  public int getAffectedTask(int resource) {
    return this.affectation[permutation.get(resource)];
  }

  public int getSimpleAffectedTask(int resource) {
	  return this.affectation[resource];
  }


  public int getCost() {
    return cost;
  }

  /*public static void main(String[] args) {
    int[][] couts = new int[][]{{90, 75, 75, 80}, {35, 85, 55, 65}, {125, 95, 90, 105}, {45, 110, 95, 115}};
    int cost = 0;
    Hongrois monHongrois = new Hongrois(couts);
    monHongrois.solve();
    for(int task = 0; task < 4; task++) {
      System.out.println("Tache " + task + " associe  " + monHongrois.getAffectedTask(task) +
          " (" + couts[task][monHongrois.getAffectedTask(task)] + ")");
      cost+= couts[task][monHongrois.getAffectedTask(task)];
    }

    System.out.println("Cout global = " + cost);
  }*/
}
