package localsearch.graphalgos.matchings;

import choco.util.IntIterator;

import java.util.Arrays;

/**
 * Created by IntelliJ IDEA.
 * User: grichaud
 * Date: 13 oct. 2005
 * Time: 12:43:32
 * To change this template use File | Settings | File Templates.
 */

/**
 * DoubleLinkedList: implemnts a double linked list permiting access, add et del in o(1)
 * The size of the list is constante
 */
public class DoubleLinkedList implements IntIterator {
	// Successors table
	protected int[] nextT;

	// Predecessors table
	protected int[] prevT;


	// Current number of elements
	protected int size = 0;

	/**
	 * Utilises pour l'iteration
	 */
	protected int currentT = 0;
	//protected int currentI = 0;

	// Size of the list (premier pointeur)
	protected int listSize;


	// Constructor
	public DoubleLinkedList(int listSize) {
		this.listSize = listSize;
		this.nextT = new int[listSize + 1];
		this.prevT = new int[listSize + 1];
		for (int i = 0; i < nextT.length; i++) {
			nextT[i] = -1;
			prevT[i] = -1;
		}
	}

	// Constructor (copy)
	public DoubleLinkedList(DoubleLinkedList origin) {
		this.nextT = new int[origin.nextT.length];
		this.prevT = new int[origin.prevT.length];
		System.arraycopy(origin.nextT, 0, this.nextT, 0, this.nextT.length);
		System.arraycopy(origin.prevT, 0, this.prevT, 0, this.prevT.length);
		this.size = origin.size;
		this.listSize = origin.listSize;
	}

	// Add an element "val"
	public void addVal(int val) {
		if (!isIn(val)) {
			if (size == 0) {
				nextT[listSize] = val;
				prevT[listSize] = val;
			} else {
				prevT[nextT[listSize]] = val;
				nextT[val] = nextT[listSize];
				nextT[listSize] = val;
			}

			size = size + 1;
		}
	}


	// Remove an element "val"
	public void removeVal(int val) {
		if (isIn(val)) {
			if (size == 1) {
				nextT[listSize] = -1;
				prevT[listSize] = -1;
			} else if (nextT[listSize] == val) {
				nextT[listSize] = nextT[val];
				prevT[nextT[val]] = -1;
			} else if (prevT[listSize] == val) {
				prevT[listSize] = prevT[val];
				nextT[prevT[val]] = -1;
			} else {
				nextT[prevT[val]] = nextT[val];
				prevT[nextT[val]] = prevT[val];
			}
			nextT[val] = -1;
			prevT[val] = -1;
			size--;
		}
	}

	// Get current number of element
	public int getSize() {
		return size;
	}

	/**
	 * reset
	 */
	public void reset() {
		size = 0;
		Arrays.fill(nextT, -1);
		Arrays.fill(prevT, -1);
	}

	public void empty() {
		//System.out.println("" + toString());
		restart();
		while (hasNext()) {
			next();
			remove();
		}
	}

	//to check
	public boolean isIn(int val) {
		return nextT[val] != -1 || prevT[listSize] == val;
	}

	// Iterator
	public void restart() {
		currentT = listSize;
	}

	public void restartFrom(int val) {
		currentT = val;
	}

	public boolean hasNext() {
		return ((currentT != -1) && (prevT[currentT] != -1));
	}

	public boolean hasNextTo(int val) {
		return ((currentT != -1) && (prevT[currentT] != val));
	}


	public int next() {
		currentT = prevT[currentT];
		return currentT;
	}


	public int read() {
		return currentT;
	}

	// Remove current element
	public void remove() {
		removeVal(currentT);
		restart();
	}


	// Display the table
	public void AfficheTab() {
		for (int i = 0; i < nextT.length; i++) {
			System.out.println(nextT[i] + " | " + prevT[i]);
		}
		System.out.println("---");
	}


	// Display the table
	public String toString() {
		String n = "next    :";
		restart();
		while (hasNext()) {
			n += next() + " ";
		}
		return ("first :" + nextT[listSize] + "| " + "last :" + prevT[listSize] + "| " + n);
	}


	public static void main(String[] args) {
		DoubleLinkedList dl = new DoubleLinkedList(10);
		//dl.empty();
		dl.addVal(4);
		dl.addVal(3);
		System.out.println(dl.currentT + " " + dl.size);
		dl.AfficheTab();
		/*dl.addVal(7);
				dl.addVal(9);
				System.out.println(dl.toString());
				System.out.println("is in " + dl.isIn(3));
				System.out.println("is in " + dl.isIn(1));
				System.out.println("is in " + dl.isIn(7));
				System.out.println("is in " + dl.isIn(9));
				dl.removeVal(3);
				System.out.println(dl.toString());
				dl.empty();
				System.out.println(dl.toString());*/
	}

}