package localsearch.graphalgos.matchings;

import localsearch.SolutionSpace;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: 17 nov. 2007
 * Time: 09:22:58
 * To change this template use File | Settings | File Templates.
 */
public class Node {

	protected SolutionSpace space;

	//the list in terms of idxInGraph
	protected DoubleLinkedList succ;

	//the index of the corresponding node
	private int ref;

	//index of the node (room or event) in the bipartite graph
	//for rooms ref = idxInGraph
	protected int idxInGraph = -1;

	//index of the matched node
	private int matchNode = -1;

	public Node(int ref, int idx, SolutionSpace space) {
		this.setRef(ref);
		this.idxInGraph = idx;
		this.succ = new DoubleLinkedList(space.R);
		this.space = space;
	}

	/**
	 * add all compatible rooms with event ref as successors
	 * of this node
	 * @param space
	 */
	public void createLeftSucc(SolutionSpace space) {
		for (int i = 0; i < space.R; i++) {
			  if(space.roomFeas[getRef()][i] == 0)
				  succ.addVal(i);
		}
	}

	public void partialInitLeft(int i, Node[] right) {
		this.idxInGraph = i;
		this.initRightNodes(right);
	}

	/**
	 * @return true if mode match at initialization
	 */
	public boolean initLeftNode(int i, int croom, Node[] right) {
		partialInitLeft(i,right);
		if (right[croom].getMatchNode() == -1 &&
			space.roomFeas[getRef()][croom] == 0) { //initialize the matching			System.out.println("croom " + croom);
			this.setMatchNode(croom);
		    right[croom].setMatchNode(i);
			return true;
		} else {
			this.setMatchNode(-1);
			return false;
		}
	}

	/**
	 * add the node as a neighbour of all the rooms
	 * where it can fit
	 * @param rights the nodes representing the rooms
	 */
	public void initRightNodes(Node[] rights) {
		succ.restart();
		while(succ.hasNext()) {
			rights[succ.next()].succ.addVal(idxInGraph);
		}
	}

	/**
	 * remove this node from the bipartite graph
	 * @param rights the nodes representing the rooms
	 */
	public void resetRightNodes(Node[] rights) {
		if (getMatchNode() != -1) {
			rights[matchNode].setMatchNode(-1);
		}
		succ.restart();
		while(succ.hasNext()) {
			rights[succ.next()].succ.removeVal(idxInGraph);
		}
	}

	public String toString() {
		return "(" + getRef() + "," + idxInGraph + ")";// + " match:" + matchNode + " succ:" + succ;
	}

	public void setMatchNode(int matchNode) {
		this.matchNode = matchNode;
	}

	public int getMatchNode() {
		return matchNode;
	}

	public void setRef(int ref) {
		this.ref = ref;
	}

	public int getRef() {
		return ref;
	}
}
