package localsearch.greedy;

import localsearch.solvers.complete.HardSolver;
import localsearch.SolutionSpace;

import java.util.*;

import instance.TimetablingInstance;
import instance.ColourableConflictGraph;
import instance.colouring.MinNodeSelector;
import util.IntSet;

/**
 * A greedy base on graph coloring heuristics
 */
public class ColoringGreedy {

	public SolutionSpace space;
	public TimeslotComparator tcomp;
	public ColourableConflictGraph cgraph;

	public ColoringGreedy(SolutionSpace sp, TimetablingInstance ins) {
		space = sp;
		cgraph = ins.getFullConflictGraph();
		cgraph.setRandom(sp.rand);
	}

	/**
	 * initialize the solution of the search space based on
	 */
	public void fillAssignmentMatrix() {
		space.initSolutionSpace();
		space.restart();
		MinNodeSelector ms = new MinNodeSelector(cgraph);
		ms.setRandom(space.rand);
		ms.greedyList();
		int[] rooms = new int[space.T];
		List<Integer> unassignedEvts = new LinkedList<Integer>();
		//System.out.println("ratio: " + ((double) cgraph.getNumberOfAssignedNodes() / cgraph.getNNodes()));
		for (int i = 0; i < space.E; i++) {
			IntSet colours = cgraph.colourList(i);
			if (colours.size() == 1) {
				space.addEvent(i, colours.first(), rooms[colours.first()]);
				rooms[colours.first()]++;
			} else unassignedEvts.add(i);
		}
		Collections.shuffle(unassignedEvts, space.rand);
		randomlyFilledRemainingEvts(unassignedEvts,rooms);
		space.initMissingCosts();
		assert (space.assertAllHardCosts());
	}

	/*
	 * randomly filled the rest of the solution by trying to balance the
	 * timeslots
	 * @param events
	 */
	public void randomlyFilledRemainingEvts(List<Integer> events, int[] rooms) {
		Integer[] slots = new Integer[space.T];
		for (int i = 0; i < slots.length; i++) {
			slots[i] = i;
		}
		Iterator it = events.iterator();
		Arrays.sort(slots, tcomp); //sort the timeslots by increasing number of events
 	    int tidx = 0;
		while(it.hasNext()) {
			int evt = (Integer) it.next();
			int timeslot = slots[tidx];
		    while(rooms[timeslot] == space.R) {
			    tidx++;
			    timeslot = slots[tidx];
		    }
			space.addEvent(evt, timeslot, rooms[timeslot]);
			rooms[timeslot]++;
			tidx = (tidx + 1) % space.T;
	    }

	}


	public class TimeslotComparator implements Comparator {
		public int compare(Object o1, Object o2) {
			int t1s = space.nbept[(Integer) o1];
			int t2s = space.nbept[(Integer) o2];
			if (t1s > t2s)
				return 1;
			else if (t1s == t2s) {
				return 0;
			} else
				return -1;
		}
	}

	public static void main(String[] args) {
		HardSolver s = new HardSolver("data/Track2/comp-2007-2-1.tim", 0);
		ColoringGreedy greedy = new ColoringGreedy(s.space, s.ins);
		greedy.fillAssignmentMatrix();
		s.space.printCost();
	}
}
