package localsearch.greedy.acogreedy;

import choco.integer.IntDomainVar;
import choco.integer.search.ValSelector;
import choco.util.DisposableIntIterator;

import java.util.Random;
import java.util.Iterator;
import java.util.Arrays;

import cpmodel.chocoext.IndexedIntVar;

/**
 * Choose a value regarding a non uniform value distribution 
 **/
public class ProbabilisticValueChooser implements ValSelector {

	/**
     * tau[i][j]: the probability of assigning event i in timeslot j
     * tau is maintained to reflect the information accumulated on w
     */
    public float[][] w;

    public Random rand;
    
    public ProbabilisticValueChooser(float[][] w, Random rand) {
        this.w = w;
        this.rand = rand;
    }


	/**
     * Returns a value x randomly following a non uniform distribution 
     **/
    public int getBestVal(IntDomainVar x) {
        //System.out.println("choose value for " + x);
        int evt = ((IndexedIntVar) x).getIdx();
        int[] dom = getDomain(x);
        float[] proba = new float[dom.length];
        float total = 0f;
        for (int i = 0; i < proba.length; i++) {
            proba[i] = w[evt][dom[i] - 1];
            total += proba[i];
        }
        for (int i = 0; i < proba.length; i++) {
            proba[i] = proba[i] / total;
        }
//        float p = 0;
//        for (int i = 0; i < proba.length; i++) {
//            p += proba[i];
//        }
//        if (p < 0.99f)
//            System.out.println("" + p);
        //if (proba.length < 10) System.out.println(Arrays.toString(proba));
        return dom[nextInt(proba)];
	}

    public int[] getDomain(IntDomainVar x) {
        int[] dom = new int[x.getDomainSize()];
        int cpt = 0;
        DisposableIntIterator it = x.getDomain().getIterator();
        while(it.hasNext()) {
            int val = it.next();
            dom[cpt] = val;
            cpt++;
        }
        it.dispose();
        return dom;
    }

    //*************************************************************//
    //*************** Non uniform distribution ********************//
    //*************************************************************//

    /**
     * @param distribution
     * @return an integer between 0 and distribution.lenght following the random
     *         distribution given in argument
     */
    public int nextInt(float[] distribution) {
        float rval = rand.nextFloat();
        float cumulProba = 0;
        for (int i = 0; i < distribution.length; i++) { //dichotomy could be used if the density function was computed in advance
            cumulProba += distribution[i];
            if (cumulProba >= rval) return i;
        }
        System.out.println("rval: " + rval);
        for (int i = 0; i < distribution.length; i++) {
            System.out.println(distribution[i]);
        }
        throw new Error("nextInt for non uniform distribution didn't terminate properly"); //should never occur if the distribution is a probability (sum to 1)
    }


    /**
     * Rejection method in two steps
     * @param distribution
     * @return an integer between 0 and distribution.lenght following the random
     *         distribution given in argument
     */
    public int nextRejectionInt(float[] distribution) {
        while(true) {
            int idx = rand.nextInt(distribution.length);
            float rval = rand.nextFloat();
            if (rval <= distribution[idx]) return idx;
        }
    }


    //*************************************************************//
    //*************** Tests ***************************************//
    //*************************************************************//

    public void checkNonUniformDist(float[] dist, int nbTirage) {
        int[] nbVal = new int[dist.length];
        for (int i = 0; i < nbTirage; i++) {
            nbVal[nextInt(dist)]++;
        }
        for (int i = 0; i < nbVal.length; i++) {
            System.out.print(" " + (float) nbVal[i] / (float) nbTirage);
        }
        System.out.println("");
    }


}
