package localsearch.metah;

import localsearch.moves.complete.*;
import localsearch.moves.Move;
import localsearch.solvers.complete.AbstractSolver;

/**
 * The tabu heuristic augmented with a mechanism to
 * avoid a too long search on a plateau by adding a probability
 * to accept a side walk move that decrease with the lenght of the
 * plateau.
 **/
public class ProbabilisticTabu extends Tabu {

	public int nbSideWalk = 0;
	public double invk;

	public ProbabilisticTabu(AbstractSolver absolver, int tabusize) {
		super(absolver, tabusize);
		invk = 1d / 200000d;
	}

	// the decrease of the probability is linear so that
	// p(x) = - 1/k * x + 1
	// p(x) is the probability of accepting a sidewalk step
	// after having taking x previous sidewalk step
	public boolean probabilisticAccept() {
		//System.out.println("probabilistic accept " + space.rand.nextDouble() + " vs " + (1 - invk*nbSideWalk));
		return space.rand.nextDouble() < (1 - invk*nbSideWalk);
	}

	public void informOfMove(Move mov) {
		super.informOfMove(mov);
		if (mov.delta == 0) {
			nbSideWalk++;
		} else {
			nbSideWalk = 0;
		}
    }

	//assumption : called before performing the move !
    public boolean accept(SwapEvents swe, int delta) {
		boolean a = super.accept(swe,delta);
		if (a) {
			if (delta == 0) return probabilisticAccept();
			else return true;
		} else {
			 return false;
		}
	}

    //assumption : called before performing the move !
    public boolean accept(TranslateEvent tre, int delta) {
	    boolean a = super.accept(tre,delta);
	    if (a) {
		    if (delta == 0) return probabilisticAccept();
		    else return true;
	    } else {
		     return false;
	    }
    }

	//assumption : called before performing the move !
	public boolean accept(SwapTimeslots swt, int delta) {
		boolean a = super.accept(swt,delta);
		if (a) {
			if (delta == 0) return probabilisticAccept();
			else return true;
		} else {
			 return false;
		}
	}

	//assumption : called before performing the move !
    public boolean accept(VHungarianMove vmove, int delta) {
		boolean a = super.accept(vmove,delta);
		if (a) {
			if (delta == 0) return probabilisticAccept();
			else return true;
		} else {
			 return false;
		}
    }


	public boolean accept(CPMove cpm, int delta) {
		boolean a = super.accept(cpm,delta);
		if (a) {
			if (delta == 0) return probabilisticAccept();
			else return true;
		} else {
			 return false;
		}
	}
}
