package localsearch.metah;

import localsearch.solvers.complete.AbstractSolver;
import localsearch.moves.complete.*;
import localsearch.moves.Move;

import java.util.Arrays;

/**
 *
 */
public class ReinforcementLearning extends MetaHeuristic {

	protected float[] noise;

	protected int level; //range from 0 to 20

	//three paramters of the R-learning algorithm
	protected float alpha = 0.1f, epsilon = 0.01f, beta= 0.001f;

	/**
	 * Qsa[i][j] gives the reward (estimation) of taking action j
	 * (a noise level) in state i.
	 */
	protected float[][] Qsa;

	/**
	 * The average reward
	 */
	protected float rho;

	public ReinforcementLearning(AbstractSolver abs) {
		super(abs);
		noise = new float[19];
		noise[0] = 0.05f;
		for (int i = 1; i < noise.length; i++) {
			noise[i] = noise[i - 1] + 0.05f;
		}
		Qsa = new float[getUbObj()][noise.length];
	}

	//todo
	public int getUbObj() {
		return 1000;
	}

	/**
	 * return the maximum reward of a given cost
	 *
	 * @param c
	 * @return
	 */
	public float getMaxReward(int c) {
		float maxreward = Float.MIN_VALUE;
		int cobj = abs.getCost();
		for (int i = 0; i < noise.length; i++) {
			if (maxreward < Qsa[cobj][i]) {
				maxreward = Qsa[cobj][i];
			}
		}
		return maxreward;
	}

	public static int[] levelCount = new int[19];

	/**
	 * epsilon-greedy policy:
	 * choose a random noise with a probability epsilon
	 * otherwise choose follow the advice of the action value table
	 */
	public void chooseNoiseLevel() {
		if (space.rand.nextFloat() <= epsilon) {
			level = space.rand.nextInt(noise.length);
			//System.out.println("random noise " + noise[level] + " at iteration " + abs.getNbIter());
		} else {
			float bestnoise = Float.MIN_VALUE;
			int cobj = abs.getCost();
			for (int i = 0; i < noise.length; i++) {
				if (bestnoise < Qsa[cobj][i]) {
					bestnoise = Qsa[cobj][i];
					level = i;
				}
			}
		}
		levelCount[level] ++;
		if (abs.getNbIter() % 100000 == 0) {
			System.out.println(rho + " " + Arrays.toString(levelCount));
		}
	}

	public void RLearning(int delta) {
		int sprime = abs.getCost();
		int s = sprime - delta;
		int r = -delta; //abs.getBestCost() - sprime;
		float maxrprime = getMaxReward(sprime);
		Qsa[s][level] = Qsa[s][level] +
						alpha * (r - rho + maxrprime - Qsa[s][level]);
		float maxr = getMaxReward(s);
		if (Qsa[s][level] == maxr) {
			rho += beta*(delta - rho + maxrprime - maxr);
		}
	}

	public void initMhParameters() {

	}

	public boolean genericAccept(int delta) {
		if (delta <= 0) {
			return true;
		} else {
			return space.rand.nextFloat() < noise[level];
		}
	}

	public boolean accept(SwapEvents swe, int delta) {
		return genericAccept(delta);
	}

	public boolean accept(TranslateEvent tre, int delta) {
		return genericAccept(delta);
	}

	public boolean accept(SwapTimeslots swt, int delta) {
		return genericAccept(delta);
	}

	public boolean accept(VHungarianMove vmove, int delta) {
		return genericAccept(delta);
	}

	public boolean accept(CPMove cpm, int delta) {
		return genericAccept(delta);
	}

	public void informOfMove(Move mov) {
		RLearning(mov.delta);
		chooseNoiseLevel();
	}
}
