package localsearch.metah;

import localsearch.solvers.complete.AbstractSolver;
import localsearch.moves.complete.*;
import localsearch.moves.Move;

import java.util.LinkedList;
import java.util.BitSet;


/**
 * A Tabu controller to guide the local search
 */
public class Tabu extends MetaHeuristic {

    //The tabu structure
    //a timeslot is forbidden to an event for a number of iteration
    protected BitSet[] tabutimeslots;
    protected LinkedList<Tabupair> tabulist;
    protected int tabusize = 10;

    public Tabu(AbstractSolver absolver, int tabusize) {
        super(absolver);
        this.tabulist = new LinkedList<Tabupair>();
        this.tabusize = tabusize;
        this.tabutimeslots = new BitSet[space.E];
        for (int i = 0; i < space.E; i++) {
            tabutimeslots[i] = new BitSet(space.T);
        }
    }


	public void initMhParameters() {
		
	}

	public class Tabupair {
        protected int evt, timeslot;

        public Tabupair(int e, int t) {
            this.evt = e;
            this.timeslot = t;
        }

	    public String toString() {
		    return "(" + evt + "," + timeslot + ")";
	    }
    }

	public String toString() {
		return "Tabu List: " + tabulist.toString();
	}

    public void setTabuSize(int s) {
        tabusize = s;
    }

    public void informOfMove(Move mov) {
        switch (mov.id) {
            case AbstractSolver.TRANSLATE_EVENTS :
                TranslateEvent tre = (TranslateEvent) mov;
                addForbiddenPair(tre.evt,tre.oldtimeslot);
                break;
            case AbstractSolver.SWAP_EVENTS :
                SwapEvents swe = (SwapEvents) mov;
                addForbiddenPair(swe.getE1(),space.timeslot[swe.getE2()]);
	            addForbiddenPair(swe.getE2(),space.timeslot[swe.getE1()]);
                break;
        }
	    //System.out.println("metah: " + toString());	    
    }


    public void addForbiddenPair(int evt, int timeslot) {
        Tabupair tp = new Tabupair(evt,timeslot);
        if (tabulist.size() == tabusize) {
           Tabupair leavingpair = tabulist.removeLast();
           int tslot = leavingpair.timeslot;
           int e = leavingpair.evt;
           tabutimeslots[e].clear(tslot);
        }
        tabulist.addFirst(tp);
        tabutimeslots[evt].set(timeslot);
    }

//**************************************************************//
//***************** Check the tabu status of moves *************//
//**************************************************************//

    //assumption : called before performing the move !
    public boolean accept(SwapEvents swe, int delta) {
       return aspiration(delta) ||
              (!tabutimeslots[swe.getE1()].get(space.timeslot[swe.getE2()]) &&
              !tabutimeslots[swe.getE2()].get(space.timeslot[swe.getE1()])) ;
    }

    //assumption : called before performing the move !
    public boolean accept(TranslateEvent tre, int delta) {
        return aspiration(delta) ||
               !tabutimeslots[tre.getEvt()].get(tre.timeslot);
    }

	//assumption : called before performing the move !    	
	public boolean accept(SwapTimeslots swt, int delta) {
		if (aspiration(delta)) return true;
        for (int i = 0; i < space.R; i++) {
            int evt = swt.timet1[i];
            if (evt != -1) {
                if (tabutimeslots[evt].get(swt.t2))
                    return false;
            }
			int evt2 = swt.timet2[i];
            if (evt2 != -1) {
                if (tabutimeslots[evt2].get(swt.t1))
                    return false;
            }
		}
		return true;
	}

	//assumption : called before performing the move !
    public boolean accept(VHungarianMove vmove, int delta) {
        if (aspiration(delta)) return true;
        for (int i = 0; i < space.T; i++) {
            int evt = vmove.evts[i];
            if (evt != -1) {
                int timeslot = vmove.HungarishAlgo.getAffectedTask(i);
                if (tabutimeslots[evt].get(timeslot))
                    return false;
            }
        }
        return true;
    }


	public boolean accept(CPMove cpm, int delta) {
		return true;
	}
}
