package localsearch.moves.coloring;

import localsearch.moves.complete.SwapEvents;
import localsearch.solvers.coloring.HardColoringSolutionSpace;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: Sep 28, 2008
 * Time: 1:38:14 PM
 * To change this template use File | Settings | File Templates.
 */
public class ColoringSwapEvents extends SwapEvents {

    public ColoringSwapEvents(HardColoringSolutionSpace space) {
        super(space);
    }

    /**
     * Evaluate the variation of the objective function implied by
     * swapping two events
     */
    public int evalCostMove() {
        int e1t = space.timeslot[e1];
        int e2t = space.timeslot[e2];
        //availability
        int ca = space.timeslotFeas[e1][e2t] + space.timeslotFeas[e2][e1t];
        //neigbhours costs
        int cn = 0;
        for (int i = 0; i < space.nbept[e1t]; i++) {
            int ei = space.listevts[e1t][i];
            cn += space.conflicts[ei][e2];
        }
        for (int i = 0; i < space.nbept[e2t]; i++) {
            int ei = space.listevts[e2t][i];
            cn += space.conflicts[ei][e1];
        }
        cn -= space.conflicts[e1][e2] + space.conflicts[e2][e1]; //these quantities were added during the previous loops
        space.timeslot[e1] = e2t;
        space.timeslot[e2] = e1t;
        int cp = space.evalPrecedenceCostFromMove(e1, e2, e1t, e2t) +
                space.evalPrecedenceCostFromMove(e2, e1, e2t, e1t);
        space.timeslot[e1] = e1t;
        space.timeslot[e2] = e2t;
        //System.out.println("eval swap of " + e1);
        delta = (ca + cn + cp) - (space.eventcost[e1] + space.eventcost[e2]);
        return delta;
    }


    /**
     * Perform the move by updating the data structure of the solution
     * space to reflect the change
     */
    public void performMove() {
        //System.out.println("perform " + this);
        int e1t = space.timeslot[e1];
        int e2t = space.timeslot[e2];
        //update room, timeslot, matrix, eventcost and global cost
        //neigbhours costs
        if (e1t != e2t) {
            space.removeEvent(e1);
            space.removeEvent(e2);
            ((HardColoringSolutionSpace) space).addEvent(e1, e2t);
            ((HardColoringSolutionSpace) space).addEvent(e2, e1t);
            space.maintainPrecedenceCostFromMove(e1, e2, e1t, e2t);
            space.maintainPrecedenceCostFromMove(e2, e1, e2t, e1t);
        } 
    }

    public String toString() {
        return "swap " + e1 + " (t:" + space.timeslot[e1] + ") and (t:" + space.timeslot[e2] + ") " + e2 + " for " + delta;
    }
}
