package localsearch.moves.coloring.softmoves;

import localsearch.moves.complete.softmoves.SoftSwapEvents;
import localsearch.SolutionSpace;
import localsearch.solvers.complete.AbstractSolver;
import localsearch.solvers.coloring.ColoringSolutionSpace;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: Sep 30, 2008
 * Time: 10:39:57 AM
 * To change this template use File | Settings | File Templates.
 */
public class ColoringSoftSwapEvents extends SoftSwapEvents {

    public ColoringSoftSwapEvents(SolutionSpace space) {
        this.id = AbstractSolver.SWAP_EVENTS;
        this.space = space;

    }

    public void setFromEvt(ColoringSoftSwapEvents sse) {
        this.e1day = sse.e1day;
        this.e2day = sse.e2day;
        this.e1 = sse.e1;
        this.e2 = sse.e2;
        this.idxe1 = sse.idxe1;
        this.idxe2 = sse.idxe2;
        this.delta = sse.delta;
    }

    /**
     * Is the swap keeping feasability of the current solution
     *
     * @return true if the move is feasible
     */
    public boolean isKeepingFeas() {
        assert (!space.before[e1][e2] && !space.before[e2][e1]);

        int e1t = space.timeslot[e1];
        int e2t = space.timeslot[e2];

        //1-) availability check
        if (space.timeslotFeas[e1][e2t] > 0 || space.timeslotFeas[e2][e1t] > 0)
            return false;

        //2-) compatibility check
        for (int i = 0; i < space.nbept[e1t]; i++) {
            int ei = space.listevts[e1t][i];
            if (ei != e1 && space.conflicts[ei][e2] > 0)
                return false;
        }
        for (int i = 0; i < space.nbept[e2t]; i++) {
            int ei = space.listevts[e2t][i];
            if (ei != e2 && space.conflicts[ei][e1] > 0)
                return false;
        }

        //3-) precedence checks
        space.timeslot[e1] = e2t;
        space.timeslot[e2] = e1t;
        if (!space.evalFeasPrecedenceFromMove(e1, e2, e1t, e2t) ||
            !space.evalFeasPrecedenceFromMove(e2, e1, e2t, e1t)) {
            space.timeslot[e1] = e1t;
            space.timeslot[e2] = e2t;
            return false;
        }
        space.timeslot[e1] = e1t;
        space.timeslot[e2] = e2t;

        return true;
    }


    /**
     * Perform the swap while taking the matching into account
     */
    public void performMove() {
        int oldcost = space.softObj; //for assertion
        //System.out.println("perform " + this);
        int e1t = space.timeslot[e1];
        int e2t = space.timeslot[e2];
        assert (e1t != e2t);

        space.softRemoveEvent(e1);
        space.softRemoveEvent(e2);
        ((ColoringSolutionSpace) space).softAddEvent(e1, e2t);
        ((ColoringSolutionSpace) space).softAddEvent(e2, e1t);

        if (e1day != e2day) {
            //update room, timeslot, matrix, eventcost and global cost
            //neigbhours costs
            space.removeEventSoftCost(e1, e1t);
            space.removeEventSoftCost(e2, e2t);
            space.addEventSoftCost(e1, e2t);
            space.addEventSoftCost(e2, e1t);
        } else {
            space.fastRemoveEventSoftCost(e1, e1t);
            space.fastRemoveEventSoftCost(e2, e2t);
            space.fastAddEventSoftCost(e1, e2t);
            space.fastAddEventSoftCost(e2, e1t);
        }
        assert (oldcost + delta == space.softObj);
    }

}
