package localsearch.moves.complete;

import localsearch.SolutionSpace;
import localsearch.moves.Move;
import localsearch.solvers.complete.AbstractSolver;

import java.util.BitSet;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: 25 janv. 2008
 * Time: 10:29:39
 * To change this template use File | Settings | File Templates.
 */
public class CPMove extends Move {

	public AbstractSolver s;

	//partial solution kept
	public BitSet freeEvents;
	public int[] partialAssignment;

	//new solution stored
	public int[] feasibleTimeslots;
	public int[] feasibleRooms;
	
	public LightModelMove lmm;

	public CPMove(SolutionSpace space, AbstractSolver s) {
		super(space);
		partialAssignment = new int[space.E];
		feasibleTimeslots = new int[space.E];
		feasibleRooms = new int[space.E];
		this.s = s;
		id = 6;		
	}

	public void initialize() {
		lmm = new LightModelMove(s.ins,s);
		lmm.initialize();
	}

	// we assume that the set of events contains
	// at least one conflicting events
	public void freeEvents(BitSet events) {
		freeEvents = events;
		for (int i = 0; i < space.E; i++) {
		   if (!events.get(i))
			   partialAssignment[i] = space.timeslot[i];
		   else partialAssignment[i] = -1;
		}
	}

	public int evalCostMove() {
		//System.out.println("eval cpmove with " + freeEvents.cardinality());
		assert(assertAllConflictingEvent());
		boolean realsolve = lmm.solve(partialAssignment);
		if (lmm.isFeasSolutionFound()) {
			for (int i = 0; i < space.E; i++) {
				feasibleTimeslots[i] = lmm.getTimeSlot(i);
				feasibleRooms[i] = lmm.getRoom(i); 
			}
			return -1;
			//no matter what it the exact improvement, 
			//it is a feasible solution
		} else {
			if (realsolve) return 1;
			else return 2;
		}
	}

	/**
	 * We reach a feasible solution by performing this move
	 */
	public void performMove() {
		for (int i = 0; i < space.E; i++) {
			space.removeEvent(i);
		}
		for (int i = 0; i < space.E; i++) {
			space.addEvent(i,feasibleTimeslots[i],feasibleRooms[i]);
		}
		for (int i = 0; i < space.E; i++) {
			space.eventcost[i] = 0;
			space.hardObj = 0;
		}
		assert(space.assertHardCostsWithoutFreeSpaces());
	}


	public boolean assertAllConflictingEvent() {
		for (int i = 0; i < partialAssignment.length; i++) {
			if (partialAssignment[i] != -1) {
				assert(space.eventcost[i] == 0);
			}
		}
		return true;
	}

	public String toString() {
		return "LNS move by CP for feasibility";
	}
}
