package localsearch.moves.complete;

import localsearch.solvers.complete.AbstractSolver;
import localsearch.SolutionSpace;
import localsearch.moves.Move;
import localsearch.graphalgos.matchings.Matching;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: 23 aot 2007
 * Time: 09:26:10
 * To change this template use File | Settings | File Templates.
 */
public class HHungarianMove extends Move {

    //A table of events to reallocate
    public int[] evts;
    //The timeslot to reorganize
    public int timeslot;
    public int nbEvt;

    // the Hungarish algorithm
	Matching matching;

	public HHungarianMove(SolutionSpace space) {
        super(space);
        this.id = AbstractSolver.HORIZ_HUNGARIAN_MOVE;
		matching = new Matching(space);
    }

    //the event choosed in timeslot i
    public void setTimeSlot(int i) {
        assert (i <= space.T);
        assert (i >= 0);
        timeslot = i;
	    evts = space.listevts[i];
	    nbEvt = space.nbept[i];
    }

    /**
     * Compute the cost of the current assignment
     *
     * @return
     */
    public int computeCurrentCost() {
        int cost = 0;
        for (int i = 0; i < nbEvt; i++) {
            int evt = evts[i];
            cost += space.roomFeas[evt][space.rooms[evt]];
        }
        return cost;
    }

    /**
     * Evaluate the objective function variation by optimally reassigning a set of
     * events in their timeslot.
     *
     * @return
     */
    public int evalCostMove() {
        int cost = 0;
        int ccost = computeCurrentCost();
        if (ccost == 0) return 1; // no way to do better, it is already optimal
	    matching.init(timeslot);
	    matching.computePerfectMatching();
	    cost = matching.getMatchingCost();
	    delta = cost - ccost;
        return delta;
    }

    public void performMove() {
        for (int i = 0; i < nbEvt; i++) {
            space.fastRemoveEvent(evts[i]);
        }
        for (int i = 0; i < nbEvt; i++) {
            int evt = evts[i];
            int room = matching.getMatch(i);
	        space.fastAddEvent(evt, room);
        }
        space.maintainFreePlacesOnTimeslot(timeslot);
    }

    public String toString() {
        return "Hungarian move intra timeslot " + timeslot + " for " + delta;
    }

}
