package localsearch.moves.complete;

import cpmodel.LightModel;
import cpmodel.PrimalModel;
import instance.TimetablingInstance;
import localsearch.solvers.complete.AbstractSolver;
import localsearch.solvers.complete.HardSolver;
import localsearch.SolutionSpace;

import java.util.ArrayList;
import java.util.Collections;

import choco.ContradictionException;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: 25 janv. 2008
 * Time: 10:37:00
 * To change this template use File | Settings | File Templates.
 */
public class LightModelMove extends LightModel {

	public boolean feasiblesolution;
	protected SolutionSpace space;
	protected ArrayList eventsList;
	

	public LightModelMove(TimetablingInstance ins, AbstractSolver s) {
		super(ins, s, LightModel.Impact);
		space = s.space;
		eventsList = new ArrayList<Integer>();
		for (int i = 0; i < ins.getNEvents(); i++) {
			eventsList.add(i);
		}
	}

	public boolean isFeasSolutionFound() {
		return feasiblesolution;
	}

	public int getTimeSlot(int i) {
		return eventTime[i].getVal() - 1;
	}

	public int getRoom(int i) {
		return solution[i];
	}


	public void initialize() {
		buildModel(PrimalModel.R_ALLDIFF);

		//Initialize Impact
		setAndInitializeImpact();
		
		//Store Initial propagation
		pb.worldPush();
		try {
			pb.propagate();
		} catch (ContradictionException e) {
			throw new Error("LightModelMove: instance should be consistent at initial Propagation");
		}
		pb.worldPush();
	}
	
	/**
	 * check if there exist a solution completing the partial
	 * assignment given by pa (value of the timeslot if known),
	 * -1 otherwise
	 *
	 * @param evts
	 * @param pa
	 * @return true if a real attempt has been made, false if the
	 *         relaxation found was not big enough
	 */
	public boolean solve(int[] pa) {
		boolean timeout = false;
		feasiblesolution = false;
		int level = pb.getWorldIndex();
		pb.worldPush();
		pb.getSolver().setTimeLimit(20000);
		int size = relax(pa);
		if (size <= HardSolver.sspacesize) {
			pb.solve();
			timeout = pb.getSolver().isEncounteredLimit() || !pb.isFeasible();
			while (!timeout && !allocateRooms()) {
				timeout = (pb.getSolver().isEncounteredLimit()) ||
						  (pb.nextSolution() == Boolean.FALSE);
			}
			if (timeout) {
				resetSolver();
				pb.worldPopUntil(level);
			} else {
				feasiblesolution = true;
			}
			return true;
		} else {
			resetSolver();
			pb.worldPopUntil(level);
			return false;
		}
	}


	public void resetSolver() {
		pb.feasible = null;
	}

	/**
	 * @param assignment
	 */
	public int relax(int[] assignment) {
		int nbFreeEvts = 0;
		Collections.shuffle(eventsList, space.rand);
		for (int j = 0; j < eventTime.length; j++) {
			int i = (Integer) eventsList.get(j);
			if (assignment[i] != -1) {
				pb.worldPush();
				boolean contradict = false;
				try {
					eventTime[i].setVal(assignment[i] + 1);
					pb.propagate();
				} catch (ContradictionException e) {
					contradict = true;
				}
				pb.worldPop();
				pb.getPropagationEngine().flushEvents(); //todo : why do I need this here ???								
				if (contradict) {
					nbFreeEvts++;
					assignment[i] = -1;
					try {
						eventTime[i].remVal(assignment[i] + 1);
						pb.propagate();
					} catch (ContradictionException e) {
						return Integer.MAX_VALUE;
					}
				} else {
					try {
						eventTime[i].setVal(assignment[i] + 1);
						pb.propagate();
					} catch (ContradictionException e) {
						throw new Error("error in relax (LightModelMove) " + eventTime[i]);
					}
				}
			} else nbFreeEvts++;
		}
		return nbFreeEvts;
	}
}
