package localsearch.moves.complete;

import localsearch.solvers.complete.AbstractSolver;
import localsearch.SolutionSpace;
import localsearch.moves.Move;


public class SwapEvents extends Move {

	//the two events to be swapped
	protected int e1, e2;

    public SwapEvents() {}

    public SwapEvents(SolutionSpace space) {
		super(space);
		this.id = AbstractSolver.SWAP_EVENTS;
	}

	public int getE1() {
		return e1;
	}

	public void setE1(int e1) {
		this.e1 = e1;
	}

	public int getE2() {
		return e2;
	}

	public void setE2(int e2) {
		this.e2 = e2;
	}

	public void setFromEvt(SwapEvents sse) {
		this.e1 = sse.e1;
		this.e2 = sse.e2;
		this.delta = sse.delta;
	}

	/**
	 * Evaluate the variation of the objective function implied by
	 * swapping two events
	 */
	public int evalCostMove() {
		int e1t = space.timeslot[e1];
		int e1r = space.rooms[e1];
		int e2t = space.timeslot[e2];
		int e2r = space.rooms[e2];
		if (e1t != e2t) {
			//cost projection on e1 and e2 themself
			//room capacities
			int cc = space.roomFeas[e1][e2r] + space.roomFeas[e2][e1r];
			//availability
			int ca = space.timeslotFeas[e1][e2t] + space.timeslotFeas[e2][e1t];
			//neigbhours costs
			int cn = 0;
			for (int i = 0; i < space.nbept[e1t]; i++) {
				int ei = space.listevts[e1t][i];
				cn += space.conflicts[ei][e2];
			}
			for (int i = 0; i < space.nbept[e2t]; i++) {
				int ei = space.listevts[e2t][i];
				cn += space.conflicts[ei][e1];
			}
			cn -= space.conflicts[e1][e2] + space.conflicts[e2][e1]; //these quantities were added during the previous loops
			space.timeslot[e1] = e2t;
			space.timeslot[e2] = e1t;
			int cp = space.evalPrecedenceCostFromMove(e1, e2, e1t, e2t) +
					space.evalPrecedenceCostFromMove(e2, e1, e2t, e1t);
			space.timeslot[e1] = e1t;
			space.timeslot[e2] = e2t;
			//System.out.println("eval swap of " + e1);
			delta = (cc + ca + cn + cp) - (space.eventcost[e1] + space.eventcost[e2]);
		} else {
			delta = space.roomFeas[e1][e2r] + space.roomFeas[e2][e1r]
					- space.roomFeas[e1][e1r] - space.roomFeas[e2][e2r];
		}
		return delta;
	}



	/**
	 * Perform the move by updating the data structure of the solution
	 * space to reflect the change
	 */
	public void performMove() {
		//System.out.println("perform " + this);
		int e1t = space.timeslot[e1];
		int e1r = space.rooms[e1];
		int e2t = space.timeslot[e2];
		int e2r = space.rooms[e2];
		//update room, timeslot, matrix, eventcost and global cost
		//neigbhours costs
		if (e1t != e2t) {
			space.removeEvent(e1);
			space.removeEvent(e2);
			space.addEvent(e1, e2t, e2r);
			space.addEvent(e2, e1t, e1r);
			space.maintainPrecedenceCostFromMove(e1, e2, e1t, e2t);
			space.maintainPrecedenceCostFromMove(e2, e1, e2t, e1t);
			//space.specialMaintainForSwap(e1,e2);
		} else {
			space.fastRemoveEvent(e1);
			space.fastRemoveEvent(e2);
			space.fastAddEvent(e1, e2r);
			space.fastAddEvent(e2, e1r);
		}
	}

	public String toString() {
		return "swap " + e1 + " (t:"+ space.timeslot[e1] +") and (t:"+ space.timeslot[e2] + ") " + e2 + " for " + delta;
	}
}
