package localsearch.moves.complete;

import localsearch.solvers.complete.AbstractSolver;
import localsearch.SolutionSpace;
import localsearch.moves.Move;


public class SwapTimeslots extends Move {

	//the two timeslots to be swapped
	public int t1, t2;

	//the timeslot defined as tables
	public int[] timet1, timet2;

	//the cost of each event regarding its precedence
	//before and after the swap of timeslots
	protected int[] opc, npc;


	public SwapTimeslots(SolutionSpace space) {
		super(space);
		opc = new int[space.E];
		npc = new int[space.E];
		this.id = AbstractSolver.SWAP_TIMESLOTS;
	}


	public int getT1() {
		return t1;
	}

	public void setT1(int t1) {
		this.t1 = t1;
		this.timet1 = space.matrix[t1];
	}

	public int getT2() {
		return t2;
	}

	public void setT2(int t2) {
		this.t2 = t2;
		this.timet2 = space.matrix[t2];
	}

	public int computeCost(int ta1, int ta2) {
		int tot = 0;
		for (int i = 0; i < space.nbept[t1]; i++) {
			int e1i = space.listevts[t1][i];
			tot += space.evalPrecCost(e1i, t1, t2) + space.timeslotFeas[e1i][ta1];
		}
		for (int i = 0; i < space.nbept[t2]; i++) {
			int e2i = space.listevts[t2][i];
			tot += space.evalPrecCost(e2i, t1, t2) + space.timeslotFeas[e2i][ta2];
		}
		return tot;
	}

	public void temporarySwapping(int firstt, int secondt) {
		for (int i = 0; i < space.nbept[t1]; i++) {
			int e1i = space.listevts[t1][i];
			space.timeslot[e1i] = secondt;
		}
		for (int i = 0; i < space.nbept[t2]; i++) {
			int e2i = space.listevts[t2][i];
			space.timeslot[e2i] = firstt;
		}
	}

	public int evalCostMove() {
		int pc = 0, nc = 0;
		//store the previous cost
		pc =  computeCost(t1, t2);
		temporarySwapping(t1, t2);
		//store the new cost
		nc = computeCost(t2, t1);
		temporarySwapping(t2, t1);
		delta = nc - pc;
		return delta;
	}


	public void performMove() {
		int a = space.nbept[t1];
		int b = space.nbept[t2];
		int[] la = space.listevts[t1];
		int[] lb = space.listevts[t2];
		for (int i = 0; i < space.R; i++) {
			int e1i = timet1[i];
			int e2i = timet2[i];
			if (e1i != -1) {
				if (e2i != -1) {
					space.fastSwitchTimeslot(e1i, e2i);
				} else {
					space.switchFreePlace(t2,space.rooms[e1i],t1,space.rooms[e1i]);
					space.fastMoveTimeslot(e1i, t2);
				}
			} else if (e2i != -1) {
				space.switchFreePlace(t1,space.rooms[e2i],t2,space.rooms[e2i]);
				space.fastMoveTimeslot(e2i, t1);
			}
		}
		for (int i = 0; i < space.R; i++) {
			int e2i = timet1[i];
			int e1i = timet2[i];
			if (e1i != -1) {
				space.maintainPrecedenceCostFromMove(e1i, t1, t2);
			}
			if (e2i != -1) {
				space.maintainPrecedenceCostFromMove(e2i, t2, t1);
			}
		}
		space.nbept[t2] = a;
		space.nbept[t1] = b;
		space.listevts[t1] = lb;
		space.listevts[t2] = la;
	}

	public String toString() {
		return "swap timeslots " + t1 + " and " + t2;
	}

}
