package localsearch.moves.complete;

import localsearch.solvers.complete.AbstractSolver;
import localsearch.Place;
import localsearch.SolutionSpace;
import localsearch.moves.Move;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: 20 aot 2007
 * Time: 10:00:04
 * To change this template use File | Settings | File Templates.
 */
public class TranslateEvent extends Move {

    public int evt, timeslot, room;
    public int oldtimeslot; //remember the timeslot where the evt comes from
    public Place freeslot;

    public TranslateEvent(){}

    public TranslateEvent(SolutionSpace space) {
        super(space);   
        this.id = AbstractSolver.TRANSLATE_EVENTS;
    }

    public int getEvt() {
        return evt;
    }

    public void setEvt(int evt) {
        this.evt = evt;
        this.oldtimeslot = space.timeslot[evt];
    }

	public void setFromEvt(TranslateEvent tre) {
		this.timeslot = tre.timeslot;
		this.room = tre.room;
		this.delta = tre.delta;
		this.freeslot = tre.freeslot;
	}

    public Place getFreeslot() {
        return freeslot;
    }

    public void setFreeslot(Place freeslot) {
        this.freeslot = freeslot;
        this.timeslot = freeslot.t;
        this.room = freeslot.r;
    }


    public String toString() {
        return "translate " + evt + " from timeslot " + oldtimeslot + " to timeslot " + timeslot + " and room " + room + " for " + delta;
    }


    public int evalCostMove() {
        int e1t = space.timeslot[evt];
        if (e1t != timeslot) {
            //room capacities
            int cc = space.roomFeas[evt][room];
            //availability
            int ca = space.timeslotFeas[evt][timeslot];
            //neigbhours costs
            int cn = 0;
			for (int i = 0; i < space.nbept[timeslot]; i++) {
				int ei = space.listevts[timeslot][i];
				cn += space.conflicts[ei][evt];
			}
            int cglobal = cc + ca + cn + space.evalPrecedenceCostFromMove(evt,-1,e1t,timeslot);
            delta = cglobal - space.eventcost[evt];
        } else {
	        delta = space.roomFeas[evt][room] - space.roomFeas[evt][space.rooms[evt]];
        }
	    return delta;  	    
    }

    public void performMove() {
        //System.out.println("perform " + this);
        int e1t = space.timeslot[evt];
        int e1r = space.rooms[evt];
        if (timeslot == e1t) {
            space.fastRemoveEvent(evt);
            space.fastAddEvent(evt,room);                
        } else {
            space.removeEvent(evt);
            space.addEvent(evt,timeslot,room);
            space.maintainPrecedenceCostFromMove(evt,-1,e1t,timeslot);
        }
        space.switchFreePlace(freeslot.t, freeslot.r, e1t, e1r);
    }
}
