package localsearch.moves.complete;

import localsearch.SolutionSpace;
import localsearch.graphalgos.MatchingInterface;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: 22 nov. 2007
 * Time: 17:03:47
 * To change this template use File | Settings | File Templates.
 */
public class TranslateMatchingEvent extends TranslateEvent {

	protected MatchingInterface matching;
	protected int[] matchingFound;

	public TranslateMatchingEvent(SolutionSpace space) {
		super(space);
		matching = new MatchingInterface(space);
	}

	public void setTimeslot(int t) {
		timeslot = t;
	}

	public String toString() {
		return "Matching translate " + evt + " from timeslot " + oldtimeslot + " to timeslot " + timeslot + " for " + delta;
	}

	public int getRoomCost() {
		int roomcost = 0;
		for (int i = 0; i < space.nbept[timeslot]; i++) {
			int oevt = space.listevts[timeslot][i];
			roomcost += space.roomFeas[oevt][space.rooms[oevt]];
		}
		return roomcost;
	}

	public int evalCostMove() {
		int e1t = space.timeslot[evt];
		assert (e1t != timeslot);
		//availability
		int ca = space.timeslotFeas[evt][timeslot];
		//neigbhours costs
		int cn = 0;
		for (int i = 0; i < space.nbept[timeslot]; i++) {
			int ei = space.listevts[timeslot][i];
			cn += space.conflicts[ei][evt];
		}
		//room capacities
		matching.setTimeSlot(timeslot, evt);
		if (matching.solve()) {
			matchingFound = matching.getMatching();
			room = matchingFound[matchingFound.length - 1];

			int cglobal = ca + cn + space.evalPrecedenceCostFromMove(evt, -1, e1t, timeslot);
			delta = cglobal - space.eventcost[evt] - getRoomCost();

		} else {
			delta = Integer.MAX_VALUE;
		}
		return delta;
	}

	public void performMove() {
		int oldcost = space.hardObj; //for assertion				

		//System.out.println("perform " + this);
		int e1t = space.timeslot[evt];
		int e1r = space.rooms[evt];

		//reallocate the events moved because of the matching
		for (int i = 0; i < space.nbept[timeslot]; i++) {
			int oevt = space.listevts[timeslot][i];
			space.fastRemoveEvent(oevt);
		}
		for (int i = 0; i < space.nbept[timeslot]; i++) {
			int oevt = space.listevts[timeslot][i];
			int newroom = matchingFound[i];
			space.fastAddEvent(oevt, newroom);
		}

		space.removeEvent(evt);
		space.addEvent(evt, timeslot, room);
		space.maintainPrecedenceCostFromMove(evt, -1, e1t, timeslot);

		space.addFreePlace(e1t,e1r);
		space.maintainFreePlacesOnTimeslot(timeslot);
		//space.assertAllSoftCosts();
		assert (oldcost + delta == space.hardObj);
		
	}

}
