package localsearch.moves.complete.softmoves;

import localsearch.moves.complete.CPMove;
import localsearch.SolutionSpace;
import localsearch.solvers.complete.AbstractSolver;

import java.util.BitSet;

/**
 *
 * A move calling the CP solver to achieve a Large Neighbourhood search 
 **/
public class SoftCPMove extends CPMove implements SoftMove{

	protected SoftLightModelMove lsmm;

	public SoftCPMove(SolutionSpace space, AbstractSolver s) {
		super(space, s);
		
	}

	public void initialize() {
		lsmm = new SoftLightModelMove(s.ins,s);
		lsmm.initialize();
	}


	public int evalSoftCost() {
		lsmm.solve(partialAssignment);
		if (lsmm.isFeasSolutionFound()) {
			for (int i = 0; i < space.E; i++) {
				feasibleTimeslots[i] = lsmm.getTimeSlot(i);
				feasibleRooms[i] = lsmm.getRoom(i);
			}
			delta = lsmm.getOptimalSolCost() - space.softObj;
			//System.out.println("delta : " + delta);
			return delta;
		} else {
			   return Integer.MAX_VALUE;
		}
	}


	public double getSSpaceSize() {
		return lsmm.ssd;
	}

	public int getNbNode() {
		return lsmm.nbnode;
	}


	/**
	 * We know there is at least the current solution
	 * as a feasible solution so this move is ensured to keep
	 * feasibility
	 * @return
	 */
	public boolean isKeepingFeas() {
		return true;
	}


	/**
	 * We are sure to reach a feasible solution
	 * by performing this move
	 */
	public void performMove() {
		//System.out.println("perform the move");
		BitSet impactedEvtToReplace = new BitSet();
		for (int i = 0; i < space.E; i++) {
			int t = space.timeslot[i];
			int r = space.rooms[i];
			if (freeEvents.get(i)){
				space.softRemoveEvent(i);
				space.removeEventSoftCost(i,t);
			} else if (feasibleRooms[i] != r) {
				impactedEvtToReplace.set(i);
				space.fastSoftRemoveEvent(i);
			}
		}
		for (int i = 0; i < space.E; i++) {
			if (freeEvents.get(i)) {
				space.softAddEvent(i, feasibleTimeslots[i], feasibleRooms[i]);
				space.addEventSoftCost(i, feasibleTimeslots[i]);
			} else if (impactedEvtToReplace.get(i)) {
				space.fastSoftAddEvent(i,feasibleRooms[i]);
			}
		}
		assert(space.assertAllSoftCosts());
	}

}
