package localsearch.moves.complete.softmoves;

import cpmodel.SoftLightModel;
import cpmodel.LightModel;
import cpmodel.PrimalModel;
import instance.TimetablingInstance;
import localsearch.solvers.complete.AbstractSolver;
import localsearch.solvers.complete.SoftSolver;
import localsearch.SolutionSpace;
import choco.ContradictionException;

import java.util.ArrayList;


/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: 11 fvr. 2008
 * Time: 09:15:11
 * To change this template use File | Settings | File Templates.
 */
public class SoftLightModelMove extends SoftLightModel {

	protected SoftSolver ss;
	protected SolutionSpace space;

	protected ArrayList eventsList;
	public boolean feasiblesolution;

	public SoftLightModelMove(TimetablingInstance ins, AbstractSolver s) {
		super(ins, s, LightModel.Impact);
		ss = (SoftSolver) s;
		space = s.space;
		eventsList = new ArrayList<Integer>();
		for (int i = 0; i < ins.getNEvents(); i++) {
			eventsList.add(i);
		}
		printCPResolution = false;
	}

	public boolean isFeasSolutionFound() {
		return feasiblesolution;
	}

	public int getTimeSlot(int i) {
		return storedTime[i];
	}

	public int getRoom(int i) {
		return storedRoom[i];
	}

	public int getOptimalSolCost() {
		return super.computeCurrentCost();
	}

	public void initialize() {
		buildModel(PrimalModel.R_ALLDIFF);
		//Initialize Impact
		setAndInitializeImpact(SoftSolver.discrepency);
		restart = false;
		//Store Initial propagation for all moves
		pb.worldPush();
		try {
			pb.propagate();
		} catch (ContradictionException e) {
			throw new Error("LightModelMove: instance should be consistent at initial Propagation");
		}
		pb.worldPush();
	}


	/**
	 * check if there exist a solution completing the partial
	 * assignment given by pa (value of the timeslot if known),
	 * -1 otherwise
	 *
	 * @param evts
	 * @param pa
	 * @return true if a real attempt has been made, false if the
	 *         relaxation found was not big enough
	 */
	public boolean solve(int[] pa) {
		feasiblesolution = false;
		stopAtFirstSol = false;
		boolean initContradiction = false;
		int level = pb.getWorldIndex();
		pb.worldPush();
		pb.getSolver().setNodeLimit(SoftSolver.nodelimit);
		pb.getSolver().setTimeLimit(ss.timelimit - ((int) System.currentTimeMillis() - ss.starttime));
		relax(pa);
		if (ss.LNS_SA) {
			try {
				objective.setSup(ss.getCost() - 1);
				pb.propagate();
			} catch (ContradictionException e) {
				initContradiction = true;
				//System.out.println("contradiction");
			}
		}

		if (!initContradiction)
			mainSoftLoop();

		nbnode = getNode();
		if (initContradiction)
			nbnode = 0;
		//System.out.println("nbNode " + nbnode + " vs " + SoftSolver.nodelimit + " " + initContradiction);
		resetSolver();
		pb.worldPopUntil(level);
		return feasiblesolution;

	}


	protected void saveImprovingSolution() {
		super.saveImprovingSolution();
		feasiblesolution = true;
		if (SoftSolver.stopAtFirstImprovingSol &&
			((storedCost1 + storedCost2 + storedCost3) < (ss.getCost() - 1))) {
			stopAtFirstSol = true;
		}
	}

	public void resetSolver() {
		pb.feasible = null;

	}

	/**
	 * @param assignment
	 */
	public int relax(int[] assignment) {
		int nbFreeEvts = 0;
		for (int j = 0; j < eventTime.length; j++) {
			int i = (Integer) eventsList.get(j);
			if (assignment[i] != -1) {
				try {
					eventTime[i].setVal(assignment[i] + 1);
					pb.propagate();
				} catch (ContradictionException e) {
					throw new Error("error in relax (SoftLightModelMove) " + eventTime[i] + " " + objective.getSup());
				}
				pb.getPropagationEngine().flushEvents();
			}
		}
		return nbFreeEvts;
	}


	public double ssd;
	public int nbnode;

	public void computeSSize() {
		ssd = 1d;
		for (int i = 0; i < eventTime.length; i++) {
			if (!eventTime[i].isInstantiated())
				ssd *= eventTime[i].getDomainSize();
		}
	}

}
