package localsearch.moves.complete.softmoves;

import localsearch.moves.complete.SwapEvents;
import localsearch.SolutionSpace;
import localsearch.graphalgos.MatchingInterface;

import java.util.Arrays;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: 24 oct. 2007
 * Time: 10:56:07
 * To change this template use File | Settings | File Templates.
 */
public class SoftSwapEvents extends SwapEvents implements SoftMove {

	protected int e1day, e2day;

	protected MatchingInterface matching;
	protected int[] matchingFound1;
	protected int[] matchingFound2;

	//index of e1 and e2 within the table encoding the result of the matching
	protected int idxe1, idxe2;

    public SoftSwapEvents(){}

    public SoftSwapEvents(SolutionSpace space) {
		super(space);
		matching = new MatchingInterface(space);
	}

	public void setE1(int e1) {
		this.e1 = e1;
		this.e1day = space.getDay(space.timeslot[e1]);
	}


	public void setE2(int e2) {
		this.e2 = e2;
		this.e2day = space.getDay(space.timeslot[e2]);
	}

	public void setFromEvt(SoftSwapEvents sse) {
		this.e1day = sse.e1day;
		this.e2day = sse.e2day;
		this.matchingFound1 = sse.matchingFound1;
		this.matchingFound2 = sse.matchingFound2;
		this.e1 = sse.e1;
		this.e2 = sse.e2;
		this.idxe1 = sse.idxe1;
		this.idxe2 = sse.idxe2;
		this.delta = sse.delta;
	}

	/**
	 * Eval the variation of the soft cost due to the swap
	 *
	 * @return the value of the variation
	 */
	public int evalSoftCost() {
		int c1 = 0; //cost of moving toward or from a final timeslot
		int e1t = space.timeslot[e1];
		int e2t = space.timeslot[e2];
		if (space.isFinalTimeslot(e1t)) {
			c1 -= space.evtSize[e1];
			c1 += space.evtSize[e2];
		}
		if (space.isFinalTimeslot(e2t)) {
			c1 -= space.evtSize[e2];
			c1 += space.evtSize[e1];
		}
		int c2 = 0; //cost of having a student with a single event in a day.
		if (e1day != e2day) {
			for (int i = 0; i < space.deltaStudentlist[e1][e2].length; i++) {
				int s = space.deltaStudentlist[e1][e2][i];
				if (space.studentPlanning[s][e1day][0] == 2) c2 += 1;
				if (space.studentPlanning[s][e2day][0] == 0) c2 += 1;
				if (space.studentPlanning[s][e1day][0] == 1) c2 -= 1;
				if (space.studentPlanning[s][e2day][0] == 1) c2 -= 1;
			}
			for (int i = 0; i < space.deltaStudentlist[e2][e1].length; i++) {
				int s = space.deltaStudentlist[e2][e1][i];
				if (space.studentPlanning[s][e2day][0] == 2) c2 += 1;
				if (space.studentPlanning[s][e1day][0] == 0) c2 += 1;
				if (space.studentPlanning[s][e2day][0] == 1) c2 -= 1;
				if (space.studentPlanning[s][e1day][0] == 1) c2 -= 1;
			}
		}
		int c3 = 0; //cost of students having more than two courses in a row.
		for (int i = 0; i < space.deltaStudentlist[e1][e2].length; i++) {
			int s = space.deltaStudentlist[e1][e2][i];
			c3 -= space.computeSoftCost2Variation(s, e1t, e1);
			c3 += space.computeSoftCost2Variation(s, e2t, e1);
		}
		for (int i = 0; i < space.deltaStudentlist[e2][e1].length; i++) {
			int s = space.deltaStudentlist[e2][e1][i];
			c3 -= space.computeSoftCost2Variation(s, e2t, e2);
			c3 += space.computeSoftCost2Variation(s, e1t, e2);
		}
		delta = c1 + c2 + c3;
		return delta;
	}

	/**
	 * Is the swap keeping feasability of the current solution
	 *
	 * @return true if the move is feasible
	 */
	public boolean isKeepingFeas() {
		assert (!space.before[e1][e2] && !space.before[e2][e1]);

		int e1t = space.timeslot[e1];
		int e2t = space.timeslot[e2];

		//1-) availability check
		if (space.timeslotFeas[e1][e2t] > 0 || space.timeslotFeas[e2][e1t] > 0)
			return false;

		//2-) compatibility check
		for (int i = 0; i < space.nbept[e1t]; i++) {
			int ei = space.listevts[e1t][i];
			if (space.conflicts[ei][e2] > 0)
				return false;
		}
		for (int i = 0; i < space.nbept[e2t]; i++) {
			int ei = space.listevts[e2t][i];
			if (space.conflicts[ei][e1] > 0)
				return false;
		}

		//3-) precedence checks
		space.timeslot[e1] = e2t;
		space.timeslot[e2] = e1t;
		if (!space.evalFeasPrecedenceFromMove(e1, e2, e1t, e2t) ||
				!space.evalFeasPrecedenceFromMove(e2, e1, e2t, e1t)) {
			space.timeslot[e1] = e1t;
			space.timeslot[e2] = e2t;
			return false;
		}
		space.timeslot[e1] = e1t;
		space.timeslot[e2] = e2t;

		//4-) room availability check by matching
		idxe2 = matching.setTimeSlot(e1t, e1, e2);
		if (!matching.solve()) return false;
		matchingFound1 = matching.getMatching();
		idxe1 = matching.setTimeSlot(e2t, e2, e1);
		if (!matching.solve()) return false;
		matchingFound2 = matching.getMatching();

		return true;
	}


	/*
	 * reallocate the events according to the given matching
	 * except the outsider event
	 */
	public void reallocateByMatching(int timeslo, int evt, int[] matchingFound) {
		for (int i = 0; i < space.nbept[timeslo]; i++) {
			int oevt = space.listevts[timeslo][i];
			space.fastSoftRemoveEvent(oevt);
		}
		for (int i = 0; i < space.nbept[timeslo]; i++) {
			int oevt = space.listevts[timeslo][i];
			int newroom = matchingFound[i];
			space.fastSoftAddEvent(oevt, newroom);
		}
	}

	/**
	 * Perform the swap while taking the matching into account
	 */
	public void performMove() {
		int oldcost = space.softObj; //for assertion
		//System.out.println("perform " + this);
		int e1t = space.timeslot[e1];
		int e2t = space.timeslot[e2];
		assert (e1t != e2t);
		//reallocate the events moved because of the matching
		reallocateByMatching(e1t, e1, matchingFound1);
		reallocateByMatching(e2t, e2, matchingFound2);

		space.softRemoveEvent(e1);
		space.softRemoveEvent(e2);
		space.softAddEvent(e1, e2t, matchingFound2[idxe1]);
		space.softAddEvent(e2, e1t, matchingFound1[idxe2]);

		if (e1day != e2day) {
			//update room, timeslot, matrix, eventcost and global cost
			//neigbhours costs
			space.removeEventSoftCost(e1, e1t);
			space.removeEventSoftCost(e2, e2t);
			space.addEventSoftCost(e1, e2t);
			space.addEventSoftCost(e2, e1t);
		} else {
			space.fastRemoveEventSoftCost(e1, e1t);
			space.fastRemoveEventSoftCost(e2, e2t);
			space.fastAddEventSoftCost(e1, e2t);
			space.fastAddEventSoftCost(e2, e1t);
		}
		assert (oldcost + delta == space.softObj);
	}

	//******************************************************//
	//*************** Print ********************************//	
	//******************************************************//

	public void printday() {
		System.out.println(space.evtSize[e1] + " " + space.evtSize[e2]);
		int e1t = space.timeslot[e1];
		int e2t = space.timeslot[e2];
		for (int i = 0; i < space.deltaStudentlist[e1][e2].length; i++) {
			int s = space.deltaStudentlist[e1][e2][i];
			System.out.println(s + " " + Arrays.toString(space.studentPlanning[s][e1day]) + "-" +
					space.computeSoftCost2Variation(s, e1t, e1) + " +" +
					space.computeSoftCost2Variation(s, e2t, e1));
		}
		System.out.println("------");
		for (int i = 0; i < space.deltaStudentlist[e2][e1].length; i++) {
			int s = space.deltaStudentlist[e2][e1][i];
			System.out.println(s + " " + Arrays.toString(space.studentPlanning[s][e1day]) + "-" +
					space.computeSoftCost2Variation(s, e2t, e2) + " +" +
					space.computeSoftCost2Variation(s, e1t, e2));
		}
	}


}
