package localsearch.moves.complete.softmoves;

import localsearch.moves.complete.SwapTimeslots;
import localsearch.SolutionSpace;

import java.util.BitSet;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: 26 oct. 2007
 * Time: 16:59:21
 * To change this template use File | Settings | File Templates.
 */
public class SoftSwapTimeslots extends SwapTimeslots implements SoftMove {

	protected int d1, d2;

	public SoftSwapTimeslots(SolutionSpace space) {
		super(space);
	}

	public void setT1(int t1) {
		this.t1 = t1;
		this.timet1 = space.matrix[t1];
		this.d1 = space.getDay(t1);
	}

	public void setT2(int t2) {
		this.t2 = t2;
		this.timet2 = space.matrix[t2];
		this.d2 = space.getDay(t2);
	}


	public int evalSoftCost() {
		int c1 = 0;  //cost1
		if (space.isFinalTimeslot(t1) && !space.isFinalTimeslot(t2)) {
			for (int i = 0; i < space.nbept[t1]; i++)
				c1 -= space.evtSize[space.listevts[t1][i]];
			for (int i = 0; i < space.nbept[t2]; i++)
				c1 += space.evtSize[space.listevts[t2][i]];
		} else if (space.isFinalTimeslot(t2) && !space.isFinalTimeslot(t1)) {
			for (int i = 0; i < space.nbept[t2]; i++)
				c1 -= space.evtSize[space.listevts[t2][i]];
			for (int i = 0; i < space.nbept[t1]; i++)
				c1 += space.evtSize[space.listevts[t1][i]];
		}
		int c2 = 0;  //cost2		
		BitSet b1 = space.studentsAttendingTimeslot(t1);
		BitSet b2 = space.studentsAttendingTimeslot(t2);
		BitSet deltat1Butt2 = space.studentsAttendingB1butB2(b1, b2);
		BitSet deltat2Butt1 = space.studentsAttendingB1butB2(b2, b1);

		if (d1 != d2) {
			for (int s = deltat1Butt2.nextSetBit(0); s >= 0; s = deltat1Butt2.nextSetBit(s + 1)) {
				if (space.studentPlanning[s][d1][0] == 2) c2 += 1;
				if (space.studentPlanning[s][d2][0] == 0) c2 += 1;
				if (space.studentPlanning[s][d1][0] == 1) c2 -= 1;
				if (space.studentPlanning[s][d2][0] == 1) c2 -= 1;
			}
			for (int s = deltat2Butt1.nextSetBit(0); s >= 0; s = deltat2Butt1.nextSetBit(s + 1)) {
				if (space.studentPlanning[s][d2][0] == 2) c2 += 1;
				if (space.studentPlanning[s][d1][0] == 0) c2 += 1;
				if (space.studentPlanning[s][d2][0] == 1) c2 -= 1;
				if (space.studentPlanning[s][d1][0] == 1) c2 -= 1;
			}
		}
		int c3 = 0;
		for (int s = deltat1Butt2.nextSetBit(0); s >= 0; s = deltat1Butt2.nextSetBit(s + 1)) {
			int e1 = space.eventOfStudentInTimeslot(s, t1);
			c3 -= space.computeSoftCost2Variation(s, t1, e1);
			c3 += space.computeSoftCost2Variation(s, t2, e1);
		}
		for (int s = deltat2Butt1.nextSetBit(0); s >= 0; s = deltat2Butt1.nextSetBit(s + 1)) {
			int e2 = space.eventOfStudentInTimeslot(s, t2);
			c3 -= space.computeSoftCost2Variation(s, t2, e2);
			c3 += space.computeSoftCost2Variation(s, t1, e2);
		}
		delta = c1 + c2 + c3;
		return delta;
	}

	public boolean isKeepingFeas() {
		//check availability of timeslots
		for (int i = 0; i < space.nbept[t1]; i++) {
			int e1i = space.listevts[t1][i];
			if (space.timeslotFeas[e1i][t2] > 0) return false;
		}
		for (int i = 0; i < space.nbept[t2]; i++) {
			int e2i = space.listevts[t2][i];
			if (space.timeslotFeas[e2i][t1] > 0) return false;
		}

		//check precedences
		for (int i = 0; i < space.nbept[t1]; i++) {
			int e1i = space.listevts[t1][i];
			if (!space.evalFeasPrecCost(e1i, t1, t2)) return false;
		}
		for (int i = 0; i < space.nbept[t2]; i++) {
			int e2i = space.listevts[t2][i];
			if (!space.evalFeasPrecCost(e2i, t1, t2)) return false;
		}
		return true;
	}


	public void performMove() {
		int oldcost = space.softObj; //for assertion
		space.fastSwapTimeSlot(t1,t2);
		if (d1 != d2) {
			removeEventSoftCost(t1,t2);
			removeEventSoftCost(t2,t1);
			addEventSoftCost(t1);
			addEventSoftCost(t2);
		} else {
			fastRemoveEventSoftCost(t2,t1);
			fastRemoveEventSoftCost(t1,t2);
			fastAddEventSoftCost(t1);
			fastAddEventSoftCost(t2);			
		}
		assert (space.assertAllSoftCosts());
		assert (oldcost + delta == space.softObj);
	}

	public void removeEventSoftCost(int t1, int t2) {
		for (int i = 0; i < space.nbept[t1]; i++) {
			int e1 = space.listevts[t1][i];
			space.removeEventSoftCost(e1, t2);
		}
	}

	public void fastRemoveEventSoftCost(int t1, int t2) {
		for (int i = 0; i < space.nbept[t1]; i++) {
			int e1 = space.listevts[t1][i];
			space.fastRemoveEventSoftCost(e1, t2);
		}
	}

	public void addEventSoftCost(int tadd) {
		for (int i = 0; i < space.nbept[tadd]; i++) {
			int e1 = space.listevts[tadd][i];
			space.addEventSoftCost(e1, tadd);
		}
	}

	public void fastAddEventSoftCost(int tadd) {
		for (int i = 0; i < space.nbept[tadd]; i++) {
			int e1 = space.listevts[tadd][i];
			space.fastAddEventSoftCost(e1, tadd);
		}
	}

	public String toString() {
		return "swapping timeslots " + t1 + " and " + t2 + " for " + delta;
	}
}
