package localsearch.moves.complete.softmoves;

import localsearch.moves.complete.TranslateEvent;
import localsearch.moves.complete.softmoves.SoftMove;
import localsearch.SolutionSpace;
import localsearch.Place;
import localsearch.graphalgos.MatchingInterface;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: 20 oct. 2007
 * Time: 11:38:07
 * To change this template use File | Settings | File Templates.
 */
public class SoftTranslateEvent extends TranslateEvent implements SoftMove {

	protected int oldday, newday;

	protected MatchingInterface matching;
	protected int[] matchingFound;

    public SoftTranslateEvent() {}

    public SoftTranslateEvent(SolutionSpace space) {
		super(space);
		matching = new MatchingInterface(space);
	}

	public void setEvt(int evt) {
		super.setEvt(evt);
		oldday = space.getDay(oldtimeslot);
	}

	public void setFromEvt(SoftTranslateEvent tre) {
		this.timeslot = tre.timeslot;
		this.room = tre.room;
		this.oldday = tre.oldday;
		this.newday = tre.newday;
		this.matchingFound = tre.matchingFound;
		this.delta = tre.delta;
	}

	public void setTimeslot(int t) {
		timeslot = t;
		newday = space.getDay(timeslot);
	}

	public void setRoom(int r) {
		room = r;
	}

	public void setFreeslot(Place freeslot) {
		super.setFreeslot(freeslot);
		newday = space.getDay(timeslot);
	}

	/**
	 * Eval the variation of the soft cost due to a translation
	 * @return
	 */
	public int evalSoftCost() {
		//System.out.println("Eval " + this);
		int c1 = 0; //cost of moving toward or from a final timeslot
		if (space.isFinalTimeslot(oldtimeslot)) {
			c1 -= space.evtSize[evt];
		}
		if (space.isFinalTimeslot(timeslot)) {
			c1 += space.evtSize[evt];
		}
		int c2 = 0; //cost of having a student with a single event in a day.
		if (oldday != newday) {
			//c2 = -space.singleEvt[evt];
			for (int i = 0; i < space.studentlist[evt].length; i++) {
				int s = space.studentlist[evt][i];
				if (space.studentPlanning[s][oldday][0] == 2)
					c2 += 1;
				if (space.studentPlanning[s][newday][0] == 0)
					c2 += 1;
				if (space.studentPlanning[s][oldday][0] == 1)
					c2 -= 1;
				if (space.studentPlanning[s][newday][0] == 1)
					c2 -= 1;
			}
		}
		int c3 = 0; //cost of students having more than two courses in a row.
		for (int i = 0; i < space.studentlist[evt].length; i++) {
			int s = space.studentlist[evt][i];
			c3 -= space.computeSoftCost2Variation(s, oldtimeslot, evt);
			c3 += space.computeSoftCost2Variation(s, timeslot, evt);
		}
		delta = c1 + c2 + c3;
		return delta;
	}

	/**
	 * check the conflict with the neigbourhs, precedences
	 * and call the matching to see if there is a way to reallocate
	 * the other events.
	 * WARNING: timeslot feasability is not checked and assumed by search
	 */
	public boolean isKeepingFeas() {
		assert(space.timeslotFeas[evt][timeslot] == 0);
		if (space.nbept[timeslot] == space.R) //quick test of room availability
				return false;
		for (int i = 0; i < space.nbept[timeslot]; i++) {
			int ei = space.listevts[timeslot][i];
			if (space.conflicts[ei][evt] > 0) return false;
		}
		if (!space.evalFeasPrecedenceFromMove(evt, -1, oldtimeslot, timeslot))
			return false;

		matching.setTimeSlot(timeslot,evt);
		if (!matching.solve()) {
			return false;
		} 
		matchingFound = matching.getMatching();
		room = matchingFound[matchingFound.length - 1];
		return true;
	}

	/**
	 * Perform the translation while maintaing the changes
	 * over the events in the timeslot that might need to be moved
	 * to satisfy the machting including the outsider event
	 */
	public void performMove() {
		int oldcost = space.softObj; //for assertion
		//reallocate the events moved because of the matching
		for (int i = 0; i < space.nbept[timeslot]; i++) {
			int oevt = space.listevts[timeslot][i];
			space.fastSoftRemoveEvent(oevt);
		}
		for (int i = 0; i < space.nbept[timeslot]; i++) {
			int oevt = space.listevts[timeslot][i];
			int newroom = matchingFound[i];
			space.fastSoftAddEvent(oevt,newroom);
		}
		//moving the outsider event
		int e1t = space.timeslot[evt];
		assert (e1t != timeslot);
		space.softRemoveEvent(evt);
		space.softAddEvent(evt, timeslot, room);
		if (oldday != newday) {
			space.removeEventSoftCost(evt, e1t);
			space.addEventSoftCost(evt, timeslot);
		} else {
			space.fastRemoveEventSoftCost(evt, e1t);
			space.fastAddEventSoftCost(evt, timeslot);
		}

		//space.assertAllSoftCosts();
		assert (oldcost + delta == space.softObj);
	}
}
