package localsearch.solvers.coloring;

import instance.TimetablingInstance;

import java.util.Arrays;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: Sep 29, 2008
 * Time: 6:53:57 PM
 * To change this template use File | Settings | File Templates.
 */
public class ColoringSolutionSpace extends HardColoringSolutionSpace {


    public ColoringSolutionSpace(TimetablingInstance ins, int seed) {
        super(ins, seed);
    }

    //**********************************************************//
    //******** Operation on Events for Hard data ***************//
    //**********************************************************//

    /**
     * Remove event evt from its current location while
     * maintaining only the hard data structures needed by softmoves
     *
     * @param evt the event removed
     */
    public void softRemoveEvent(int evt) {
        int e1t = timeslot[evt];
        timeslot[evt] = -1;
        int pevt = -1;
        for (int i = 0; i < R; i++) { //todo : improve with a real list data structure
            if (listevts[e1t][i] == evt) {
                pevt = i;
                break;
            }
        }
        maintainListEvt(evt, e1t, pevt);
        nbept[e1t]--;
    }

    /**
     * Place event evt in the EMPTY place corresponding to the given
     * timeslot and room while maintaining only the hard data structures needed
     * by soft moves.
     *
     * @param evt     the event to add
     * @param timeslo the timeslot where we want to put the event
     * @param room    the room where the event has to go
     */
    public void softAddEvent(int evt, int timeslo) {
        timeslot[evt] = timeslo;
        listevts[timeslo][nbept[timeslo]] = evt;
        nbept[timeslo]++;
    }


    //**********************************************************//
    //******** Api for swapping timeslots **********************//
    //**********************************************************//


    public void fastSwapTimeSlot(int t1, int t2) {
        int[] levtst1 = listevts[t1];
        int[] levtst2 = listevts[t2];
        for (int i = 0; i < nbept[t1]; i++) {
            int e = levtst1[i];
            timeslot[e] = t2;
        }
        for (int i = 0; i < nbept[t2]; i++) {
            int e = levtst2[i];
            timeslot[e] = t1;
        }
        listevts[t1] = levtst2;
        listevts[t2] = levtst1;
        int nbept1 = nbept[t1];
        nbept[t1] = nbept[t2];       
        nbept[t2] = nbept1;
    }


//**********************************************************//
//******** assert data structures **************************//
//**********************************************************//

    public boolean assertFeasible() {
        int[] cardEvtsPerTimeslot = new int[T];
        for (int e = 0; e < E; e++) {
            int et = timeslot[e];
            cardEvtsPerTimeslot[et]++;
            assert (et != -1);
            assert (timeslotFeas[e][et] == 0);
            for (int r = 0; r < nbept[et]; r++) {
                int e2 = listevts[et][r];
                if (e2 != -1) {
                    assert (conflicts[e][e2] == 0);
                }
            }
            for (int i = 0; i < greaterthan[e].length; i++) {
                assert (timeslot[greaterthan[e][i]] < et);
            }
        }
        for (int i = 0; i < T; i++) {
            assert (nbept[i] == cardEvtsPerTimeslot[i]);
        }
        for (int i = 0; i < E; i++) {
            int eit = timeslot[i];
            boolean isInList = false;
            for (int j = 0; j < nbept[eit] && !isInList; j++) {
                if (listevts[eit][j] == i) isInList = true;
            }
            if (!isInList) System.out.println("evt " + i + " in " + eit);
            assert (isInList);
        }
        return true;
    }


}
