package localsearch.solvers.coloring;

import localsearch.solvers.complete.HardSolver;
import localsearch.solvers.complete.AbstractSolver;
import localsearch.solvers.coloring.HardColoringSolutionSpace;
import localsearch.moves.coloring.ColoringTranslation;
import localsearch.moves.coloring.ColoringSwapEvents;
import localsearch.moves.coloring.ColoringSwapTimeslots;
import localsearch.moves.coloring.ColoringVHungarianMove;
import localsearch.moves.complete.*;
import localsearch.SolutionSpace;
import localsearch.Tracer;
import localsearch.metah.TrueMH;
import localsearch.graphalgos.matchings.Matching;

import java.util.Collections;
import java.util.LinkedList;

import instance.TimetablingInstance;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: Sep 28, 2008
 * Time: 11:41:35 AM
 * To change this template use File | Settings | File Templates.
 */
public class HardColoringSolver extends HardSolver {


    public HardColoringSolver(String ins, int seed) {
        super(ins, seed);
    }


    public HardColoringSolver(TimetablingInstance ins, int seed) {
        super(ins, seed);
    }


    public void buildHardSolver() {
        evts = new Integer[space.E];
        for (int i = 0; i < space.E; i++) {
            evts[i] = i;
        }
    }

    public void buildAbstractSolver(TimetablingInstance ins, int seed) {
        this.starttime = (int) System.currentTimeMillis();
        this.seed  = seed;
        this.ins = ins;
        this.space = new ColoringSolutionSpace(ins, seed);
        this.moves = new LinkedList<Integer>();
        this.bestsolution = new int[space.T][space.R];
        this.mh = new TrueMH(this);
        this.swape = new ColoringSwapEvents((HardColoringSolutionSpace) space);
        this.transe = new ColoringTranslation((HardColoringSolutionSpace) space);
        this.swapt = new ColoringSwapTimeslots(space);
        this.vhunge = new ColoringVHungarianMove(space);
        this.ldTranslation = new ColoringTranslation((HardColoringSolutionSpace) space);
        this.ldSwap = new ColoringSwapEvents((HardColoringSolutionSpace) space);
    }


    public void storeNewImprovingSol() {
        for (int i = 0; i < space.T; i++) {
            for (int j = 0; j < space.R; j++) {
                bestsolution[i][j] = space.listevts[i][j];
            }
        }
        bestCost = getCost();
    }

    /**
     * Define the neighbourhood used at the current iteration
     * it is controlled by a probability p for each kind of move
     */
    public void defineNeighbourhood() {
        if (dynamicNeihbourhood) {
            moves.clear();
            moves.add(TRANSLATE_EVENTS);
            moves.add(SWAP_EVENTS);
            if (space.rand.nextDouble() <= pst) {
                moves.add(SWAP_TIMESLOTS);
            }
            if (space.rand.nextDouble() <= pvh) {
                moves.add(VERT_HUNGARIAN_MOVE);
            }
        }
        Collections.shuffle(moves, space.rand);
    }

    public void matchingTimeSlots() {
        //do nothing when asked to solve the matchings
    }



    public void addNeighbourhood(int m) {
        if (m == TRANSLATE_EVENTS ||
            m == SWAP_EVENTS ||
            m == SWAP_TIMESLOTS ||
            m == VERT_HUNGARIAN_MOVE) {
           moves.addFirst(m);
        } 
    }

    public void setDefaultNeighbourhood() {
        addNeighbourhood(AbstractSolver.SWAP_TIMESLOTS);
        addNeighbourhood(AbstractSolver.SWAP_EVENTS);
        addNeighbourhood(AbstractSolver.TRANSLATE_EVENTS);
    }    
    /**
     * Intialization of the assignment before entering
     * the local search
     */
    public void start() {
        if (turnTracerOn)
            this.tracy = new Tracer(name + "coloring",false);        
        if (firstRun) {
            space.initSolutionSpace();
        }
        iter = 0;
        bestCost = Integer.MAX_VALUE;
        if (!dynamicNeihbourhood && moves.isEmpty()) {
            throw new Error("no neighbourhood defined !");
        }
        //this.tracy = new Tracer(name,false);
        reinitLDMove();
        if (chocoinit) {
            space.initFromTimeslotSol(eventTime);
        } else {
            space.randomInit();
        }
        if (firstRun) {
            mh.initMhParameters();
            firstRun = false;
        }
        assert (space.assertAllHardCosts());
    }

    //****************************************************************//
    //*************** Translating structure **************************//
    //****************************************************************//

    //remember the last event translated to start the scanning
    //from the next one
    protected int tevt1;

    /**
     * @return true if an improving translation has been found
     */
    public boolean exploreTranslationDeterministically() {
        if (exploreTranslationDeterministically(tevt1 + 1, space.E)) {
            return true;
        } else return exploreTranslationDeterministically(0, tevt1 + 1);
    }


    public boolean exploreTranslationDeterministically(int deb, int fin) {
        for (int i = deb; i < fin; i++) {
            int e = evts[i];
            transe.setEvt(e);
            for (int t = 0; t < space.T; t++) {
                if (t != space.timeslot[e] && space.nbept[t] < space.R) {
                    ((ColoringTranslation) transe).setTimeslot(t);
                    int ctrans = transe.evalCostMove();
                    boolean accept = mh.accept(transe, ctrans);
                    if (ctrans <= 0 && accept) {  //accept side walk steps
                        tevt1 = e;
                        return true;
                    } else if (accept) {
                        storeLDMove(transe, ctrans);
                    }
                }
            }
        }
        return false;
    }

    //****************************************************************//
    //*************** Swapping events structure **********************//
    //****************************************************************//
    //remember the last event swapped to start the scanning
    //from the next one
    protected int sevt1;

    /**
     * @return true if an improving swap has been found
     */
    public boolean exploreSwapsDeterministically() {
        if (exploreSwapsDeterministically(sevt1 + 1, space.E)) {
            return true;
        } else return exploreSwapsDeterministically(0, sevt1 + 1);
    }

    public boolean exploreSwapsDeterministically(int deb, int fin) {
        for (int i = deb; i < fin; i++) {
            int e1 = evts[i];
            swape.setE1(e1);
            for (int j = i + 1; j < space.E; j++) {
                int e2 = evts[j];
                if (space.timeslot[e1] != space.timeslot[e2]) {
                    swape.setE2(e2);
                    int cswap = swape.evalCostMove();
                    boolean accept = mh.accept(swape, cswap);
                    if (cswap <= 0 && accept) { //accept side walk steps
                        sevt1 = e1;
                        return true;
                    } else if (accept) {
                        storeLDMove(swape, cswap);
                    }
                }
            }
        }
        return false;
    }

    

  

}
