package localsearch.solvers.coloring;

import localsearch.solvers.complete.SoftSolver;
import localsearch.solvers.complete.HardSolver;
import localsearch.solvers.complete.SoftCostProjection;
import localsearch.metah.TrueMH;
import localsearch.metah.SimulatedAnnealing;
import localsearch.moves.complete.softmoves.SoftSwapEvents;
import localsearch.moves.complete.softmoves.SoftSwapTimeslots;
import localsearch.moves.complete.softmoves.SoftTranslateEvent;
import localsearch.moves.complete.softmoves.SoftCPMove;
import localsearch.moves.coloring.softmoves.ColoringSoftSwapEvents;
import localsearch.moves.coloring.softmoves.ColoringSoftTranslation;
import localsearch.SolutionSpace;

import java.util.LinkedList;
import java.util.Collections;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: Sep 30, 2008
 * Time: 10:58:52 AM
 * To change this template use File | Settings | File Templates.
 */
public class SoftColoringSolver extends SoftSolver {

    public SoftColoringSolver(HardColoringSolver solver) {
        hsolve = solver;
        space = solver.space;
        bestsolution = new int[space.T][space.R];
        for (int i = 0; i < space.T; i++) {
            System.arraycopy(solver.bestsolution[i], 0, bestsolution[i], 0, space.R);
        }
        ins = solver.ins;
        name = solver.name;
        timelimit = solver.timelimit;
        seed = solver.seed;
        mh = new TrueMH(this);
        moves = new LinkedList<Integer>();
        swape = new ColoringSoftSwapEvents(space);
        //swapt = new SoftSwapTimeslots(space);
        transe = new ColoringSoftTranslation(space);
        ldTranslation = new ColoringSoftTranslation(space);
        ldSwap = new ColoringSoftSwapEvents(space);
    }


//    public HardSolver hardS = null;
//
//    public void setHardSolverToRegisterFeasibleSol(HardSolver hs) {
//        hardS = hs;
//        hardS.space.initSolutionSpace();
//    }
//
//
    public void storeNewImprovingSol() {
        for (int i = 0; i < space.T; i++) {
            for (int j = 0; j < space.R; j++) {
                bestsolution[i][j] = space.listevts[i][j];
            }
        }
        bestCost = getCost();
    }
//        //try to transform this solution into a feasible one
//        //by solving T matchings (because this does not affect the soft cost)
//        if (hardS != null) {
//           //System.out.print("try to make it feasible");
//           hardS.space.initFromTimeslotSol(space.timeslot);
//           hardS.matchingTimeSlots();
//           if (hardS.getCost() <= hardS.getBestCost())
//               hardS.storeNewImprovingSol();
//
//           //System.out.println(" " + hardS.getCost() + " " + hardS.getBestCost());
//        }
//    }
    
    /**
	 * Define the neighbourhood used at the current iteration
	 * it is controlled by a probability p for each kind of move
	 */
	public void defineNeighbourhood() {
		if (dynamicNeihbourhood) {
			moves.clear();
			moves.add(TRANSLATE_EVENTS);
			if (space.rand.nextDouble() <= phh) {
				moves.add(SWAP_EVENTS);
			}
//			if (space.rand.nextDouble() <= pst) {
//				moves.add(SWAP_TIMESLOTS);
//			}
		} 
		Collections.shuffle(moves, space.rand);
	}

    public void start() {
        //tracy = new Tracer(name,true);
        starttime = hsolve.starttime;
        if (firstRun) {
            sav = new HardColoringSolSAV(space);
            sav.save(space);
        }
        if (finalIntensification) {
            restoreBestSolution();
            mh = new TrueMH(this);
        }
        space.baseSoftInit();
        space.softInit();
        space.computeSoftCostFromScratch();
        iter = 0;
        if (space.softObj < bestCost)
            bestCost = space.softObj;
        if (trace)
            System.out.println("Start with soft cost " + bestCost);
        reinitLDMove();
        if (firstRun) {
            mh.initMhParameters();
            firstRun = false;
        }
        assert (space.assertAllSoftCosts());
    }
    

    /**
	 * A structure to sav a feasible solution of the coloring
     * and restore it easily
	 */
	public class HardColoringSolSAV extends HardSolSAV {
		public int[] nbept;
		public int[] timeslot;
		public int[][] listevts;

		public HardColoringSolSAV(SolutionSpace space) {
			listevts = new int[space.T][space.R];
			timeslot = new int[space.E];
			nbept = new int[space.T];

		}

		public void save(SolutionSpace space) {
			System.arraycopy(space.timeslot, 0, timeslot, 0, space.E);
			System.arraycopy(space.nbept, 0, nbept, 0, space.T);
			for (int i = 0; i < space.T; i++) {
				System.arraycopy(space.listevts[i], 0, listevts[i], 0, space.R);
			}
		}

		public void restore(SolutionSpace space) {
			System.arraycopy(timeslot, 0, space.timeslot, 0, space.E);
			System.arraycopy(nbept, 0, space.nbept, 0, space.T);
			for (int i = 0; i < space.T; i++) {
				System.arraycopy(listevts[i], 0, space.listevts[i], 0, space.R);
			}
        }

	}

}
