package localsearch.solvers.coloring;

import localsearch.solvers.LSSolver;
import localsearch.solvers.complete.HardSolver;
import localsearch.solvers.complete.AbstractSolver;
import localsearch.solvers.complete.SearchLoop;
import localsearch.solvers.complete.SoftSolver;
import localsearch.metah.SimulatedAnnealing;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: Sep 30, 2008
 * Time: 11:35:35 AM
 * To change this template use File | Settings | File Templates.
 */
public class TwoStepColoringSolver implements LSSolver {

    public HardColoringSolver hcs;
    public SoftColoringSolver scs;
    public HardSolver hs;
    public SoftSolver ss;

    public TwoStepColoringSolver(String ins, int seed) {
        hcs = new HardColoringSolver(ins, seed);
        scs = new SoftColoringSolver(hcs);
        hs = new HardSolver(hcs.ins, seed);
        hs.setStartTime(hcs.starttime);
        ss = new SoftSolver(hs);
        hs.name = ins;
        ss.name = ins;
    }


    public void addNeighbourhood(int move) {
        throw new Error("not yet available on TwoStepColoringSolver");
    }

    public void setDefaultNeighnourhood() {
        setGreedy();
        hcs.setDefaultNeighbourhood();
        hs.setDefaultNeighbourhood();
        scs.setDefaultNeighbourhood();
        ss.setDefaultNeighbourhood();
    }

    public void dynamicNeighbourhood() {
        hcs.dynamicNeighbourhood();
        hs.dynamicNeighbourhood();
    }

    public int getStartTime() {
        return hcs.starttime;
    }

    public void setGreedy() {
        hcs.setGreedy();
        hs.setGreedy();
    }

    public void setTimelimit(int timel) {
        hcs.setTimelimit(timel);
        scs.setTimelimit(timel - (int) (0.15f * (float) timel));
        hs.setTimelimit(timel);
        ss.setTimelimit(timel);
    }

    public void setTabuMetaHeuristic(int tsize) {
        hcs.setTabuMetaHeuristic(tsize);
        hs.setTabuMetaHeuristic(tsize);
    }

    public void setSAMetaHeuristic(double ratio) {
        scs.setSAMetaHeuristic(ratio);
        ss.setSAMetaHeuristic(ratio);
    }

    public int getTotalTime() {
        return hs.getTotalTime();
    }

    public int getNbIter() {
        return hcs.getNbIter() + hs.getNbIter() + scs.getNbIter() + ss.getNbIter();
    }

    public int getBestCost() {
        return ss.getBestCost();
    }

    public boolean isFeasible() {
        return hs.getBestCost() == 0;
    }

    public void solve() {

        //COLORING PHASE
        hcs.solve();
        if (AbstractSolver.trace)
            System.out.println(" End of Hard Coloring Solver (time elapsed : " + ((int) System.currentTimeMillis() - hcs.starttime) + ")");
//        scs.setHardSolverToRegisterFeasibleSol(hs);
        scs.solve();

        if (AbstractSolver.trace)
            System.out.println(" End of Soft Coloring Solver (time elapsed : " + ((int) System.currentTimeMillis() - hcs.starttime) + ")");

        //MATCHING + COLORING PHASE
        int[] timeslot = new int[scs.space.E];
        for (int i = 0; i < scs.space.T; i++) {
            for (int j = 0; j < scs.space.R; j++) {
                if (scs.bestsolution[i][j] != -1) {
                    timeslot[scs.bestsolution[i][j]] = i;
                }
            }
        }
//        if (hs.getBestCost() != 0) {
        hs.setGreedyInit(timeslot);
        hs.solve();
//        } else {
//        hs.space.initFromGivenSol(hs.bestsolution);
//        hs.solve();
//        if (AbstractSolver.trace)
//            System.out.println(" Solution already stored of hard cost " + hs.getCost());
//        }
        if (AbstractSolver.trace)
            System.out.println(" End of Hard Solver (time elapsed : " + ((int) System.currentTimeMillis() - hcs.starttime) + ")");
        if (hs.getBestCost() == 0) {
            ss.solve();
            if (AbstractSolver.trace)
                System.out.println(" End of Soft Solver (time elapsed : " + ((int) System.currentTimeMillis() - hcs.starttime) + ")");
            ss.outputSolution(true);
        }
    }

    public void outputSolution(boolean soft) {

    }

    public static void main(String[] arg) {
        String name = "data/Track2/comp-2007-2-10.tim";
        long time = System.currentTimeMillis();
        int timelimit = 420000;
        TwoStepColoringSolver s = new TwoStepColoringSolver(name, 2);
        AbstractSolver.trace = true;
        AbstractSolver.debug = false;
        SearchLoop.searchLoopTrace = true;
        SimulatedAnnealing.hardPrediction = false;
        s.setTimelimit(timelimit);
        s.setDefaultNeighnourhood();
        s.setTabuMetaHeuristic(10);
        s.setSAMetaHeuristic(1);
        s.solve();
        System.out.println("" + s.getBestCost() + " " + (System.currentTimeMillis() - time) + " " + s.getTotalTime());

    }

//ss.setSAMetaHeuristic(1 - ratioIntens);
//ss.setDefaultNeighbourhood();

}
