package localsearch.solvers.coloring;

import localsearch.solvers.LSSolver;
import localsearch.solvers.complete.HardSolver;
import localsearch.solvers.complete.AbstractSolver;
import localsearch.solvers.complete.SoftSolver;
import localsearch.Tracer;

/**
 *
 * Solver that split the resolution of the feasibility by solving
 * first the coloring relaxation the problem and then repairing the rooms
 * with the complete solver
 */
public class TwoStepHardSolver implements LSSolver {

    public HardColoringSolver hcs;
    public HardSolver hs;

    public TwoStepHardSolver(String ins, int seed) {
        hcs = new HardColoringSolver(ins,seed);
        hs = new HardSolver(hcs.ins,seed);
        hs.setStartTime(hcs.starttime);
        hs.name = ins;
    }

    public void setTabuMetaHeuristic(int tsize) {
		hcs.setTabuMetaHeuristic(tsize);
	    hs.setTabuMetaHeuristic(tsize);
    }

    public void setSAMetaHeuristic(double ratio) {
        hcs.setSAMetaHeuristic(ratio);
        hs.setSAMetaHeuristic(ratio);
    }

     public void setDefaultNeighbourhood() {
        hcs.setDefaultNeighbourhood();
        hs.setDefaultNeighbourhood();
        setGreedy();
    }

    public int getTotalTime() {
        return hs.getTotalTime();
    }

    public int getStartTime() {
        return hcs.getStartTime();
    }

    public int getNbIter() {
        return hs.getNbIter() + hcs.getNbIter();
    }

    public int getBestCost() {
        return hs.getBestCost();
    }

    public void dynamicNeighbourhood() {
        hs.dynamicNeighbourhood();
        hcs.dynamicNeighbourhood();
    }

    public void addNeighbourhood(int m) {
        hs.addNeighbourhood(m);
        hcs.addNeighbourhood(m);
    }

    public void setGreedy() {
        hs.setGreedy();
        hcs.setGreedy();
    }

    public void setTimelimit(int timel) {
        hs.setTimelimit(timel);
        hcs.setTimelimit(timel);
    }

    public void outputSolution(boolean soft) {
        hs.outputSolution(soft);
    }

    /**
     * Solve in two steps by calling first the coloring solver
     * and then the complete solver.
     */
    public void solve() {
        hcs.solve();
        hs.setGreedyInit(hcs.space.timeslot);
        if (AbstractSolver.trace)
            System.out.println(" End of Coloring Solver, apply integral solver (time elapsed : " + ((int) System.currentTimeMillis() - hcs.starttime) + ")");
        if (AbstractSolver.turnTracerOn)
            hs.tracy = new Tracer(hs.name,false,hcs.iter);

        hs.solve();
//        if (hcs.getBestCost() == 0)  //to solve the coloring only
//           hs.setBestCost(0);
//        hs.endtime = (int) System.currentTimeMillis();
    }

    public static void main(String[] arg) {
        String name = "data/Track2/comp-2007-2-1.tim";
        long time = System.currentTimeMillis();
        int timelimit = 420000;
        TwoStepHardSolver s = new TwoStepHardSolver(name, 0);
        AbstractSolver.trace = true;
        AbstractSolver.debug = false;
        s.dynamicNeighbourhood();
        s.setGreedy();
        s.setTimelimit(timelimit);
        s.solve();
        System.out.println("" + s.getBestCost() + " " + (System.currentTimeMillis() - time) + " "+ s.getTotalTime());

    }
}
