package localsearch.solvers.complete;

import instance.TimetablingInstance;
import localsearch.Place;
import localsearch.Tracer;
import localsearch.greedy.HungarianGreedy;
import localsearch.moves.complete.SwapEvents;
import localsearch.moves.complete.TranslateEvent;
import localsearch.moves.Move;

import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: 20 aot 2007
 * Time: 11:24:52
 * To change this template use File | Settings | File Templates.
 */
public class HardSolver extends AbstractSolver {

    public HungarianGreedy greedy;
    public static int threshold = 1000;

    //default constructor
    public HardSolver() {

    }

    public HardSolver(String ins, int seed) {
        super(ins, seed);
        buildHardSolver();
    }

    public HardSolver(TimetablingInstance ins, int seed) {
        super(ins, seed);
        buildHardSolver();
    }

    public void buildHardSolver() {
        ldSwap = new SwapEvents(space);
        ldTranslation = new TranslateEvent(space);
        evts = new Integer[space.E];
        for (int i = 0; i < space.E; i++) {
            evts[i] = i;
        }
    }

    /**
     * Intialization of the assignment before entering
     * the local search
     */
    public void start() {
        super.start();
        if (turnTracerOn && tracy == null)
            this.tracy = new Tracer(name, false);
        reinitLDMove();
        if (chocoinit) {
            space.initFromTimeslotSol(eventTime);
            matchingTimeSlots();
        } else {
            space.randomInit();
            if (initmatching)
                matchingTimeSlots();
        }
        if (firstRun) {
            mh.initMhParameters();
            firstRun = false;
        }
        assert (space.assertAllHardCosts());
    }

    /**
     * Main method to call the local search process
     */
    public void solve() {
        start();
        boolean improvingMove = true;
        while (!doWeStop()) {
            defineNeighbourhood();
            improvingMove = false;
            Move move = null;
            for (Iterator<Integer> it = moves.iterator(); it.hasNext();) {
                Integer m = it.next();
                switch (m) {
                    case SWAP_EVENTS:
                        move = swape;
                        improvingMove = exploreSwapsDeterministically();
                        break;
                    case TRANSLATE_EVENTS:
                        move = transe;
                        improvingMove = exploreTranslationDeterministically();
                        break;
                    case TRANSLATE_EVENTS_MATCHING:
                        move = transem;
                        improvingMove = exploreTranslationMatchingDeterministically();
                        break;
                    case SWAP_TIMESLOTS:
                        move = swapt;
                        improvingMove = exploreTSwapsDeterministically();
                        break;
                    case HORIZ_HUNGARIAN_MOVE:
                        move = hhunge;
                        improvingMove = exploreHHungarianDeterministically();
                        break;
                    case VERT_HUNGARIAN_MOVE:
                        move = vhunge;
                        improvingMove = exploreHungarianStochastically();
                        break;
                    case CP_MOVE:
                        move = cpmove;
                        improvingMove = exploreLNS();
                        break;
                    default:
                        throw new Error(m + " is an unknown kind of move");
                }
                if (improvingMove) break;
            }
            if (improvingMove) {
                move.performMove();
                mh.informOfMove(move);
                if (isCostImproved()) {
                    nonImprovIter = 0;
                    storeNewImprovingSol();
                    if (trace)
                        System.out.println(iter + " new best sol of cost " + bestCost + " after move " + move + " " + getCompInfo() + " " + ((int) System.currentTimeMillis() - starttime));
                } else {
                    if (debug)
                        System.out.println(iter + " move " + move + " to " + getCost() + "[" + bestCost + "]" + " " + getCompInfo());
                    nonImprovIter++;
                }
            } else {
                int oldobj = getCost();
                if (!performLDMove()) {
                    diversificationStep();
                }
                nonImprovIter++;
                if (debug)
                    System.out.println(iter + " less degrading move from " + oldobj + " towards " + getCost() + " by " + getLDMove() + " " + getCompInfo());
            }
            assert (assertCostComputation());
            reinitLDMove();
            if (turnTracerOn)
                tracy.print(getCost(), iter);
            iter++;
            if (GREEDY && (nonImprovIter != 0 && nonImprovIter % threshold == 0)) {
                fastDescentStep(false);
            }
            //restart eventually if the threshold is reached
            restartFunction();
        }
        end();
    }

    public boolean isCostImproved() {
        //assert (space.hardObj <= bestCost);
        return (space.hardObj < bestCost);
    }

    public int getCost() {
        return space.hardObj;
    }

    public int nbrestart = 1;

    public void restartFunction() {
        int spenttime = (int) System.currentTimeMillis() - starttime;
        if (restartHardSolver && spenttime >= nbrestart * restartThreshold) {
            space.rand = new Random(seed + 100 * nbrestart);
            space.randomInit();
            nbrestart++;
            bestCost = space.hardObj;
        }
    }

    public void setDefaultNeighbourhood() {
        //dynamicNeighbourhood();
        addNeighbourhood(AbstractSolver.SWAP_TIMESLOTS);
        addNeighbourhood(AbstractSolver.TRANSLATE_EVENTS_MATCHING);
        addNeighbourhood(AbstractSolver.HORIZ_HUNGARIAN_MOVE);
        addNeighbourhood(AbstractSolver.TRANSLATE_EVENTS);
        addNeighbourhood(AbstractSolver.SWAP_EVENTS);
    }

    public void setLNS() {
        LNS = true;
        cpmove.initialize();
    }

    //****************************************************************//
    //************* A fast Hungarian based descent *******************//
    //****************************************************************//

    public void fastDescentStep(boolean completesolve) {
        if (greedy == null) {
            greedy = new HungarianGreedy(this);
        }
        if (completesolve) {
            bestCost = Integer.MAX_VALUE;
            iter = 0;
            space.initSolutionSpace();
            greedy.fillAssignmentMatrix();
            while (!doWeStop()) {
                matchingTimeSlots();
                boolean improve = greedy.mainHLoop();
                if (improve) nonImprovIter = 0;
                else nonImprovIter++;
                if (nonImprovIter != 0 && nonImprovIter % 20 == 0) {
                    diversificationStep();
                }
            }
            iter = greedy.nbhc;
            greedy.nbhc = 0;
            greedy = null;
            end();
        } else {
            matchingTimeSlots();
            while (greedy.mainHLoop()) {
            }
        }
    }

    //****************************************************************//
    //************* Opimitizing the timeslots ************************//
    //****************************************************************//

    public void matchingTimeSlots() {
//        boolean hasImproved = false;
        for (int t = 0; t < space.T; t++) {
            hhunge.setTimeSlot(t);
            int ctrans = hhunge.evalCostMove();
            if (ctrans < 0) { //do NOT accept side walk steps !
                hhtime = t;   //This move can only be improving so no need to check it with the metaheuristic
                hhunge.performMove();
//                hasImproved = true;
            }
        }
        if (isCostImproved()) storeNewImprovingSol();
    }

    //****************************************************************//
    //************* Diversification step *****************************//
    //****************************************************************//

    /**
     * Perform k random swaps
     */
    public void diversificationStep() {
        int k = 0;
        while (k < 50) {
            int e1 = space.rand.nextInt(space.E);
            int e2 = space.rand.nextInt(space.E);
            if (e1 == e2) { //bad random here but do not matter
                if (e1 < (space.E - 1)) e1++;
                else e1--;
            }
            swape.setE1(e1);
            swape.setE2(e2);
            swape.performMove();
            if (isCostImproved()) {
                storeNewImprovingSol();
            }
            k++;
        }
    }

    //****************************************************************//
    //************* Dealing with least degrading move ****************//
    //****************************************************************//

    protected SwapEvents ldSwap;
    protected TranslateEvent ldTranslation;
    protected int delta;
    protected int ldmove;

    public void reinitLDMove() {
        delta = Integer.MAX_VALUE;
        ldmove = -1;
    }

    public Move getLDMove() {
        if (delta < Integer.MAX_VALUE) {
            switch (ldmove) {
                case SWAP_EVENTS:
                    return ldSwap;
                case TRANSLATE_EVENTS:
                    return ldTranslation;
                default:
                    return null;
            }
        } else return null;
    }

    /**
     * if no degrading move has been found, we switch to a diversification
     * phase.
     *
     * @return true if a move has been performed, false to go to a diversification
     *         step
     */
    public boolean performLDMove() {
        Move move = getLDMove();
        if (move != null) {
            move.delta = delta;
            move.performMove();
            mh.informOfMove(move);
            return true;
        } else {
            return false;
        }
    }

    public void storeLDMove(SwapEvents swp, int delt) {
        if (delt < delta) {
            ldmove = SWAP_EVENTS;
            ldSwap.setFromEvt(swp);
            delta = delt;
        }
    }

    public void storeLDMove(TranslateEvent tre, int delt) {
        if (delt < delta) {
            ldmove = TRANSLATE_EVENTS;
            ldTranslation.setFromEvt(tre);
            delta = delt;
        }
    }

    //****************************************************************//
    //*************** Neighbourhood parameters ***********************//
    //****************************************************************//

    //the order in which events will be considered for translation
    //and swap
    protected Integer[] evts;

    //****************************************************************//
    //*************** Translating structure **************************//
    //****************************************************************//

    //remember the last event translated to start the scanning
    //from the next one
    protected int tevt1;

    /**
     * @return true if an improving translation has been found
     */
    public boolean exploreTranslationDeterministically() {
        if (exploreTranslationDeterministically(tevt1 + 1, space.E)) {
            return true;
        } else return exploreTranslationDeterministically(0, tevt1 + 1);
    }


    public boolean exploreTranslationDeterministically(int deb, int fin) {
        for (int i = deb; i < fin; i++) {
            int e = evts[i];
            transe.setEvt(e);
            for (Iterator<Place> it = space.freeSlots.iterator(); it.hasNext();) {
                Place place = it.next();
                transe.setFreeslot(place);
                int ctrans = transe.evalCostMove();
                boolean accept = mh.accept(transe, ctrans);
                if (ctrans <= 0 && accept) {  //accept side walk steps
                    tevt1 = e;
                    return true;
                } else if (accept) {
                    storeLDMove(transe, ctrans);
                }
            }
        }
        return false;
    }

    //****************************************************************//
    //*************** Translating with matching **********************//
    //****************************************************************//

    //remember the last event translated to start the scanning
    //from the next one
    protected int tevt1m;

    /**
     * @return true if an improving translation has been found
     */
    public boolean exploreTranslationMatchingDeterministically() {
        if (exploreTranslationMatchingDeterministically(tevt1m + 1, space.E)) {
            return true;
        } else return exploreTranslationMatchingDeterministically(0, tevt1m + 1);
    }


    public boolean exploreTranslationMatchingDeterministically(int deb, int fin) {
        for (int i = deb; i < fin; i++) {
            int e = evts[i];
            transem.setEvt(e);
            for (int t = 0; t < space.T; t++) {
                if (t != space.timeslot[e]) {
                    transem.setTimeslot(t);
                    int ctrans = transem.evalCostMove();
                    if (ctrans < Integer.MAX_VALUE) {
                        boolean accept = mh.accept(transem, ctrans);
                        if (ctrans <= 0 && accept) {  //accept side walk steps
                            tevt1m = e;
                            return true;
                        }
                    }
                }
            }
        }
        return false;
    }


    //****************************************************************//
    //*************** Swapping events structure **********************//
    //****************************************************************//
    //remember the last event swapped to start the scanning
    //from the next one
    protected int sevt1;

    /**
     * @return true if an improving swap has been found
     */
    public boolean exploreSwapsDeterministically() {
        if (exploreSwapsDeterministically(sevt1 + 1, space.E)) {
            return true;
        } else return exploreSwapsDeterministically(0, sevt1 + 1);
    }

    public boolean exploreSwapsDeterministically(int deb, int fin) {
        for (int i = deb; i < fin; i++) {
            int e1 = evts[i];
            swape.setE1(e1);
            for (int j = i + 1; j < space.E; j++) {
                int e2 = evts[j];
                //boolean b = space.timeslot[e1] != space.timeslot[e2];
                swape.setE2(e2);
                int cswap = swape.evalCostMove();
                boolean accept = mh.accept(swape, cswap);
                if (cswap <= 0 && accept) { //accept side walk steps
                    sevt1 = e1;
                    return true;
                } else if (accept) {
                    storeLDMove(swape, cswap);
                }
            }
        }
        return false;
    }

    //****************************************************************//
    //*************** Swapping Timeslots structure *******************//
    //****************************************************************//
    //remember the last timeslot swapped to start the scanning
    //from the next one
    protected int time1;

    /**
     * @return true if an improving swap has been found
     */
    public boolean exploreTSwapsDeterministically() {
        if (exploreTSwapsDeterministically(time1 + 1, space.T)) {
            return true;
        } else return exploreTSwapsDeterministically(0, time1 + 1);
    }

    public boolean exploreTSwapsDeterministically(int deb, int fin) {
        for (int t1 = deb; t1 < fin; t1++) {
            swapt.setT1(t1);
            for (int t2 = t1 + 1; t2 < space.T; t2++) {
                swapt.setT2(t2);
                int cswap = swapt.evalCostMove();
                if (cswap < 0 && mh.accept(swapt, cswap)) { //accept side walk steps
                    time1 = t1;
                    return true;
                }
            }
        }
        return false;
    }

    //****************************************************************//
    //*************** Horizontal Hungarian structure *****************//
    //****************************************************************//

    //Remember the last timeslot improved by matching
    protected int hhtime;

    /**
     * @return true if an improving inter timeslot hungarian move has been found
     */

    public boolean exploreHHungarianDeterministically() {
        if (exploreHHungarianDeterministically(hhtime, space.T)) {
            return true;
        } else return exploreHHungarianDeterministically(0, hhtime);
    }


    public boolean exploreHHungarianDeterministically(int deb, int fin) {
        for (int t = deb; t < fin; t++) {
            hhunge.setTimeSlot(t);
            int ctrans = hhunge.evalCostMove();
            if (ctrans < 0) { //do NOT accept side walk steps !
                hhtime = t;   //This move can only be improving so no need to check it with the metaheuristic
                return true;
            }
        }
        return false;
    }

    //****************************************************************//
    //*************** Vertical Hungarian structure *******************//
    //****************************************************************//

    /**
     * @return a random event (with conflict if possible) in the given timeslot
     *         -1 if no event is choosed but there is a free room
     *         -2 if no event can't be choosed because of precedences and
     *         there is no free room in the timeslot
     */
    protected int pickARandomEvtInTimeslot(int timeslot, LinkedList<Integer> l) {
        int nbevt = space.nbept[timeslot];
        if (nbevt == 0) return -1;
        int rdevt = space.rand.nextInt(nbevt);
        int bestre = -1;
        for (int i = rdevt; i < nbevt; i++) {
            int e = space.listevts[timeslot][i];
            if (precedenceFree(e, l)) {
                if (space.eventcost[e] > 0) return e;
                else if (bestre == -1) bestre = e;
            }
        }
        for (int i = 0; i < rdevt; i++) {
            int e = space.listevts[timeslot][i];
            if (precedenceFree(e, l)) {
                if (space.eventcost[e] > 0) return e;
                else if (bestre == -1) bestre = e;
            }
        }
        if (bestre != -1) return bestre;
        if (nbevt == space.R)
            return -2;
        return -1;
    }

    /**
     * check if the event evt has no precedence relationships within any
     * of the events of list l
     *
     * @param evt
     * @param l
     * @return
     */
    public boolean precedenceFree(int evt, List<Integer> l) {
        if (space.isPrecedencFree(evt)) {
            return true;
        } else {
            for (Iterator<Integer> it = l.iterator(); it.hasNext();) {
                Integer e = it.next();
                if (space.isThereAPrecRelationShip(evt, e))
                    return false;
            }
        }
        return true;
    }


    /**
     * @return true if an improving hungarian move has been found
     */
    public boolean exploreHungarianStochastically() {
        LinkedList<Integer> levt = new LinkedList<Integer>();
        for (int i = 0; i < nbRandomHungarian; i++) {
            levt.clear();
            int j;
            for (j = 0; j < space.T; j++) {
                int e = pickARandomEvtInTimeslot(j, levt);
                if (e != -1)
                    levt.add(e);
                vhunge.setEvt(j, e);
            }
            int chung = vhunge.evalCostMove();
            if (chung <= 0 && mh.accept(vhunge, chung)) {
                return true;
            }
        }
        return false;
    }

    //****************************************************************//
    //*************** LNS CP move ************************************//
    //****************************************************************//

    public int nbiterlns = 0;
    public int nbRealAttempt = 0;
    public int nbAttempt = 0;
    public boolean isLNSDecisive = false;

    public static int frequency = 1000;
    public static int thresholdLNS = 5;
    public static int sspacesize = 80;
    public static int freeEvt = 30;


    public boolean addLNSStep() {
        nbiterlns++;
        if (nbiterlns >= frequency && getCost() <= thresholdLNS) {
            nbiterlns = 0;
            return true;
        }
        return false;
    }

    /**
     * Attempt a LNS move calling the CP solver
     */
    public boolean exploreLNS() {
        //System.out.println("Explore LNS");
        BitSet ct = getConflictingTimeslot();
        BitSet fevts = getEventsTRelax(ct);
        cpmove.freeEvents(fevts);
        int cpcost = cpmove.evalCostMove();
        nbAttempt++;
        if (cpcost < 0) { //do NOT accept side walk steps !
            isLNSDecisive = true;
            nbRealAttempt++;
            return true;
        } else {
            if (cpcost == 1) {
                nbRealAttempt++;
            }
            return false;
        }
    }

    public BitSet getConflictingTimeslot() {
        BitSet conflictingTimeslot = new BitSet();
        for (int i = 0; i < space.E; i++) {
            if (space.eventcost[i] > 0) {
                conflictingTimeslot.set(space.timeslot[i]);
            }
        }
        return conflictingTimeslot;
    }

    public BitSet getEventsTRelax(BitSet ct) {
        BitSet freeEvents = new BitSet();
        //relax conflicting time
        for (int i = 0; i < space.E; i++) {
            if (ct.get(space.timeslot[i])) {
                freeEvents.set(i);
            }
        }
        //add promising timeslots ?
        while (freeEvents.cardinality() < freeEvt) { //30 before
            int mint = -1;
            int minval = Integer.MAX_VALUE;
            for (int i = 0; i < space.T; i++) {
                if (!ct.get(i) && minval > space.nbept[i]) {
                    mint = i;
                    minval = space.nbept[i];
                }
            }
            if (mint == -1) break;
            for (int i = 0; i < space.nbept[mint]; i++) {
                freeEvents.set(space.listevts[mint][i]);
            }
            ct.set(mint);
        }
        return freeEvents;
    }


}
