package localsearch.solvers.complete;

import localsearch.solvers.coloring.TwoStepHardSolver;
import localsearch.Benchmark;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: 13 dc. 2007
 * Time: 17:46:59
 * To change this template use File | Settings | File Templates.
 */
public class SearchLoop {

	public static boolean searchLoopTrace = false;

	protected int tlimit;
	protected int faketlimit;
	public TwoStepHardSolver hardS;
	public SoftSolver softS;

	protected double ratioIntens = 0d;
	protected int intensificationPhase = 0;


	/**
	 * @param from2002
	 * @param name
	 * @param seed
	 * @param intensif : the pourcentage of the timelimit we want to spend on
	 *        a pure LNS phase.
	 */
	public SearchLoop(String name, int seed, double intensif) {
		hardS = new TwoStepHardSolver(name,seed);
		ratioIntens = intensif;
	}

	public void setParamSolver(int timelimit, int tabusize) {
		tlimit = timelimit;
		if (ratioIntens >= 0d) {
			intensificationPhase = (int) (ratioIntens * (double) timelimit);
		}
		//System.out.println("Intensification phase: " + intensificationPhase);
		faketlimit = tlimit - intensificationPhase;
		hardS.setTabuMetaHeuristic(tabusize);
		hardS.setTimelimit(tlimit);
		hardS.setDefaultNeighbourhood();
	}

    public void setParamSoftSolver() {
        softS = new SoftSolver(hardS.hs);
        if (Benchmark.pureLNS) softS.setLNS();
        else if (Benchmark.SA_LNS) softS.setLNS_SA();        
        softS.setTimelimit(faketlimit);
        softS.setDefaultNeighbourhood();
        if (!softS.LNS)
            softS.setSAMetaHeuristic(1 - ratioIntens);
    }

    public void stopByChangingTimeLimit() {
        this.faketlimit = 0;
        this.tlimit = 0;
    }

    public boolean solveLoop() {
		boolean feasreached = false;
		while (((int) System.currentTimeMillis() - hardS.getStartTime()) < faketlimit && ((softS == null) || softS.getBestCost() != 0)) {
			if (searchLoopTrace) System.out.println("NEXT FEASIBLE SOLUTION");
			hardS.solve();     //solve the feasible part
			if (hardS.getBestCost() == 0) {//feasible solution found
				if (searchLoopTrace) System.out.println("SOLVE SOFT ");
				feasreached = true;
				setParamSoftSolver();
                softS.solve(); //solve the soft part
				if (searchLoopTrace) System.out.println("REACHED " + softS.getBestCost() + " " + ((int) System.currentTimeMillis() - hardS.getStartTime()));
			}
		}
		if (feasreached
			&& ((int) System.currentTimeMillis() - hardS.getStartTime()) < tlimit
			&& softS.getBestCost() != 0) {
			if (searchLoopTrace)
				System.out.println("Final intensification with time left : " + (tlimit - ((int) System.currentTimeMillis() - hardS.getStartTime())));
			softS.setTimelimit(tlimit);
			softS.setLNS();
			softS.setDefaultNeighbourhood();
			softS.finalIntensification = true;
			softS.solve();
			if (searchLoopTrace)
				System.out.println("REACHED " + softS.getBestCost() + " " + ((int) System.currentTimeMillis() - hardS.getStartTime()));
		}
		return feasreached;
	}


}
