package localsearch.solvers.complete;

import cpmodel.constraints.ConsecutiveConstraint;
import localsearch.SolutionSpace;

import java.util.*;

/**
 * A class that allow to project the current state of the cost
 * on days / events / timeslots
 */
public class SoftCostProjection {

	// defining the number of conflicting timeslots and randomtimeslots
	// we want to remove
	public static int s_nbct = 1;
	public static int s_nbrt = 2;
	public static int s_nbroom = 1;

	protected int nbct = 1;
	protected int nbrt = 2;
	protected int nbroom = 1;


	protected SoftSolver ssolver;
	protected SolutionSpace space;
	protected ArrayList<Integer> events;
	protected ArrayList<Integer> timesl;

	//cost1_X_[i] gives the contribution of X (event or timeslot)
	//i in cost 1
	protected int[] cost1_event_P;
	protected int[] cost1_times_P;

	//cost2_X_[i] gives the contribution of X (event or timeslot)
	//i in cost 2
	protected int[] cost2_event_P;
	protected int[] cost2_times_P;

	//cost3_X_[i] gives the contribution of X (event or timeslot)
	//i in cost 3
	protected int[] cost3_event_P;
	protected int[] cost3_times_P;


	public SoftCostProjection(SoftSolver ss) {
		this.ssolver = ss;
		this.space = ss.space;
		nbct = s_nbct;
		nbrt = s_nbrt;
		cost1_event_P = new int[space.E];
		cost1_times_P = new int[space.T];
		cost2_event_P = new int[space.E];
		cost2_times_P = new int[space.T];
		cost3_event_P = new int[space.E];
		cost3_times_P = new int[space.T];
		event_pcost = new int[space.T];
		events = new ArrayList<Integer>();
		timesl = new ArrayList<Integer>();
		for (int i = 0; i < space.E; i++) {
			events.add(i);
		}
		for (int i = 0; i < space.T; i++) {
			timesl.add(i);
		}

	}

    public int[] getCost1_event_P() {
        return cost1_event_P;
    }

    public int[] getCost1_times_P() {
        return cost1_times_P;
    }

    public int[] getCost2_event_P() {
        return cost2_event_P;
    }

    public int[] getCost2_times_P() {
        return cost2_times_P;
    }

    public int[] getCost3_event_P() {
        return cost3_event_P;
    }

    public int[] getCost3_times_P() {
        return cost3_times_P;
    }

    public LinkedList<Integer> getConfictsEvts() {
		LinkedList<Integer> etslist = new LinkedList<Integer>();
		for (int e = 0; e < space.E; e++) {
			int c1 = cost1_event_P[e] + cost2_event_P[e] + cost3_event_P[e];
			if (c1 > 0) {
				etslist.add(e);
			}
		}
		return etslist;
	}

	public LinkedList<Integer> getConfictsTims() {
		LinkedList<Integer> etslist = new LinkedList<Integer>();
		for (int e = 0; e < space.T; e++) {
			int c1 = cost1_times_P[e] + cost2_times_P[e] + cost3_times_P[e];
			if (c1 > 0) {
				etslist.add(e);
			}
		}
		return etslist;
	}


	public void growNeigbourhood() {
		if (SoftSolver.roomn) {
			if (nbct >= 2*nbroom) {
				nbroom ++;
			} else {
				nbct++;
			}
		} else {
			if (nbct + nbrt < 30) {
				if (nbct >= nbrt) {
					nbrt++;
				} else {
					nbct++;
				}
			}
		}
	}

	//*******************************************************//
	//**************** Computing the relaxation *************//
	//*******************************************************//


	public BitSet computeEventsToRelaxBasedOnTimeslot() {
		BitSet freeTimSet = new BitSet();
		BitSet freeEvents = new BitSet();
		projectCost();
		//findKMostConflictingElement(timesl, new TimeslotComparator(),freeTimSet,nbct);
		List<Integer> lctimes = getConfictsTims();
		findKRandomElementAmong(lctimes, new TimeslotComparator(), freeTimSet, nbct, 3);
		addKRandomElements(freeTimSet, nbrt, space.T);

		for (int i = 0; i < space.T; i++) {
			if (freeTimSet.get(i))
				addTimeSlot(freeEvents, i);
		}
		//System.out.println("timeslots choosed " + freeTimSet + " " + freeEvents.cardinality());
		return freeEvents;
	}

	public BitSet computeEventsToRelaxBasedOnTimeslotAndRooms() {
		BitSet freeTimSet = new BitSet();
		BitSet freeEvents = new BitSet();
		projectCost();
		//findKMostConflictingElement(timesl, new TimeslotComparator(),freeTimSet,nbct);
		List<Integer> lctimes = getConfictsTims();
		findKRandomElementAmong(lctimes, new TimeslotComparator(), freeTimSet, nbct, 3);

		//addKRandomElements(freeTimSet, 1, space.T);
		addKRandomRoom(freeEvents, nbroom);

		for (int i = 0; i < space.T; i++) {
			if (freeTimSet.get(i))
				addTimeSlot(freeEvents, i);
		}
		//System.out.println("timeslots choosed " + freeTimSet + " " + freeEvents.cardinality());
		return freeEvents;
	}


	public void addKRandomRoom(BitSet freeEvents, int k) {
		int cpt = 0;
		BitSet freeRooms = new BitSet();
		while (cpt < k) {
			int room_idx = space.rand.nextInt(space.R - cpt);
			int rr, tempIdx = -1;
			for (rr = 0; rr < space.R; rr++) {
				if (!freeRooms.get(rr)) {
					tempIdx++;
				}
				if (tempIdx == room_idx) break;
			}
			freeRooms.set(rr);
			cpt++;
		}
		for (int rr = freeRooms.nextSetBit(0); rr > 0; rr = freeRooms.nextSetBit(rr + 1)) {
			for (int i = 0; i < 45; i++) {
				if (space.matrix[i][rr] != -1 && space.rand.nextDouble() <= 0.25d) {
					freeEvents.set(space.matrix[i][rr]);
				}
			}
		}
	}

	//add the k most conflicting elements (events or timeslots)
	//in the BitSet freeElts
	public void findKMostConflictingElement(List<Integer> list, Comparator comp,
	                                        BitSet freeElts, int k) {
		Collections.sort(list, comp);
		Iterator<Integer> it = list.iterator();
		int cpt = 0;
		while (it.hasNext() && cpt < k) {
			freeElts.set(it.next());
			cpt++;
		}
	}


	//add k random elements (events or timeslots) of the list
	//in the BitSet freeElts
	public void findKRandomElementAmong(List<Integer> list, Comparator comp, BitSet freeElts, int k, int ratio) {
		Collections.sort(list, comp);
		List<Integer> nlist = list.subList(0, list.size() / ratio + 1);
		if (nlist.size() < k) nlist = list;
		Collections.shuffle(nlist, space.rand);
		Iterator<Integer> it = nlist.iterator();
		int cpt = 0;
		while (it.hasNext() && cpt < k) {
			int elt = it.next();
			if (!freeElts.get(elt)) {
				freeElts.set(elt);
				cpt++;
			}
		}
	}

	//add k random element among [0..size] except those already
	//included in the BitSet freeEvts. They are therefore removed from the
	//choices
	public void addKRandomElements(BitSet freeElts, int k, int size) {
		int cste = freeElts.cardinality();
		int cpt = 0;
		while (cpt < k) {
			int event_idx = space.rand.nextInt(size - cpt - cste);
			int e, tempIdx = -1;
			for (e = 0; e < size; e++) {
				if (!freeElts.get(e)) {
					tempIdx++;
				}
				if (tempIdx == event_idx) break;
			}
			freeElts.set(e);
			cpt++;
		}
	}

	/**
	 * add timeslot t in the set of events freeEvents
	 *
	 * @param freeEvents
	 * @param t
	 */
	public void addTimeSlot(BitSet freeEvents, int t) {
		for (int j = 0; j < space.nbept[t]; j++) {
			freeEvents.set(space.listevts[t][j]);
		}
	}

	public void addTimeSlot(List<Integer> freeEvents, int t) {
		for (int j = 0; j < space.nbept[t]; j++) {
			freeEvents.add(space.listevts[t][j]);
		}
	}


	public void printList() {
		Iterator<Integer> it = events.iterator();
		while (it.hasNext()) {
			int e = it.next();
			int c1 = cost1_event_P[e] + cost2_event_P[e] + cost3_event_P[e];
			System.out.println(e + " " + c1);
		}
	}

	public class EventComparator implements Comparator {

		public int compare(Object o, Object o1) {
			Integer e1 = (Integer) o;
			Integer e2 = (Integer) o1;
			int c1 = cost1_event_P[e1] + cost2_event_P[e1] + cost3_event_P[e1];
			int c2 = cost1_event_P[e2] + cost2_event_P[e2] + cost3_event_P[e2];
			if (c1 < c2) {
				return 1;
			} else if (c1 > c2) {
				return -1;
			} else return 0;
		}
	}

	public class TimeslotComparator implements Comparator {

		public int compare(Object o, Object o1) {
			Integer e1 = (Integer) o;
			Integer e2 = (Integer) o1;
			int c1 = cost1_times_P[e1] + cost2_times_P[e1] + cost3_times_P[e1];
			int c2 = cost1_times_P[e2] + cost2_times_P[e2] + cost3_times_P[e2];
			if (c1 < c2) {
				return 1;
			} else if (c1 > c2) {
				return -1;
			} else return 0;
		}
	}

	//*******************************************************//
	//**************** Projecting the costs *****************//
	//*******************************************************//

	public void projectCost() {
		computeProjectionC1();
		computeProjectionC2();
		computeProjectionC3();
	}

	// project cost1 on the events
	public void computeProjectionC1() {
		Arrays.fill(cost1_event_P, 0);
		Arrays.fill(cost1_times_P, 0);
		for (int i = 0; i < space.D; i++) {
			int t = (i + 1) * space.H - 1;
			int st = 0;
			for (int j = 0; j < space.nbept[t]; j++) {
				int evt = space.listevts[t][j];
				cost1_event_P[evt] = space.evtSize[evt];
				st += space.evtSize[evt];
			}
			cost1_times_P[t] = st;
		}
	}

	// project cost2 on the events and timeslots
	public void computeProjectionC2() {
		Arrays.fill(cost2_event_P, 0);
		Arrays.fill(cost2_times_P, 0);
		int i, j, x, y, z;

		for (i = 0; i < 5; ++i) {
			for (j = 0; j < 7; ++j) {
				int t1 = i * 9 + j;
				int t2 = t1 + 1;
				int t3 = t2 + 1;
				for (x = 0; x < space.nbept[t1]; ++x) {
					for (y = 0; y < space.nbept[t2]; ++y) {
						int evt1 = space.listevts[t1][x];
						int evt2 = space.listevts[t2][y];
						int idx2 = ConsecutiveConstraint.get2Index(evt1, evt2);
						if (ConsecutiveConstraint.costList[idx2] != null) {
							for (z = 0; z < space.nbept[t3]; ++z) {
								int evt3 = space.listevts[t3][z];
								int idx3 = ConsecutiveConstraint.get3Index(evt1, evt2, evt3);
								int cost = ConsecutiveConstraint.tripletCost[idx3];
								cost2_event_P[evt1] += cost;
								cost2_event_P[evt2] += cost;
								cost2_event_P[evt3] += cost;
								cost2_times_P[t1] += cost;
								cost2_times_P[t2] += cost;
								cost2_times_P[t3] += cost;
							}
						}
					}
				}
			}
		}
	}

	// project cost3 on the events and timeslots
	public void computeProjectionC3() {
		Arrays.fill(cost3_event_P, 0);
		Arrays.fill(cost3_times_P, 0);
		for (int s = 0; s < space.S; s++) {
			for (int d = 0; d < space.D; d++) {
				if (space.studentPlanning[s][d][0] == 1) {
					for (int i = 1; i < 10; i++) {
						int evt = space.studentPlanning[s][d][i];
						if (evt != -1) {
							cost3_event_P[evt]++;
							cost3_times_P[i - 1]++;
							break;
						}
					}
				}
			}
		}
	}

	//*******************************************************//
	//**************** Retro-Projection of costs ************//
	//*******************************************************//

	// the sum of the cost of placing each event of the conflicting timeslots
	// in the every other timeslots timeslots
	public int[] event_pcost;


	public BitSet computeCostGuidedNeighborhood() {
		BitSet freeTimSet = new BitSet();
		BitSet freeEvents = new BitSet();
		List<Integer> levts = new LinkedList<Integer>();
		projectCost();

		List<Integer> lctimes = getConfictsTims();
		findKRandomElementAmong(lctimes, new TimeslotComparator(), freeTimSet, nbct, 3);

		if (space.rand.nextDouble() <= 0.75d) {
			for (int i = freeTimSet.nextSetBit(0); i >= 0; i = freeTimSet.nextSetBit(i + 1)) {
				addTimeSlot(levts, i);
			}

			computeFProjection(levts, freeTimSet);
			findKRandomElementAmong(lctimes, new PromisingTimeslotComparator(), freeTimSet, nbrt, 7);

			for (int i = freeTimSet.nextSetBit(0); i >= 0; i = freeTimSet.nextSetBit(i + 1)) {
				addTimeSlot(freeEvents, i);
			}
		} else addKRandomElements(freeTimSet, nbrt, space.T);

		//System.out.println("Size: " + freeEvents.cardinality() + " " + freeEvents);
		return freeEvents;
	}


	/**
	 * compute the cost of placing each event in the candidate timeslots
	 *
	 * @param candidateTimeslots
	 * @param freeEvts
	 */
	public void computeFProjection(List<Integer> candidateEvents, BitSet freeTimes) {
		Arrays.fill(event_pcost, 0);
		for (int i = freeTimes.nextSetBit(0); i >= 0; i = freeTimes.nextSetBit(i + 1)) {
			event_pcost[i] = Integer.MAX_VALUE;
		}
		computeFProjectionC1(candidateEvents, freeTimes);
		computeFProjectionC2(candidateEvents, freeTimes);
		computeFProjectionC3(candidateEvents, freeTimes);
		//Collections.sort(timesl, new PromisingTimeslotComparator());
	}

	/**
	 * compute the projection of cost1 of each events in the candidate timeslots
	 *
	 * @param candidateTimeslots
	 */
	public void computeFProjectionC1(List<Integer> candidateEvents, BitSet freeTimes) {
		for (int i = 0; i < space.T; i++) {
			if (space.isFinalTimeslot(i) && !freeTimes.get(i)) {
				for (Iterator<Integer> it = candidateEvents.iterator(); it.hasNext();) {
					int evt = it.next();
					event_pcost[i] += space.evtSize[evt];
				}
			}
		}
	}

	public void computeFProjectionC2(List<Integer> candidateEvents, BitSet freeTimes) {
		for (Iterator<Integer> it = candidateEvents.iterator(); it.hasNext();) {
			int evt = it.next();
			for (int i = 0; i < space.T; i++) {
				if (!freeTimes.get(i)) {
					event_pcost[i] += getPenalty(evt, i);
				}
			}
		}
	}

	public void computeFProjectionC3(List<Integer> candidateEvents, BitSet freeTimes) {
		for (Iterator<Integer> it = candidateEvents.iterator(); it.hasNext();) {
			int evt = it.next();
			for (int i = 0; i < space.T; i++) {
				if (!freeTimes.get(i) && space.day[i] != space.day[space.timeslot[evt]])
					event_pcost[i] += getDayPenality(evt, space.day[i]);
			}
		}
	}


	public int getPenalty(final int evt, final int tst) {
		int extracost = 0;
		if (tst > 1 && space.sameDay(tst - 2, tst - 1, tst))
			extracost += getPenalty(evt, tst - 2, tst - 1);
		if (tst > 0 && tst < 44 && space.sameDay(tst - 1, tst, tst + 1))
			extracost += getPenalty(evt, tst - 1, tst + 1);
		if (tst < 43 && space.sameDay(tst, tst + 1, tst + 2))
			extracost += getPenalty(evt, tst + 1, tst + 2);
		return extracost;
	}

	private int getPenalty(final int evt, final int t1, final int t2) {
		int x, y, extracost = 0;
		for (x = 0; x < space.nbept[t1]; ++x) {
			int evt1 = space.listevts[t1][x];
			if (evt1 != evt) {
				int idx1 = ConsecutiveConstraint.get2Index(evt, evt1);
				if (ConsecutiveConstraint.costList[idx1] != null) {
					for (y = 0; y < space.nbept[t2]; ++y) {
						int evt2 = space.listevts[t2][y];
						if (evt2 != evt) {
							int idx3 = ConsecutiveConstraint.get3Index(evt1, evt2, evt);
							extracost += ConsecutiveConstraint.tripletCost[idx3];
						}
					}
				}
			}
		}
		return extracost;
	}

	public int getDayPenality(int evt, int d) {
		int penality = 0;
		int cday = space.day[space.timeslot[evt]];
		for (int i = 0; i < space.studentlist[evt].length; i++) {
			int s = space.studentlist[evt][i];
			if (space.studentPlanning[s][cday][0] == 2)
				penality++;
			if (space.studentPlanning[s][d][0] == 0)
				penality++;
			if (space.studentPlanning[s][cday][0] == 1)
				penality--;
			if (space.studentPlanning[s][d][0] == 1)
				penality--;
		}
		return penality;
	}


	public class PromisingTimeslotComparator implements Comparator {

		public int compare(Object o, Object o1) {
			Integer e1 = (Integer) o;
			Integer e2 = (Integer) o1;
			int c1 = event_pcost[e1];
			int c2 = event_pcost[e2];
			if (c1 > c2) {
				return 1;
			} else if (c1 < c2) {
				return -1;
			} else return 0;
		}
	}

}

/*computeFProjection(lts, freeEvents);
				int cpt = 0;
				LinkedList fakeEvts = new LinkedList<Integer>();
				for (Iterator<Integer> it = events.iterator(); it.hasNext();) {
					int evt = it.next();
					if (!freeEvents.get(evt)) {
						fakeEvts.addLast(evt);
					}
				}
				List<Integer> nlist = fakeEvts.subList(0, fakeEvts.size() / 3 + 1);
				if (nlist.size() < space.R) nlist = fakeEvts;
				Collections.shuffle(fakeEvts);
				for (Iterator<Integer> it = fakeEvts.iterator(); it.hasNext();) {
					int evt = it.next();
					freeEvents.set(evt);
					cpt++;
					if (cpt == space.R) break;
				} */
