package localsearch.solvers.complete;

import localsearch.SolutionSpace;
import localsearch.metah.SimulatedAnnealing;
import localsearch.metah.TrueMH;
import localsearch.moves.complete.softmoves.SoftCPMove;
import localsearch.moves.complete.softmoves.SoftSwapEvents;
import localsearch.moves.complete.softmoves.SoftSwapTimeslots;
import localsearch.moves.complete.softmoves.SoftTranslateEvent;

import java.util.BitSet;
import java.util.Collections;
import java.util.LinkedList;

/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: 20 aot 2007
 * Time: 11:29:50
 * To change this template use File | Settings | File Templates.
 */
public class SoftSolver extends HardSolver {

	/**
	 * image of the feasible solution from which the soft LS
	 * is starting
	 */
	protected HardSolSAV sav;

	public HardSolver hsolve;

	protected boolean finalIntensification = false;

	//Use the LNS move inside the SA
	public boolean LNS_SA;

	//Use the LNS move as an intensification step after the SA
	protected boolean LNS_SA_Intens;

	protected boolean newImprovingSolJustFound = false;


    public SoftSolver(){} 

    public SoftSolver(String ins, int seed) {
		super(ins, seed);
	}

	public SoftSolver(HardSolver solver) {
		//this.tracy = new Tracer(name, greedyInit);
		hsolve = solver;
		space = solver.space;
		bestsolution = new int[space.T][space.R];
		for (int i = 0; i < space.T; i++) {
			System.arraycopy(solver.bestsolution[i], 0, bestsolution[i], 0, space.R);
		}
		ins = solver.ins;
		name = solver.name;
		timelimit = solver.timelimit;
		seed = solver.seed;
		mh = new TrueMH(this);
		moves = new LinkedList<Integer>();
		swape = new SoftSwapEvents(space);
		swapt = new SoftSwapTimeslots(space);
		transe = new SoftTranslateEvent(space);
		ldTranslation = new SoftTranslateEvent(space);
		ldSwap = new SoftSwapEvents(space);
		cpmove = new SoftCPMove(space, this);
		scp = new SoftCostProjection(this);
	}

	public void start() {
		//tracy = new Tracer(name,true);
		starttime = hsolve.starttime;
		if (firstRun) {
			sav = new HardSolSAV(space);
			sav.save(space);
		}
		if (finalIntensification) {
			restoreBestSolution();
			mh = new TrueMH(this);
		}
		space.baseSoftInit();
		space.softInit();
		space.computeSoftCostFromScratch();
		iter = 0;
		if (space.softObj < bestCost) {
			bestCost = space.softObj;
            storeNewImprovingSol();
        }
        if (trace) {
            System.out.println("initial soft cost : " + bestCost);
        }
        reinitLDMove();
		if (firstRun) {
			mh.initMhParameters();
			firstRun = false;
		}
		assert (space.assertAllSoftCosts());
	}

	//no restart for the SA
	public void restartFunction() {

	}

	public void setTimelimit(int timelimit) {
		this.timelimit = timelimit;
	}

	//note that the intern time limit is set to
	//the normal time limit if not used
	public int getTimelimit() {
		return timelimit; //intermediateTLimit;
	}

	public boolean doWeStop() {
		int ctime = (int) System.currentTimeMillis();
		int timeSpent = (ctime - starttime);
		if ((bestCost == 0) || (timeSpent > timelimit)) {
			return true;
		} else {
			return false;
		}
	}

	//assertion
	public boolean assertCostComputation() {
		space.assertAllSoftCosts();
		return bestCost <= space.softObj;
	}

	public boolean isCostImproved() {
		return (space.softObj < bestCost);
	}

	public int getCost() {
		return space.softObj;
	}

	public void solve() {
		super.solve();
	}

	public String getCompInfo() {
		return " t:" + (mh instanceof SimulatedAnnealing ? ((SimulatedAnnealing) mh).currentTemp : "");
	}

	public void setDefaultNeighbourhood() {
		//dynamicNeighbourhood();
		moves.clear();
		if (LNS || LNS_SA) {
			addNeighbourhood(CP_MOVE);
			cpmove.initialize();
		} else {
			addNeighbourhood(TRANSLATE_EVENTS);
		}
	}

	public void setLNS() {
		LNS = true;
	}

	public void setLNS_SA() {
		LNS_SA = true;
	}

	public void setLNS_SA_Intens() {
		LNS_SA_Intens = true;
	}

	/**
	 * Define the neighbourhood used at the current iteration
	 * it is controlled by a probability p for each kind of move
	 */
	public void defineNeighbourhood() {
		if (dynamicNeihbourhood) {
			moves.clear();
			moves.add(TRANSLATE_EVENTS);
			if (space.rand.nextDouble() <= phh) {
				moves.add(SWAP_EVENTS);
			}
			if (space.rand.nextDouble() <= pst) {
				moves.add(SWAP_TIMESLOTS);
			}
		} else if (LNS_SA) {
			moves.clear();
			moves.add(TRANSLATE_EVENTS);
			SimulatedAnnealing smh = (SimulatedAnnealing) mh;
			if (!smh.initializationPhase && smh.isInTheLastTenPourcent()) {
				if (newImprovingSolJustFound) {
					moves.clear();
					moves.add(CP_MOVE);
					newImprovingSolJustFound = false;
				} else {
					double proba = frequencyRatio * (1d / smh.currentTemp);
					//if (iter % 50000 == 0)
					//	System.out.println(smh.currentStep + "|" + smh.nbstep + " " + smh.isInTheLastTenPourcent() + " " + nbCPMove + " proba " + proba + " " + ((SimulatedAnnealing) mh).currentTemp);
					if (space.rand.nextDouble() <= proba) {
						moves.add(CP_MOVE);
					}
				}
			}
		}
		Collections.shuffle(moves, space.rand);
	}

/*	public void end() {
		super.end();
		System.out.println("SA : " + ((SimulatedAnnealing) mh).currentTemp + " " + iter);
	}
*/
	//****************************************************************//
	//************* Dealing with least degrading move ****************//
	//****************************************************************//

	public int nbemptystep = 0;

	public boolean performLDMove() {
		if (!LNS && !LNS_SA) {
			nbemptystep += 1;
			//reheating if the neigbourhood is searched without accepting any moves
			//Note feasible moves still exist !
			if (nbemptystep == 1) {
				nbemptystep = 0;
				((SimulatedAnnealing) mh).currentTemp = ((SimulatedAnnealing) mh).tempI;
			}
			return super.performLDMove();
		} else {
			return true; //do not go in a diversification step when LNS
		}
	}

	public void storeLDMove(SoftTranslateEvent tre, int delt) {
		if (delt < delta) {
			ldmove = TRANSLATE_EVENTS;
			ldTranslation.setEvt(tre.getEvt());
			((SoftTranslateEvent) ldTranslation).setFromEvt(tre);
			delta = delt;
		}
	}

	public void storeLDMove(SoftSwapEvents tre, int delt) {
		if (delt < delta) {
			ldmove = SWAP_EVENTS;
			((SoftSwapEvents) ldSwap).setFromEvt(tre);
			delta = delt;
		}
	}

	//****************************************************************//
	//************* Diversification step *****************************//
	//****************************************************************//

	/**
	 * no feasible move left in the search space, so restart
	 * from the initial hard solution
	 */
	public void diversificationStep() {
		System.out.println("Diversification step at iter " + iter + " from " + space.softObj);
		sav.restore(space);
		space.softInit();
		space.computeSoftCostFromScratch();
		System.out.println("soft cost " + space.softObj);
	}

	//****************************************************************//
	//*************** Translating structure **************************//
	//****************************************************************//

	public boolean exploreTranslationDeterministically(int deb, int fin) {
		SoftTranslateEvent softmove = (SoftTranslateEvent) transe;
		for (int e = deb; e < fin; e++) {
			softmove.setEvt(e);
			for (int i = 0; i < space.feastimeslots[e].length; i++) {
				int feast = space.feastimeslots[e][i];
				if (feast != space.timeslot[e]) { //do not move between the same timeslots
					softmove.setTimeslot(feast);
					if (softmove.isKeepingFeas()) {
						int ctrans = softmove.evalSoftCost();
						boolean accept = mh.accept(softmove, ctrans);
						if (ctrans <= 0 || accept) {  //accept side walk steps
							tevt1 = e;
							return true;
						} else {
							storeLDMove(softmove, ctrans);
						}
					}
				}
			}
		}
		return false;
	}

	//****************************************************************//
	//*************** Swapping events structure **********************//
	//****************************************************************//


	public boolean exploreSwapsDeterministically(int deb, int fin) {
		SoftSwapEvents softmove = (SoftSwapEvents) swape;
		for (int e1 = deb; e1 < fin; e1++) {
			softmove.setE1(e1);
			for (int i = 0; i < space.possibleswap[e1].length; i++) {
				int e2 = space.possibleswap[e1][i];
				softmove.setE2(e2);
				if (space.timeslot[e1] != space.timeslot[e2]) { //do not move between the same timeslots
					if (softmove.isKeepingFeas()) {
						int cswap = softmove.evalSoftCost();
						boolean accept = mh.accept(softmove, cswap);
						if (cswap <= 0 || accept) { //accept side walk steps
							sevt1 = e1;
							return true;
						} else {
							storeLDMove(softmove, cswap);
						}
					}
				}
			}
		}
		return false;
	}

	//****************************************************************//
	//*************** Swapping Timeslots structure *******************//
	//****************************************************************//

	public boolean exploreTSwapsDeterministically(int deb, int fin) {
		SoftSwapTimeslots softmove = (SoftSwapTimeslots) swapt;
		for (int t1 = deb; t1 < fin; t1++) {
			softmove.setT1(t1);
			for (int t2 = t1 + 1; t2 < space.T; t2++) {
				softmove.setT2(t2);
				if (softmove.isKeepingFeas()) {
					int cswap = softmove.evalSoftCost();
					boolean accept = mh.accept(softmove, cswap);
					if (cswap <= 0 || accept) { //accept side walk steps
						time1 = t1;
						return true;
					} // don't store swap of timeslots
				}
			}
		}
		return false;
	}

	//****************************************************************//
	//*************** SoftLNS ****************************************//
	//****************************************************************//
	protected SoftCostProjection scp;
	public static int nodelimit = 5000;
	public static int discrepency = 45;
	public static double frequencyRatio = 1d / 200d;
	public static boolean optHeuristic = false;
	public static boolean stopAtFirstImprovingSol = false;
	//statistic on the cp solver
	public int nbCPMove = 0;
	public double sumnodes;  //sum of the search space size
	public int nbtthresholdReached;
	public double time = 0;
	public int nbNonImprovLNS = 0;

	//1- increase the size of the neigbourhood after 200
	//non improving move
	public static boolean growingSizeN = false;
	public static int grow_threshold = 40; //400
	public int nbThreshSinceLastGrowing = 0;
	public int nbZeroNodeLastGrowing = 0;

	//2- nature of the neigbourhood
	// - per student or per timeslot
	protected static boolean eventNeighbourhood = false;

	//3- diversification step with LNS
	public static int diver_threshold = 30;
	public static boolean diversification = false;
	public static int amountOfD = 10;

	//4- nature of neigbhorhood (cost+promise or cost+random
	public static boolean roomn = false;

	public void printStateLNS() {
		System.out.println(nbCPMove + " " + nbNonImprovLNS);
	}

	public void preActions() {
		if (nbNonImprovLNS != 0) {
			if (growingSizeN &&
				((nbNonImprovLNS % grow_threshold == 0 &&
				nbThreshSinceLastGrowing <= 0.15d * grow_threshold) || 
				(nbCPMove >= 10 && nbZeroNodeLastGrowing >= 0.5 * nbCPMove))) {
				//System.out.println(nbThreshSinceLastGrowing + " " + (nbZeroNodeLastGrowing) + " " + (0.4 * nbCPMove) + " " + scp.nbct + " " + scp.s_nbct);
				scp.growNeigbourhood();
				nbThreshSinceLastGrowing = 0;
			}
			if (diversification && nbNonImprovLNS % diver_threshold == 0) {
				diversificationLNS();
			}
		}
	}

	public void updateFromAction(int delta, int timeused) {
		if (delta == Integer.MAX_VALUE || getCost() + delta >= getBestCost()) {
			nbNonImprovLNS++;
		} else {
			nbNonImprovLNS = 0;
		}
		int node = ((SoftCPMove) cpmove).getNbNode();
		nbCPMove++;
		sumnodes += node;
		if (node >= nodelimit) {
			nbtthresholdReached += 1;
			nbThreshSinceLastGrowing += 1;
		}
		if (node == 0) {
			nbZeroNodeLastGrowing++;
		}
		time += timeused;
	}

	/**
	 * Attempt a LNS move calling the CP solver
	 */
	public boolean exploreLNS() {
		int stime = (int) System.currentTimeMillis();
		preActions();
		SoftCPMove softcpmove = (SoftCPMove) cpmove;

		BitSet fevts;
		if (roomn) fevts = scp.computeEventsToRelaxBasedOnTimeslotAndRooms();
		else fevts = scp.computeEventsToRelaxBasedOnTimeslot();
		cpmove.freeEvents(fevts);
		int cpcost = softcpmove.evalSoftCost();
		//System.out.println("------------");
		boolean accept = mh.accept(softcpmove, cpcost) && cpcost != Integer.MAX_VALUE;
//		System.out.println("accept LNS move by degarding " + cpcost + " ");
		updateFromAction(cpcost, ((int) System.currentTimeMillis() - stime));
		if (cpcost <= 0 || accept) { //always perform a move except if it is infeasible
			return true;
		} else {
			return false;
		}
	}

	public void diversificationLNS() {
		int k = 0;
		while (k < amountOfD) {
			exploreTranslationDeterministically();
			transe.performMove();
			k++;
		}
	}


	public void storeNewImprovingSol() {
		super.storeNewImprovingSol();
		newImprovingSolJustFound = true;
	}

//****************************************************************//
//*************** Storage of solution ****************************//
//****************************************************************//


	/**
	 * A structure to sav a feasible solution and restore it easily
	 */
	public class HardSolSAV {
		public int[][] matrix;
		public int[] nbept;
		public int[] rooms;
		public int[] timeslot;
		public int[][] listevts;

        public HardSolSAV() {}        

        public HardSolSAV(SolutionSpace space) {
			matrix = new int[space.T][space.R];
			listevts = new int[space.T][space.R];
			rooms = new int[space.E];
			timeslot = new int[space.E];
			nbept = new int[space.T];

		}

		public void save(SolutionSpace space) {
			System.arraycopy(space.rooms, 0, rooms, 0, space.E);
			System.arraycopy(space.timeslot, 0, timeslot, 0, space.E);
			System.arraycopy(space.nbept, 0, nbept, 0, space.T);
			for (int i = 0; i < space.T; i++) {
				System.arraycopy(space.matrix[i], 0, matrix[i], 0, space.R);
				System.arraycopy(space.listevts[i], 0, listevts[i], 0, space.R);
			}
		}

		public void restore(SolutionSpace space) {
			System.arraycopy(rooms, 0, space.rooms, 0, space.E);
			System.arraycopy(timeslot, 0, space.timeslot, 0, space.E);
			System.arraycopy(nbept, 0, space.nbept, 0, space.T);
			for (int i = 0; i < space.T; i++) {
				System.arraycopy(matrix[i], 0, space.matrix[i], 0, space.R);
				System.arraycopy(listevts[i], 0, space.listevts[i], 0, space.R);
			}
		}

	}

}