package timetabling.graphics;

import java.awt.*;
import java.awt.event.*;
import java.util.ArrayList;

import javax.swing.*;

import timetabling.recover.DiaryEvent;

public class DiaryDisplayOption extends JPanel implements ActionListener
{
	private static final long serialVersionUID = 1L;
	
	private int[] settings;
	
	private String[] comboBoxContent = {"General Timetabling", "Student Timetabling"};
	private JComboBox choiceDisplaying;
	
	private JLabel choiceStudentEdit;
	private int nbStudent;
	private JTextField choiceStudent;
	
	private JButton applyOption;
	
	private DiaryFrame diaryFrame;
	
	public DiaryDisplayOption(DiaryFrame diaryFrame)
	{
		this.diaryFrame = diaryFrame;
		nbStudent = 0;
		
		this.setLayout(new GridLayout(4, 1, 5, 5));
		
		choiceDisplaying = new JComboBox(comboBoxContent);
		choiceDisplaying.setActionCommand("Combo");
		choiceDisplaying.addActionListener(this);
		this.add(choiceDisplaying);
	}
	
	public void displayStudentOption()
	{
		choiceStudentEdit = new JLabel("Student Selection");
		choiceStudentEdit.setHorizontalAlignment(SwingConstants.CENTER);
		choiceStudentEdit.setHorizontalTextPosition(SwingConstants.CENTER);
		this.add(choiceStudentEdit);
		
		choiceStudent = new JTextField();
		choiceStudent.setText(String.valueOf(0));
		choiceStudent.setHorizontalAlignment(SwingConstants.CENTER);
		choiceStudent.setActionCommand("S");
        choiceStudent.addActionListener(this);
        this.add(choiceStudent);      
        
        applyOption = new JButton("Apply");
		applyOption.addActionListener(this);
		this.add(applyOption);
		
		if(settings != null)
		{
			choiceStudent.setEnabled(true);
			choiceStudentEdit.setForeground(Color.BLACK);
			applyOption.setEnabled(true);
		}
		else 
		{
			choiceStudent.setEnabled(false);
			choiceStudentEdit.setForeground(Color.LIGHT_GRAY);
			applyOption.setEnabled(false);
		}
	}
	
	public void setSettings(int[] settings)
	{
		this.settings = settings;
    }
	
	public void changeStatus() 
	{
		if(choiceDisplaying != null && choiceStudent != null)
		{
			choiceDisplaying.setEnabled(!diaryFrame.superFrame.isVisible());
			choiceStudent.setEnabled(!diaryFrame.superFrame.isVisible());
			applyOption.setEnabled(!diaryFrame.superFrame.isVisible());
		}
	}
	
	public void actionPerformed(ActionEvent event)
	{
		String sentence = event.getActionCommand();
		
		/*// Text Field Event
		if(sentence == "S")
		{
			JTextField result = (JTextField) event.getSource();
			nbStudent = Integer.parseInt(result.getText());	
			
			if(settings != null && settings[3] < nbStudent)
				JOptionPane.showMessageDialog(null,
							" The defined student number does not exist	\n\n"
							+" Please retry	", "Warning", JOptionPane.ERROR_MESSAGE);				
		}*/
		
		// Combo Box Event
		if(sentence == "Combo")
		{
			if(choiceStudentEdit != null && choiceStudent != null && applyOption != null)
			{
				this.remove(choiceStudentEdit);
				this.remove(choiceStudent);
				this.remove(applyOption);
			}
			
			JComboBox combo = (JComboBox)event.getSource();
			if(combo.getSelectedIndex() == 1)
				displayStudentOption();
			else
				if(diaryFrame.launchTreatment != null)
				{
					ArrayList<DiaryEvent> listEvent = diaryFrame.launchTreatment.getListEvent();
					diaryFrame.getTimeTabling().Change(listEvent);
				}
			
			this.validate();
			this.repaint();
		}
		
		// Apply Choice Event
		if(sentence == "Apply")
		{
			ArrayList<DiaryEvent> listEvent = diaryFrame.launchTreatment.getListEvent();
			if(choiceDisplaying.getSelectedIndex() == 1)
			{
				nbStudent = Integer.parseInt(choiceStudent.getText());	
				if(settings != null && settings[3] < nbStudent)
					JOptionPane.showMessageDialog(null,
								" The defined student number does not exist	\n\n"
								+" Please retry	", "Warning", JOptionPane.ERROR_MESSAGE);
				
                diaryFrame.getTimeTabling().ChangeForStudent(listEvent, nbStudent);
			}
		}
	}
}
