package timetabling.graphics;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.io.*;
import java.util.*;

import javax.imageio.*;
import javax.swing.*;
import timetabling.recover.*;
import tools.InstanceReader;
import instance.TimetablingInstance;

public class DiaryFrame extends JFrame implements ActionListener, DiaryConst
{
	private static final long serialVersionUID = 1L;

	public Container desktop;
	private DiaryMenu menu;
	protected DiaryGlassPane superFrame;
	
	private DiaryTreatmentSettings panel1;
	private DiaryFileSettings panel2;
	private DiaryDisplayOption panel3;
	
	private JPanel panel4, subpane;
	protected DiaryPanel1 subpane1;
	
	private JLabel[] hourDisplaying;
	private Point size;
	
	private JFileChooser select;
	private DiaryFileFilter zFilter1 = new DiaryFileFilter(".tim",  "Input Format Instance");
	private DiaryFileFilter zFilter2 = new DiaryFileFilter(".sln",  "Output Format");
	private File inputFile;
	protected int[] inputFileSettings;
	
	protected SolverThread launchTreatment;
	protected TimetablingInstance instance;
    private int nbSave, nbExecution;
	
	public DiaryFrame()
	{
		this.setTitle("Timetabling Problem Project - 4C 2009");
		this.setSize(1024, 768);
		this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		this.setLocationRelativeTo(null);
		this.setResizable(false);

		//----- Loading Panel -----\\
		
		superFrame = new DiaryGlassPane(this);
		superFrame.setVisible(false);
		this.add(superFrame);
		
		desktop = this.getContentPane();
		desktop.setLayout(null);
		
		menu = new DiaryMenu(this);
		this.setJMenuBar(menu);
		
		//----- Treatment Settings subpane -----\\
		
		panel1 = new DiaryTreatmentSettings(this);
		panel1.setBorder(BorderFactory.createTitledBorder("Process settings"));
		panel1.setBounds((this.getWidth()-(2*(this.getWidth()/3)))-190, 15, 190, 130);
		desktop.add(panel1);
		
		//----- Instance Settings -----\\
		
		panel2 = new DiaryFileSettings(this);
		panel2.setBorder(BorderFactory.createTitledBorder("Instance Settings"));
		panel2.setBounds((this.getWidth()-300)/2, 15, 300, 130);
		desktop.add(panel2);
		
		//----- Displaying Options Panel -----\\
		
		panel3 = new DiaryDisplayOption(this);
		panel3.setBorder(BorderFactory.createTitledBorder("Displaying Options"));
		panel3.setBounds((this.getWidth()-(this.getWidth()/3)), 15, 190, 130);
		desktop.add(panel3);
		
		//----- Diary subpane -----\\
		
		size = new Point(this.getWidth()-20, ((this.getHeight()*2)/3)-40);
		
		panel4 = new JPanel();
		panel4.setBorder(BorderFactory.createTitledBorder("Week diary"));
		panel4.setBounds((this.getWidth()-size.x)/2, (this.getHeight()-size.y)/2+30, size.x, size.y);
		panel4.setLayout(new BorderLayout());
		desktop.add(panel4);
		
		subpane1 = new DiaryPanel1(this);
		subpane1.setBackground(Color.WHITE);
		subpane1.setBorder(BorderFactory.createEtchedBorder());
		panel4.add(subpane1, BorderLayout.CENTER);
		
		subpane = new JPanel();
		subpane.setLayout(new GridLayout(9, 1, 0, 0));
		panel4.add(subpane, BorderLayout.WEST);
		
		hourDisplaying = new JLabel[nbHour];
		for(int i = 0; i < hourDisplaying.length; i++)
		{
			if(i < 5)
				hourDisplaying[i] = new JLabel(String.valueOf(i+8)+".00 AM");
			else
				hourDisplaying[i] = new JLabel(String.valueOf(i-4)+".00 PM");
			
			hourDisplaying[i].setFont(new Font("Arial", Font.ROMAN_BASELINE, 10));
			hourDisplaying[i].setHorizontalAlignment(SwingConstants.CENTER);
			hourDisplaying[i].setVerticalAlignment(SwingConstants.CENTER);
			
			subpane.add(hourDisplaying[i]);
		}
		
		nbSave = 0;
		
		this.setVisible(true);
	}

    public TimetablingInstance getInstance() {
        return instance;
    }

    public void OpenInstanceAction()
	{		
		if(inputFile != null)
			inputFile = new File("");
		
		select = new JFileChooser();
		select.setCurrentDirectory(new File("./data/Track2"));
		select.setFileFilter(zFilter1);
		
		if (select.showOpenDialog(null)==JFileChooser.APPROVE_OPTION)
		{
			nbSave = 0; nbExecution = 0;
			inputFile = select.getSelectedFile();
			fileSettingsRecover();
            setInstanceRepresentation();
        }
    }
	
	public void OpenSolvedFileAction()
	{		
		if(inputFile != null)
			inputFile = new File("");
		
		select = new JFileChooser();
		select.setCurrentDirectory(new File("./data/Track2"));
		select.setFileFilter(zFilter2);
		
		if (select.showOpenDialog(null)==JFileChooser.APPROVE_OPTION)
		{
			nbSave = 0;
			inputFile = select.getSelectedFile();
			fileSettingsRecover();
            setInstanceRepresentation();
        }
    }

    public void setInstanceRepresentation() 
    {
        	// Update the instance representation
            InstanceReader reader = null;
		    try {
		    	reader = new InstanceReader(new FileInputStream(inputFile));
		    } catch (FileNotFoundException e) {
		    	e.printStackTrace();
	    	}
            instance = reader.getInstance();
    }

    public void fileSettingsRecover()
	{
		menu.setInfo(inputFile.getName());
		
		try 
		{
			inputFileSettings = new int[4];
			Scanner scanner = new Scanner(inputFile);
			
			for(int i = 0; i < inputFileSettings.length; i++)
				inputFileSettings[i] = scanner.nextInt();
			
			subpane1.setSettings(inputFileSettings);
			panel2.setSettings(inputFileSettings);
			panel3.setSettings(inputFileSettings);
		} catch (FileNotFoundException e) 
		{e.printStackTrace();}
		this.validate();
	}
	
	public void launch() 
	{
		if(inputFile != null)
		{
            if(nbExecution != 0)
				fileSettingsRecover();		
			
			//Number of solver execution
			nbExecution++; 
			
			//----- Solver execution -----\\
			
			launchTreatment = new SolverThread(this);
			superFrame.Modify(panel1, launchTreatment);
			launchTreatment.start();
			
			//TODO Disable the different parts of the program during the treatment
			menu.setEnabled(false);			
			subpane1.changeStatus();
			panel1.changeStatus();
			panel2.changeStatus();
			panel3.changeStatus();
		}
		this.validate();
	}
	
	public void stopResolution()
	{
		menu.setEnabled(true);
		superFrame.finalStop();
		
		if (launchTreatment.isAlive())
		{
			launchTreatment.manualStop();
			superFrame.getCostObserver().stop();
		}
		subpane1.changeStatus();
		panel1.changeStatus();
		panel2.changeStatus();
        panel3.changeStatus();
        desktop.validate();
	}
	
	public void ScreenPrintAction() throws AWTException
	{
		nbSave++;
		Robot robot = new Robot();
		BufferedImage image;
		
		image = robot.createScreenCapture(new Rectangle(this.getX(), this.getY(), this.getWidth(), this.getHeight()));
		
		try 
		{ 
			// Save of the new image
			File f = new File("./"+inputFile.getName()+"savedImage_"+nbSave+".png"); ImageIO.write(image, "png",f);
			JOptionPane.showMessageDialog(null,
					"	Image saved into the ./images directory	", "Finished saving", JOptionPane.INFORMATION_MESSAGE);			
		} 
		catch (Exception e)
		{
			// Display of a error window in case of saving problem
			JOptionPane.showMessageDialog(null,
					"	Frame's screenshot process : An error has been meeting	\n"+"\n"
					+"	Please retry	", "Saving problem", JOptionPane.ERROR_MESSAGE);		
		}
	}
	
	public File getInputFile()
	{
		return inputFile;
	}
	
	public DiaryTreatmentSettings getSettings()
	{
		return panel1;
	}
	
	public DiaryPanel1 getTimeTabling()
	{
		return subpane1;
	}
	
	public void actionPerformed(ActionEvent event)
	{
		String sentence = event.getActionCommand();
		
		// Open Instance Event
		if(sentence.equals("O"))
			OpenInstanceAction();
		
		// Open Solved File Event
		if(sentence.equals("L"))
			OpenSolvedFileAction();
		
		// Print Event
		if(sentence.equals("S"))
			try {
				ScreenPrintAction();
			} catch (AWTException e) {
				e.printStackTrace();
			}
		
		// Close event
		if(sentence.equals("Close"))
			System.exit(0);
			
		// About event
		if(sentence.equals("About"))
		{
			JOptionPane.showMessageDialog(null,
			"Post Enrolment based Course Timetabling \n\n"+
			"Hadrien Cambazard, Emmanuel Hebrard, \n Barry O'Sullivan, Alexandre Papadopoulos,\n Nicolas S�guier \n\n"+
			"Cork Constraint Computation Centre\n"+
			"2008 - 2009 \n",
			"About this device", JOptionPane.INFORMATION_MESSAGE, new ImageIcon("./icons/logo.png"));
		}
	}
}
