package timetabling.graphics;

import java.awt.*;
import java.util.*;

import javax.swing.*;

import timetabling.recover.*;

public class DiaryPanel1 extends JPanel implements DiaryConst
{
	private static final long serialVersionUID = 1L;
	
	private DiaryGlassPane glassPane;
	private DiaryFrame mainframe;
	
	protected boolean completeDisplaying;
    private Font font;
	private JPanel subpane1, subpane2;
	
	private DiaryTimeslot[] set;
	private JLabel[] dayDisplaying;


    public DiaryPanel1(DiaryFrame frame)
	{
		this.glassPane = frame.superFrame;
        this.mainframe = frame;
        init();
	}
	
	public void init()
	{
		this.setLayout(new BorderLayout());
		
		completeDisplaying = false;
		
		font = new Font("Arial", Font.BOLD, 11);
		this.setFont(font);
		
		subpane1 = new JPanel();
		subpane1.setLayout(new GridLayout(1, 7, 2, 0));
		subpane1.setBackground(Color.LIGHT_GRAY);
		this.add(subpane1, BorderLayout.NORTH);
		
		dayDisplaying = new JLabel[days.length];
        for(int i = 0; i < dayDisplaying.length; i++)
        {
        	dayDisplaying[i] = new JLabel(days[i]);

        	dayDisplaying[i].setForeground(Color.DARK_GRAY);
        	dayDisplaying[i].setHorizontalAlignment(SwingConstants.CENTER);
        	
        	subpane1.add(dayDisplaying[i]);
        }
		
        subpane2 = new JPanel();
		subpane2.setLayout(new GridLayout(9, 5, 3, 0));
		subpane2.setBackground(Color.WHITE);
		this.add(subpane2, BorderLayout.CENTER);
		
		//----- Second part -----\\
		
		int i = 0;
		set = new DiaryTimeslot[days.length*nbHour];
		
		while(i < nbHour)
		{			
			for(int j = 0; j < days.length; j++)
			{
					set[i+(nbHour*j)] = new DiaryTimeslot(i+(nbHour*j));
					this.subpane2.add(set[i+(nbHour*j)]);
			}
			i++;
		}
	}

    public void empty() 
    {
        for (int i = 0; i < set.length; i++) 
        {
            set[i].removeTimeslots();
        }
    }

    public void Change(ArrayList<DiaryEvent> list)
    {
    	empty();
        int timeslot, room;
    	for(int i = 0; i < list.size(); i++)
    	{
    		room = list.get(i).getRoom();
    		timeslot = list.get(i).getTimeslot();
    		
    		set[timeslot].Add(i, room);
    	}
    	
    	changeDisplayingMode();    		
    	this.repaint();
        this.validate();
    }

    public void ChangeForStudent(ArrayList<DiaryEvent> list, int student)
    {
        empty();
        int timeslot, room;
    	for(int i = 0; i < list.size(); i++)
    	{
           if (mainframe.getInstance().studentAttendsEvent(student,i))
            {
                room = list.get(i).getRoom();
        		timeslot = list.get(i).getTimeslot();
                System.out.println("add event "+i+" in "+timeslot+"timeslot and room "+room+" for student "+student);
        		set[timeslot].Add(i, room);
    		}
    	}
    	
    	changeDisplayingMode();
    	this.repaint();
        this.validate();
    }

	public void setSettings(int[] settings)
	{
		for(int i = 0; i < set.length; i++)
			set[i].setSettings(settings);
	}
	
	public void changeStatus()
	{
		for(int i = 0; i < set.length; i++)
			set[i].changeStatus(glassPane.isVisible());
	}
	
	public void changeDisplayingMode()
	{
		for(int i = 0; i < set.length; i++)
			set[i].changeMode(completeDisplaying);
	}
	
	public DiaryTimeslot[] getTimeTabling()
	{
		return set;
	}
}
