package timetabling.graphics;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;

public class DiaryTimeslot extends JPanel implements MouseListener
{
	private static final long serialVersionUID = 1L;
	
	private int indice, number;
	private int[] settings;
	private ArrayList<Integer> listRoom, listEvent;
	
	private JLabel nbEventLabel;
	private JLabel[] drawRoom;
	
	private Color color;
	private boolean glassPaneVisible;
	private boolean completeDisplayActivated;
	
	public DiaryTimeslot(int indice)
	{
		color = new Color(230, 230, 250);
		this.setFont(new Font("Arial", Font.TRUETYPE_FONT, 11));
		this.setBorder(BorderFactory.createEtchedBorder());
		this.setLayout(new BorderLayout());
		this.setBackground(Color.WHITE);
        this.setForeground(Color.DARK_GRAY);
        this.setOpaque(true);
        this.addMouseListener(this);
        
		this.indice = indice;
		number = 0;
		listRoom = new ArrayList<Integer>();
		listEvent = new ArrayList<Integer>();
		glassPaneVisible = false;
		completeDisplayActivated = false;
		
		nbEventLabel = new JLabel();
		nbEventLabel.setHorizontalAlignment(SwingConstants.CENTER);
		nbEventLabel.setHorizontalTextPosition(SwingConstants.CENTER);
		nbEventLabel.setVisible(false);
		this.add(nbEventLabel, BorderLayout.CENTER);
    }

    public void removeTimeslots() 
    {
        listEvent.clear();
        listRoom.clear();
        
        nbEventLabel.setText("");
        number = 0;
        this.setBackground(Color.WHITE);
        
        if(drawRoom != null)
        	for(int i = 0; i < drawRoom.length; i++)
        		if(drawRoom[i] != null)
        			this.remove(drawRoom[i]);
    }


    public void Add(int event, int room)
	{
		listEvent.add(event);
		listRoom.add(room);
		
		if(number != 0)
			this.setCursor(new Cursor(Cursor.HAND_CURSOR));
		number++;
		
		nbEventLabel.setVisible(true);
		nbEventLabel.setText("Number of Events : "+number);
		partialDisplayMode();
		this.setBackground(color);
		
		this.repaint();
	}
	
	public void setSettings(int[] settings)
	{
		this.settings = settings;
	}
	
	public void changeStatus(boolean statusGlassPane)
	{
		glassPaneVisible = statusGlassPane;
		this.validate();
	}
	
	public void changeMode(boolean statusMode)
	{
		completeDisplayActivated = statusMode;
		if(completeDisplayActivated)
			completeDisplayMode();
		else partialDisplayMode();
		this.repaint();
	}
	
	public void completeDisplayMode()
	{
		if(!glassPaneVisible)
		{
			this.remove(nbEventLabel);
			this.setLayout(new GridLayout(2, 10));
			if(settings != null)
			{
				drawRoom = new JLabel[settings[1]];
			    for(int i = 0; i < drawRoom.length; i++)
			    {
			    	drawRoom[i] = new JLabel();
			    	drawRoom[i].setBorder(BorderFactory.createEtchedBorder());
			    	drawRoom[i].setFont(new Font("Arial", Font.TRUETYPE_FONT, 8));
			    	drawRoom[i].setHorizontalAlignment(SwingConstants.CENTER);
			    	drawRoom[i].setHorizontalTextPosition(SwingConstants.CENTER);
			    	
			    	this.add(drawRoom[i]);
			    }
			    if(number != 0)
			    	drawUsedRoom();
			}
			this.repaint();
		}
	}
	
	public void partialDisplayMode()
	{
		if(!glassPaneVisible)
		{
			this.removeAll();
			if(number != 0)
				this.setBackground(color);
			else this.setBackground(Color.WHITE);
			
			nbEventLabel.setHorizontalAlignment(SwingConstants.CENTER);
			nbEventLabel.setHorizontalTextPosition(SwingConstants.CENTER);
			this.add(nbEventLabel, BorderLayout.CENTER);
			this.repaint();
		}
	}
	
	public void drawUsedRoom()
	{
		for(int j = 0; j < listRoom.size(); j++)
		{
			drawRoom[listRoom.get(j)].setBackground(Color.BLACK);
			drawRoom[listRoom.get(j)].setText(String.valueOf(listEvent.get(j)));
		}
	}
	
	public ArrayList<Integer> getListEvent()
	{
		return listEvent;
	}

	public ArrayList<Integer> getListRoom() 
	{
		return listRoom;
	}
	
    public void mouseMoved(MouseEvent event) 
    {
    	
    }

	public void mouseClicked(MouseEvent event) 
	{
		if(number != 0 && !glassPaneVisible)
		{
			String details = "";
			String title = "Timeslot n"+indice;
			
			for(int i = 0; i < number; i++)
				details += "Event n : "+listEvent.get(i)+"\n"+
							"Room n : "+listRoom.get(i)+"\n"+"\n";
			
			JOptionPane.showMessageDialog(null, details, title, JOptionPane.PLAIN_MESSAGE, null);
		}
	}

	public void mouseEntered(MouseEvent event) 
	{
		if(!glassPaneVisible)
		{
			this.setBackground(new Color(255, 222, 173));
			if(!completeDisplayActivated)
				completeDisplayMode();
		}
	}

	public void mouseExited(MouseEvent event)
	{
		if(!glassPaneVisible)
		{
			if(number == 0)
				this.setBackground(Color.WHITE);
			else this.setBackground(color);
			
			if(!completeDisplayActivated)
				partialDisplayMode();
		}
	}

	public void mousePressed(MouseEvent event) 
	{
		
	}

	public void mouseReleased(MouseEvent event) 
	{
		
	}
}
