package timetabling.graphics;

import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.*;

import timetabling.export.CreationExportFile;

public class DiaryToolBar extends JToolBar implements ActionListener
{
	private static final long serialVersionUID = 1L;
	
	private DiaryFrame frame;
	
	private ImageIcon[] buttonIcon;
	private JButton[] button;
	
	private JLabel editInstanceName;
	private JTextField displayInstanceName;
	
	public DiaryToolBar(DiaryFrame frame)
	{
		this.frame = frame;
		
		ImageIcon[] buttonIcon = new ImageIcon[3];
		JButton[] button = new JButton[3];
		
		buttonIcon[0] = new ImageIcon("./icons/treatment.png");
		button[0] = new JButton("Solve");
		button[0].setIcon(buttonIcon[0]);
		button[0].setCursor(new Cursor(Cursor.HAND_CURSOR));
		button[0].addActionListener(this);
		this.add(button[0]);
		
		this.addSeparator();
		
		buttonIcon[1] = new ImageIcon("./icons/export.png");
		button[1] = new JButton("Google Export");
		button[1].setIcon(buttonIcon[1]);
		button[1].setCursor(new Cursor(Cursor.HAND_CURSOR));
		button[1].addActionListener(this);
		button[1].setActionCommand("E");
		this.add(button[1]);
		
		this.addSeparator();
		
		editInstanceName = new JLabel("Actual imported file : ");
		this.add(editInstanceName);
		
		displayInstanceName = new JTextField(" No imported file");
		displayInstanceName.setEditable(false);
		this.add(displayInstanceName);
		
		this.addSeparator();
		
		buttonIcon[2] = new ImageIcon("./icons/loop.png");
		button[2] = new JButton("Detailed Mode");
		button[2].setIcon(buttonIcon[2]);
		button[2].setCursor(new Cursor(Cursor.HAND_CURSOR));
		button[2].setActionCommand("D");
		button[2].addActionListener(this);
		this.add(button[2]);
		
		this.addSeparator(new Dimension(300, 20));
		
		this.setVisible(true);
	}
	
	public void changeName(String name)
	{
		displayInstanceName.setText(" "+name);
		this.validate();
	}
	
	public void actionPerformed(ActionEvent event) 
	{
		String phrase = event.getActionCommand();
		
		// Treatment Launching Event
		if(phrase.equals("Solve"))
		{
			frame.launch();
		}
		
		// Export Calendar Event
		if(phrase.equals("E"))
		{			
			CreationExportFile export = new CreationExportFile(frame);
			export.createFile();
		}
		
		// Enable or Disable the detailed mode of timeslot's displaying
		if(phrase.equals("D"))
		{
			frame.subpane1.completeDisplaying = !frame.subpane1.completeDisplaying;
			frame.subpane1.changeDisplayingMode();
		}
	}
}
