package timetabling.recover;

import java.util.*;
import javax.swing.*;

import timetabling.graphics.*;
import localsearch.*;
import localsearch.solvers.complete.SearchLoop;
import localsearch.solvers.complete.AbstractSolver;
import localsearch.solvers.complete.SoftSolver;
import localsearch.solvers.complete.SoftCostProjection;

public class SolverThread extends Thread
{
	protected DiaryFrame frame;
	protected SearchLoop ssloop;
	
	private ArrayList<DiaryEvent> listEvent;
	
	private int timelimit;
	private boolean hardSolver, hardPrediction;
	
	protected boolean searchOver;
	
	public SolverThread(DiaryFrame frame) 
	{
		this.frame = frame;
		
		timelimit = (int) (frame.getSettings().timelimit * 1000);
		hardSolver = frame.getSettings().hardSolver;
		hardPrediction = frame.getSettings().hardPrediction;
	}
	
	public void run() 
	{
		listEvent = new ArrayList<DiaryEvent>();
		
		// Parameter settings of the treatment
		Benchmark.ech = 1;
		Benchmark.seed = 0;
        Benchmark.checkResults = false;
		
        ssloop = new SearchLoop(frame.getInputFile().getPath(), 0, 0);
        ssloop.setParamSolver(timelimit, 10);
        SearchLoop.searchLoopTrace = true;
        AbstractSolver.trace = true;
        AbstractSolver.debug = false;
        
        searchOver = false;    
        
        // Recovery of the data (if an solution is found)
        if(ssloop.solveLoop())
        {
            //SoftCostProjection sproj = new SoftCostProjection(ssloop.softS);
            //sproj.projectCost();
            //sproj.get
            for(int i = 0; i < ssloop.softS.space.E; i++)
            {
                Place p = ssloop.softS.getPosition(i);
        		listEvent.add(new DiaryEvent(p.t, p.r));
        	}
        	//TODO
            frame.getTimeTabling().Change(listEvent);
            System.out.println("Instance resolue");
        } 
        else 
        {
        	JOptionPane.showMessageDialog(null,
					"	No Solution found	\n\n"
					+"	Please retry	", "Solver result", JOptionPane.ERROR_MESSAGE);
        }
		
		// End of treatment and displaying
        searchOver = true;
        frame.stopResolution();
    }
	
	public void manualStop()
	{
        ssloop.stopByChangingTimeLimit();
        ssloop.hardS.setTimelimit(0);
    	if(ssloop.softS != null)
    		ssloop.softS.setTimelimit(0);
	}
	
	public ArrayList<DiaryEvent> getListEvent()
	{
		return listEvent;
	}
}
