/**
 *
 */
package tools;

// coucou
/**
 * C'est a moi qu'il est adress ce coucou? 
 */

import instance.Precedence;
import instance.TimetablingInstance;

import java.io.FileInputStream;
import java.io.InputStream;
import java.io.File;
import java.io.FileFilter;
import java.util.InputMismatchException;
import java.util.NoSuchElementException;
import java.util.Scanner;


/**
 * A utility that reads a tim file and builds
 * the corresponding timetabling instance.
 *
 * @author alexp
 */
public class InstanceReader {

	private Scanner in;

	protected TimetablingInstance ins;

	/**
	 * Creates a new instance reader for
	 * the specified source
	 *
	 * @param source any input stream containing
	 *               an instance in the tim format
	 */
	public InstanceReader(InputStream source) {
		in = new Scanner(source);
	}


	/**
	 * Builds the instance while reading the input,
	 * and returns it.
	 *
	 * @return the timetabling instance
	 */
	public TimetablingInstance getInstance() {
		int nEvents, nRooms, nFeatures, nStudents;
		nEvents = in.nextInt();
		nRooms = in.nextInt();
		nFeatures = in.nextInt();
		nStudents = in.nextInt();

		ins = new TimetablingInstance(nEvents, nRooms, nFeatures, nStudents);

		try {
			for (int room = 0; room < nRooms; room++)
				ins.setRoomSize(room, in.nextInt());

			for (int student = 0; student < nStudents; student++)
				for (int event = 0; event < nEvents; event++)
					if (in.nextInt() == 1)
						ins.setStudentForEvent(student, event);

			for (int room = 0; room < nRooms; room++)
				for (int feature = 0; feature < nFeatures; feature++)
					if (in.nextInt() == 1)
						ins.setRoomForFeature(room, feature);

			for (int event = 0; event < nEvents; event++)
				for (int feature = 0; feature < nFeatures; feature++)
					if (in.nextInt() == 1)
						ins.setEventForFeature(event, feature);

			for (int event = 0; event < nEvents; event++) {
				for (int timeslot = 0; timeslot < TimetablingInstance.N_TIMESLOTS; timeslot++) {
						if (in.nextInt() == 1)
							ins.setEventForTimeslot(event, timeslot);
				}
			}

			for (int firstEvent = 0; firstEvent < nEvents; firstEvent++)
				for (int secondEvent = 0; secondEvent < nEvents; secondEvent++)
					if (in.nextInt() == 1)
						ins.addPrecedence(firstEvent, secondEvent);
			
			ins.fillEventRoomMatrix();

		} catch (InputMismatchException e) {
			e.printStackTrace();
		} catch (NoSuchElementException e) {
			//e.printStackTrace();
		}

		return ins;
	}


	public static File[] listAllFiles(File dir) {
	    FileFilter fileFilter = new FileFilter() {
	        public boolean accept(File file) {
	            return !file.isDirectory() && file.getName().endsWith(".tim");
	        }
	    };
	    return dir.listFiles(fileFilter);
	}
	

	public static void main(String[] args) throws Exception {
		InstanceReader reader = new InstanceReader(new FileInputStream("data/Track2/comp-2007-2-1.tim"));
		TimetablingInstance ins = reader.getInstance();

		for (Precedence p : ins.eventPrecedences()) {
			System.out.println(ins.eventsShareStudents(p.firstEvent, p.secondEvent));
		}
	}


}
